#include "Segment.h"
#include "Treap.h"
#include "poly.h"
#include <queue>
#include <map>

#include "curve.h"

class Embedding : public RefCount {
public:
  static const double TELOMERE_LENGTH;
  static int nEmbedding;

  Embedding (const Embedding &set) { assert(0); }
  void operator= (const Embedding &set) { assert(0); }

  class Edge;

  class Vert {
  public:
    static Vert vert;

    Vert (Point p, bool dummy) : p(p) {}
    Vert (Point p) { *this = vert; this->p = p; }
    ~Vert () {}

    Point p;
    // edge[2][0/1] is CCWmost upper/lower edge
    Edge *edge[3 /*rlm*/][2 /*hilo*/];

    class CompareXY {
    public:
      bool operator() (const Vert *u, const Vert *v) {
	if (u->p.x != v->p.x)
	  return u->p.x < v->p.x;
	return u->p.y < v->p.y;
      }
    };

    class CompareYX {
    public:
      bool operator() (const Vert *u, const Vert *v) {
	if (u->p.y != v->p.y)
	  return u->p.y < v->p.y;
	return u->p.x < v->p.x;
      }
    };

    double height ();

    Edge *edgeAbove (int hilo);
  };

  typedef set<Vert *, Vert::CompareXY> Verts;
  typedef Verts::iterator Verti;
  Verts verts;

  Vert *addVert (Point p);

  class SweepNode;

  class Edge {
    Vert *vert[2 /*tailhead*/];
  public:
    // debug
    static int nextSSN;
    int ssn;

    static int nEdge;
    Ptr<Function> f;
    int mult, id;
    int loop[2];
    int height;
    int info;
    Edge *edge[2 /*tailhead*/][2 /*hilo*/];
    SweepNode *node;

    Edge (const Edge &that) {
      nEdge++;
      ssn = nextSSN++;
      *this = that;
    }

    Edge (Vert *tail, Vert *head, Ptr<Function> f,
	  int id=0, int mult=1, int info=0)
      : f(f), height(-1), id(id), info(info), node(0) {
      nEdge++;
      ssn = nextSSN++;

      assert(tail->p != head->p);
      if (Vert::CompareXY()(tail, head)) {
	vert[0] = tail;
	vert[1] = head;
	this->mult = mult;
      }
      else {
	vert[0] = head;
	vert[1] = tail;
	this->mult = -mult;
      }
      edge[0][0] = edge[0][1] = edge[1][0] = edge[1][1] = 0;
      if (f == 0)
	assert(vert[0]->p.x == vert[1]->p.x);
      else {
	assert(vert[0]->p.x < vert[1]->p.x);
	checkDistance(f, vert[0]->p);
	checkDistance(f, vert[1]->p);
      }
	
      loop[0] = loop[1] = -1;
    }

    ~Edge () { delete node; nEdge--; }

    static bool compareF (const Edge *a, const Edge *b) {
      if (a->f == 0) {
	assert(b->f == 0);
	if (a->vert[0]->p.x != b->vert[0]->p.x)
	  return a->vert[0]->p.x < b->vert[0]->p.x;
	if (a->id != b->id)
	  return a->id < b->id;
	if (a->vert[0]->p.y != b->vert[0]->p.y)
	  return a->vert[0]->p.y < b->vert[0]->p.y;
	if (a->vert[1]->p.y != b->vert[1]->p.y)
	  return a->vert[1]->p.y < b->vert[1]->p.y;
      }
      else {
	assert(!(b->f == 0));
	if (a->f != b->f)
	  return a->f < b->f;
	if (a->id != b->id)
	  return a->id < b->id;
	if (a->vert[0] != b->vert[0])
	  return Vert::CompareXY()(a->vert[0], b->vert[0]);
	if (a->vert[1] != b->vert[1])
	  return Vert::CompareXY()(a->vert[1], b->vert[1]);
      }
      return false;
    }

    bool equalF (const Edge *that) const {
      if (f != 0)
	return f == that->f;
      return vert[0]->p.x == that->vert[0]->p.x;
    }

    static bool compareX (const Edge *a, const Edge *b) {
      assert(a->f == 0 && b->f == 0);
      if (a->vert[0]->p.x != b->vert[0]->p.x)
	return a->vert[0]->p.x < b->vert[0]->p.x;
      if (a->id != b->id)
	return a->id < b->id;
      if (a->vert[0] != b->vert[0])
	return Vert::CompareXY()(a->vert[0], b->vert[0]);
      if (a->vert[1] != b->vert[1])
	return Vert::CompareXY()(a->vert[1], b->vert[1]);
      return false;
    }

    class CompareF {
    public:
      bool operator() (const Edge *a, const Edge *b) {
	return compareF(a, b);
      }
    };

    class CompareH {
    public:
      bool operator() (const Edge *a, const Edge *b) {
	return a->height < b->height;
      }
    };

    class CompareX {
    public:
      bool operator() (const Edge *a, const Edge *b) {
	return compareX(a, b);
      }
    };

    static bool compareY (const Edge *a, const Edge *b);
    static bool compareH (const Edge *a, const Edge *b);

    class CompareY {
    public:
      bool operator() (const Edge *a, const Edge *b) {
	return compareY(a, b);
      }
    };

    bool isVertical () { return f == 0; }

    double getY (double x) { return f->getY(x); }

    double getDY (double x) { return f->getDY(x); }
    double getDY (double x, double d) {
      double y = f->getY(x);
      return max(f->distance(Point(x, y+d)), f->distance(Point(x, y-d)));
    }

    double minY () { return std::min(vert[0]->p.y, vert[1]->p.y); }

    double maxY () { return std::max(vert[0]->p.y, vert[1]->p.y); }

    double bestY (Edge *that, double x) {
      Ptr<Function> g = that->f;
      if (f->nearestRoot(g, x) == x)
	return f->getRoot(g, x).y;
      double fx = f->getY(x);
      double gx = g->getY(x);
      double ret;
#define USE_DY_ROOT
#ifdef USE_DY_ROOT
      double df = fabs(f->getDY(x));
      double dg = fabs(g->getDY(x));
      if (df > dg)
	ret = gx;
      else
	ret = fx;
#else
      double distf = fabs(f->distance(Point(x, gx)));
      double distg = fabs(g->distance(Point(x, fx)));
      if (distf < distg)
	ret = gx;
      else
	ret = fx;
#endif
      return ret;
    }

    int compare (Edge *that, double x, int ler) {
      return f->compare(that->f, x, ler);
    }

    void dumpPolys (Function *f, Function *g);

    Point checkDist (Function *f, Function *g, Point p) {
      static double maxRootDist;
      assert(p.x != Function::FINFINITY);
      double distF = f->distance(p);
      double distG = g->distance(p);
      double dist = distF > distG ? distF : distG;
      if (maxRootDist < dist) {
	maxRootDist = dist;
	std::cout << "maxRootDist " << maxRootDist << std::endl;
	// debug
	double fx = f->getY(p.x);
	double gx = g->getY(p.x);
	double dfx = f->getDY(p.x);
	double dgx = g->getDY(p.x);
	double distF2 = f->distance(Point(p.x, gx));
	double distG2 = g->distance(Point(p.x, fx));
	if (maxRootDist > 1e-3) {
	  int cmp = f->compare(g, p.x, 1);
	  dumpPolys(f, g);
	  assert(0);
	}
      }
      return p;
    }

    Point checkDist2 (Function *f, Function *g, Point p) {
      static double maxRootDist2;
      assert(p.x != Function::FINFINITY);
      double distF = f->distance(p);
      double distG = g->distance(p);
      double dist = distF > distG ? distF : distG;
      if (maxRootDist2 < dist) {
	maxRootDist2 = dist;
	std::cout << "maxRootDist2 " << maxRootDist2 << std::endl;

	if (maxRootDist2 > 1e-3) {
	  dumpPolys(f, g);
	  assert(0);
	}
      }
      return p;
    }

    Point nextRoot (Edge *that, double x) {
      // Check ranges against x.
      Point p(Function::FINFINITY, Function::FINFINITY);
      assert(x >= this->vert[0]->p.x && x >= that->vert[0]->p.x);
      if (x >= this->vert[1]->p.x || x >= this->vert[1]->p.x)
	return p;
      if (this->f == that->f)
	return p;
      if (this->f->getMaxX() <= that->f->getMinX() ||
	  that->f->getMaxX() <= this->f->getMinX())
	return p;

      // Immediate swap.
      int cmp = this->f->compare(that->f, x, 1);
      assert(cmp != 0);
      if (cmp > 0) {
	p.x = x;
	p.y = bestY(that, x);
	return checkDist(this->f, that->f, p);
      }
	
      // Nominal root.
      double r = f->nextRoot(that->f, x);
      if (r >= this->vert[1]->p.x || r >= that->vert[1]->p.x)
	return p;
      return checkDist2(this->f, that->f, f->getRoot(that->f, r));
    }

    bool isTelomere () const {
      return mult == 0;
    }

    void sort(set<Edge *> *los, Edge **sorted, int &nsorted);

    Vert *tail () { return vert[0]; }
    Vert *head () { return vert[1]; }
    Vert *getVert (bool th) { return vert[th]; }

    void setTail (Vert *v) {
      // assert(v != vert[1]);
      if (f == 0)
	assert(v->p.x == tail()->p.x);
      else 
	checkDistance(f, v->p);
      vert[0] = v;

#ifdef BLEEN
      if (vert[0]->p.x == vert[1]->p.x) {
	f = 0;
	if (vert[0]->p.y > vert[1]->p.y) {
	  Vert *temp = vert[0];
	  vert[0] = vert[1];
	  vert[1] = temp;
	  mult = -mult;
	}
      }
#endif
    }
	       
    void setHead (Vert *v) {
      // assert(v != vert[0]);
      if (f == 0)
	assert(v->p.x == head()->p.x);
      else 
	checkDistance(f, v->p);
      vert[1] = v;

#ifdef BLEEN
      if (vert[0]->p.x == vert[1]->p.x) {
	f = 0;
	if (vert[0]->p.y > vert[1]->p.y) {
	  Vert *temp = vert[0];
	  vert[0] = vert[1];
	  vert[1] = temp;
	  mult = -mult;
	}
      }
#endif
    }

    void setVert (bool th, Vert *v) {
      // if (ssn == 29752)
      // cout << "this is it" << endl;
      if (f == 0)
	assert(v->p.x == vert[0]->p.x);
      else 
	checkDistance(f, v->p);
      vert[th] = v;
    }
	       
    Edge *&aboveTail () { return edge[0][0]; }
    Edge *&belowTail () { return edge[0][1]; }
    Edge *&aboveHead () { return edge[1][0]; }
    Edge *&belowHead () { return edge[1][1]; }
    Edge *&subEdge () { return edge[0][0]; }
    Edge *&nextSweep () { return edge[1][0]; }
    Edge *&prevSweep () { return edge[1][1]; }

    void setEdgeAboveBelowTail () {
      Edge *next = nextSweep();
      if (next != 0)
	subEdge()->aboveTail() = next->subEdge();
      Edge *prev = prevSweep();
      if (prev != 0)
	subEdge()->belowTail() = prev->subEdge();
    }

    void setEdgeAboveBelowHead () {
      Edge *next = nextSweep();
      if (next != 0)
	subEdge()->aboveHead() = next->subEdge();
      Edge *prev = prevSweep();
      if (prev != 0)
	subEdge()->belowHead() = prev->subEdge();
    }

    SweepNode *&nodeID () {
      return *(SweepNode **) &edge[0][0];
    }

    Vert *&above () {
      return *(Vert **) &edge[0][1];
    }

    double implicitDistance (Function *f, Point p);

    void checkDistance (Function *f, Point p) {
      // DEBUG
      return;
      static double maxDistance;
      double dist = f->getType() == Function::IMPLICIT ?
	implicitDistance(f, p) : f->distance(p);
	  
      if (maxDistance < dist) {
	maxDistance = dist;
	std::cout << "maxDistance " << maxDistance << std::endl;
	if (maxDistance > 0.001)
	  cout << "this is it" << endl;
      }
    }
  };

  // non-vertical and vertical edges
  vector<Edge *> edges, edgesV;

  int numAdded;

  Edge *addEdge (Point tail, Point head, Ptr<Function> f,
		 int id=0, int mult=1, int info=0);

  void checkMults ();

  Embedding *emb[2];
  bool ignoreIDs, sumInfo;
  void arrange1d (vector<Edge *> &edges);

  class SweepNode : public Treap {
  public:
    Edge *edge;

    SweepNode (Edge *edge) : edge(edge) { edge->node = this; }
    ~SweepNode () { edge->node = 0; }

    virtual int compare (Treap *that_);
    Edge *nextEdge () {
      Treap *nextNode = next();
      return nextNode == 0 ? 0 : dynamic_cast<SweepNode *>(nextNode)->edge;
    }
    Edge *prevEdge () {
      Treap *prevNode = prev();
      return prevNode == 0 ? 0 : dynamic_cast<SweepNode *>(prevNode)->edge;
    }
    Edge *stepEdge (int du) {
      return du == 0 ? prevEdge() : nextEdge();
    }
    void locate(Edge *edge, Edge *&prev, Edge *&next) {
      Treap *prevT, *nextT;
      Treap::locate(edge->node, prevT, nextT);
      prev = prevT == 0 ? 0 : dynamic_cast<SweepNode *>(prevT)->edge;
      next = nextT == 0 ? 0 : dynamic_cast<SweepNode *>(nextT)->edge;
    }
    SweepNode *remove () {
      return dynamic_cast<SweepNode *>(Treap::remove());
    }
    SweepNode *rebalance () {
      return dynamic_cast<SweepNode *>(Treap::rebalance());
    }
  };				

  class HeightNode : public SweepNode {
  public:
    HeightNode (Edge *edge) : SweepNode(edge) {}

    int compare (Treap *that_) {
      HeightNode *that = dynamic_cast<HeightNode *>(that_);
      return this->edge->height - that->edge->height;
    }
  };

  class RotateNode : public SweepNode {
  public:
    static double ry;
    RotateNode (Edge *edge) : SweepNode(edge) {}
    int compare (Treap *that_);
  };

  class Event {
  public:
    enum Type { REMOVE=0, TREMOVE=1, CONTINUE=2, CHECK=3, SWAP=4, INSERT=5 };
    Type type;
    Point p;
    Edge *a, *b, *c;
    
    void report ();

    Event (Type type, Point p, Edge *c, Edge *a, Edge *b)
      : type(type), p(p), a(a), b(b), c(c) {
      if (type == SWAP && a && b && a->id == 421 && b->id == 420)
	report ();
    }
    
    class Compare {
    public:
      bool operator() (const Event *a, const Event *b) const {
	if (a == b)
	  return false;
	if (a->p.x != b->p.x)
	  return a->p.x > b->p.x;
#ifdef BLEEN
	if (a->type != b->type)
	  return a->type > b->type;
#endif
	// Do REMOVEs first
	if ((a->type == REMOVE || b->type == REMOVE) && a->type != b->type)
	  return a->type > b->type;
	// Do CHECKs, SWAPs and INSERTs last
	if ((a->type >= CHECK || b->type >= CHECK) && a->type != b->type)
	  return a->type > b->type;
	// Order by y
	if (a->p.y != b->p.y)
	  return a->p.y > b->p.y;
	// Order by type (REMOVEs before CONTINUEs)
	if (a->type != b->type)
	  return a->type > b->type;
	// id == -1 (telomere) should be removed last
	if (a->c != 0 && a->c->id != b->c->id)
	  return a->c->id < b->c->id;
	if (a->c != b->c)
	  return a->c < b->c;
	if (a->a != b->a)
	  return a->a < b->a;
	if (a->b != b->b)
	  return a->b < b->b;
	assert(a->type == SWAP || a->type == CHECK);
	return false;
      }
    };

    bool equals (Event *that) {
      return (this->type == that->type && this->p == that->p &&
	      this->a == that->a && this->b == that->b && this->c == that->c);
    }
  };

  typedef std::priority_queue<Event *, vector<Event *>, Event::Compare> Queue;

  void arrange ();

  void arrange (vector<Event *> &events);
  void arrangeM (vector<Event *> &events);

  void splitEdges (vector<Event *> &events);

  void setHeights ();

  void splitBetween (Edge *edges[2], SweepNode *sweepTHS[3],
		     vector<Edge *> edgesTHV[3]);

  void pinch (Edge *edge, int du, int th,
	      SweepNode *sweepTHS[3], vector<Edge *> edgesTHV[3]);

  void embed (Edge *edge, int th, int du);

  void embed ();

  void buildLists (bool doArrange1D=true);

  void mergeEdgesV ();

  void mergeEdges ();

  class DEdge {
    Edge *e;
    bool sw;

  public:
    DEdge (Edge *edge, bool sw) : e(edge), sw(sw) {}

    DEdge next () {
      Vert *vert = e->getVert(!sw);

      if (e->isVertical()) {
	if (e->edge[sw][sw] != 0)
	  return DEdge(e->edge[sw][sw], !sw);

	if (vert->edge[!sw][!sw] != 0)
	  return DEdge(vert->edge[!sw][!sw], !sw);

	if (vert->edge[2][sw] != 0)
	  return DEdge(vert->edge[2][sw], sw);

	if (vert->edge[sw][sw] != 0)
	  return DEdge(vert->edge[sw][sw], sw);

	return DEdge(vert->edge[2][!sw], !sw);
      }
      else {
	if (vert->edge[!sw][sw] != e) {
	  assert(e->edge[!sw][sw]->getVert(!sw) == vert);
	  return DEdge(e->edge[!sw][sw], !sw);
	}
	
	if (vert->edge[2][sw] != 0)
	  return DEdge(vert->edge[2][sw], sw);
	
	if (vert->edge[sw][sw] != 0)
	  return DEdge(vert->edge[sw][sw], sw);
	
	if (vert->edge[2][!sw] != 0)
	  return DEdge(vert->edge[2][!sw], !sw);
	
	return DEdge(vert->edge[!sw][!sw], !sw);
      }
    }

    DEdge face (int th) {
      return DEdge(e->edge[th][sw], !sw);
    }

    Edge *edge () { return e; }
    bool down () { return sw; }
    bool null () { return e == 0; }
    int &loop () { return e->loop[sw]; }
    int &loop2 () { return e->loop[!sw]; }
    int mult () { return (1 - 2 * sw) * e->mult; }
    int id () { return e->id; }
    Vert *tail () { return e->getVert(sw); }
    Vert *head () { return e->getVert(!sw); }
    bool operator== (const DEdge &d) { return e == d.e && sw == d.sw; }
    bool operator!= (const DEdge &d) { return e != d.e || sw != d.sw; }
    DEdge operator++ () { *this = next(); return *this; }
    double area ();
  };

  class Loop {
  public:
    DEdge d;
    int cell;
    int chainK, chainM;
    
    Loop (DEdge d) : d(d), cell(-1), chainK(-1), chainM(-1) {}
    void edgedraw (const char *filename);
    double area ();
    int &chain (int km) { return km == 0 ? chainK : chainM; }
  };

  double area ();

  vector<Loop> loops;

  enum CellState { UNKNOWN=-1, OUTSIDE=0, INSIDE=1 };

  class Cell {
  public:
    vector<int> loops;
    int depth;
    CellState state;

    Cell () : depth(-1), state(UNKNOWN) {}
  };

  vector<Cell> cells;

  bool isUnbounded () { return cells.at(0).state == INSIDE; }

  void build (bool dontArrange1D=false);

  int upfind (int up);

  class CellTest {
  public:
    virtual bool operator() (vector<int> &cn) const = 0;
  };

  class NonZero : public CellTest {
  public:
    bool operator() (vector<int> &cn) const {
      assert(cn.size() == 1);
      return cn[0] != 0;
    }
  };

  class GTZero : public CellTest {
  public:
    bool operator() (vector<int> &cn) const {
      assert(cn.size() == 1);
      return cn[0] > 0;
    }
  };

  class Union : public CellTest {
  public:
    bool operator() (vector<int> &cn) const {
      for (int i = 0; i < cn.size(); i++)
	if (cn[i] > 0)
	  return true;
      return false;
    }
  };

  class Intersection : public CellTest {
  public:
    bool operator() (vector<int> &cn) const {
      for (int i = 0; i < cn.size(); i++)
	if (cn[i] <= 0)
	  return false;
      return true;
    }
  };

  class XOR : public CellTest {
  public:
    bool operator() (vector<int> &cn) const {
      assert(cn.size() == 2);
      return cn[0] != cn[1];
    }
  };

  int maxID () {
    int id = 0;
    for (int i = 0; i < edges.size(); i++)
      if (edges[i]->id > id)
	id = edges[i]->id;
    for (int i = 0; i < edgesV.size(); i++)
      if (edgesV[i]->id > id)
	id = edges[i]->id;
    return id;
  }

  void setCellState (vector<int> &cn, const CellTest &test, int iCell=0);

  void getBoundary ();

  Embedding (vector<Segment>  &segs, bool oddDepthIsInside=false);

  Embedding (vector<Embedding *> &in, vector<int> &cn, const CellTest &test);

  Embedding (Embedding *that, const Transformation &tr);

  Embedding () : numAdded(0), ignoreIDs(false), sumInfo(0) {
    // cout << "Embedding " << this << endl;
  }

  virtual ~Embedding ();

  void splitAtTurns (const Transformation &tr, vector<Ptr<Function> > &funcs);

  bool splitAtInconsistencies (Point u, int du);

  void arrange (const Transformation &tr, Embedding *orig,
		vector<Event *> &events);

  void draw (const char *filename);
  void draw (const char *filename, const vector<Embedding *> &regions);
  void draw (ostream &out);
  void edgedraw (const char *filename);
  void edgedraw (ostream &out);

  int getNumVerts () { return verts.size(); }
  int getNumEdges () { return edges.size() + edgesV.size(); }

  void getBBox(Point bbox[2]);

  // Determine a point inside each cell.
  // ps:  array[cells.size()], the output points
  // rs:  double[cells.size()], an estimate of point depth
  void pointInside (Point* ps, double *rs);
  Point pointInside (int iCell=-1);
  bool pointInside (Point p);

  // largest depth if inside point
  double size ();

  void complement ();

  bool isNull () { return edges.size() == 0 && cells.at(0).state == OUTSIDE; }

  CellState getOuterState () const { return cells.at(0).state; }

  bool isImplicit () {
    for (int i = 0; i < edges.size(); i++)
      if (edges[i]->f != 0)
	return edges[i]->f->getType() == Function::IMPLICIT;
    return false;
  }

  const static double BBOX;
  const static double MU;

  static bool inBBox (Point p) {
    return -BBOX < p.x && p.x < BBOX && -BBOX < p.y && p.y < BBOX;
  }

  class Chain;
  vector<Chain *> chainsK, chainsM;
  vector<Chain *> &chains (int km) { return km == 0 ? chainsK : chainsM; }
  void buildChains (bool km);

  // Return the index field of the chain incident to chains[chain]
  // at its tail / head for th == 0 / 1.
  int nextChainIndexM (int chain, int th);

  bool isHole (int iLoop) {
    return loops[iLoop].cell == 0 ||
      iLoop != cells[loops[iLoop].cell].loops[0];
  }
  Point minmaxX (int iLoop);

  void splitAtNearVerts (double epsilon);

  class PairIDSet;
  void fillMBubbles (Embedding *a, Embedding *b, PairIDSet *pairIDSet);
  void fillMSlits (Embedding *a, Embedding *b, PairIDSet *pairIDSet);
  void getMOrder (map<int, Edge *> &chains, int cct, double x,
		  vector<int> *other);

  Embedding *loop (int iLoop);

  Embedding *sumLoops
    (Embedding *that, int thisLoop, int thatLoop, bool km);
  Embedding *sumCells
    (Embedding *that, int thisCell, int thatCell, bool km);

  // 0: kinetic, 1: monotonic
  static int convolutionType;

  Embedding *sum (Embedding *that, bool km=convolutionType);
};
