#ifndef Point_h
#define Point_h
#include <math.h>
#include <iostream>
#include <vector>

using namespace std;

//#define DOUBLE long double
#define DOUBLE double

static inline DOUBLE MIN (DOUBLE x, DOUBLE y) {
  return x < y ? x : y;
}

static inline DOUBLE MAX (DOUBLE x, DOUBLE y) {
  return x > y ? x : y;
}

class Point {
public:
  DOUBLE x, y;

  Point () : x(0), y(0) {}
  Point (DOUBLE x, DOUBLE y) : x(x), y(y) {}
  Point (DOUBLE r, DOUBLE theta, bool polar)
    : x(r * cos(theta)), y(r * sin(theta)) {}
  DOUBLE &operator[] (bool xy) { return xy ? y : x; }
  bool operator== (const Point &p) const { return x == p.x && y == p.y; }
  bool operator!= (const Point &p) const { return x != p.x || y != p.y; }
  bool operator< (const Point &p) const { return x < p.x || x == p.x && y < p.y; }
  Point operator+ (const Point &p) const { return Point(x + p.x, y + p.y); }
  Point operator- () const { return Point(-x, -y); }
  Point operator- (const Point &p) const { return Point(x - p.x, y - p.y); }
  DOUBLE operator* (const Point &p) const { return x * p.x + y * p.y; }
  Point operator* (DOUBLE s) const { return Point(x * s, y * s); }
  Point operator/ (DOUBLE s) const { return Point(x / s, y / s); }
  DOUBLE operator% (const Point &p) const { return x * p.y - y * p.x; }
  DOUBLE norm () const { return *this * *this; }
  DOUBLE length () const { return sqrt(norm()); }
  DOUBLE distance (const Point &p) const { return (*this - p).length(); }
  Point unit () const { return *this / length(); }
  Point rot90 () const { return Point(-y, x); }
  DOUBLE angle () const { return atan2(y, x); }
  DOUBLE angle (Point p) const { return p(conjugate()).angle(); }

  // complex multiplication, conjugate, and inverse
  Point operator() (const Point &p) const {
    return Point(x * p.x - y * p.y, x * p.y + y * p.x);
  }
  Point conjugate () const { return Point(x, -y); }
  Point inverse () const { return conjugate() / norm(); }

  Point rint () const { return Point(::rint(x), ::rint(y)); }
};

inline ostream &operator<< (ostream &out, const Point &p) {
  return out << p.x << " " << p.y;
}

inline istream &operator>> (istream &in, Point &p) {
  return in >> p.x >> p.y;
}

inline bool xLessThan (const Point &a, const Point &b) {
  return a.x < b.x;
}

typedef vector<Point> Points;
typedef vector<Points> Curves;

class Transformation {
public:
  static const DOUBLE TINY;

  Transformation () { r.x = 1; }
  Transformation (DOUBLE q) : r(angle2point(q)) {}
  Transformation (Point t) : t(t) { r.x = 1; }
  Transformation (Point t, DOUBLE q) : t(t), r(angle2point(q)) {}
  Transformation (Point t, Point r) : t(t), r(r) {}

  Point operator() (const Point &p) const { return t + r(p); }
  Point apply (const Point &p) const { return t + r(p); }
  Transformation compose (const Transformation &that) {
    return Transformation(this->apply(that.t), this->r(that.r));
  }

  Point getT () const { return t; }
  Point getR () const { return r; }
  bool operator== (const Transformation &tr) const {
    return t == tr.t && r == tr.r;
  }
  Transformation inverse () const {
    Point rinv = r.inverse();
    Point tinv = rinv(-t);
    return Transformation(tinv, rinv);
  }

  static Point angle2point (DOUBLE q) {
    Point r(cos(q), sin(q));
    if (fabs(r.x) < TINY) r.x = 0;
    if (fabs(r.y) < TINY) r.y = 0;
    return r;
  }

private:
  Point t, r;
};

inline void writeInt (ostream &out, int x) {
  char *p = (char *) &x;
  for (int i = 0; i < sizeof(x); i++)
    out.put(p[i]);
}

inline void readInt (istream &in, int &x) {
  char *p = (char *) &x;
  for (int i = 0; i < sizeof(x); i++)
    in.get(p[i]);
}

inline void writeDouble (ostream &out, DOUBLE x) {
  char *p = (char *) &x;
  for (int i = 0; i < sizeof(x); i++)
    out.put(p[i]);
}

inline void readDouble (istream &in, DOUBLE &x) {
  char *p = (char *) &x;
  for (int i = 0; i < sizeof(x); i++)
    in.get(p[i]);
}

inline void writePoint (ostream &out, Point x) {
  char *p = (char *) &x;
  for (int i = 0; i < sizeof(x); i++)
    out.put(p[i]);
}

inline void readPoint (istream &in, Point &x) {
  char *p = (char *) &x;
  for (int i = 0; i < sizeof(x); i++)
    in.get(p[i]);
}

#endif
