#ifndef Segment_h
#define Segment_h

//#define CONSISTENT_MAXMIN

#include "Sets.h"
#include "Point.h"
#include <math.h>
#include <float.h>
#include <assert.h>
#include <iostream>
#include <vector>
#include <sys/times.h>

using std::vector;

class Function;

class RootList {
public:
  // r is roots of f - g sorted.  Makes a copy of *r.
  RootList (Function *f, Function *g, const vector<Point> &roots);

  // r is roots of f - g sorted.  Makes a copy of *r.
  RootList (Function *f, Function *g, const vector<Point> &roots, int sense);

  // This is just for Function::find.
  RootList (Function *f, Function *g);
  ~RootList ();

  int size () const { return roots.size(); }
  Point root (int i) const { return roots[i]; }

  // Make functions tangent by clearing rootlist and rebuilding based
  // on sampled midpoint value.
  void makeTangent ();

  // Returns true if two lines are made tangent.
  bool equals ();

  // compare f to g at x- (ler == -1), c (ler == 0) or at x+ (ler == 1)
  // -1: f < g, 0: f == g, 1: f > g
  int compare (DOUBLE x, int ler);

  // Return nearest root >= x or INFINITY if none.
  DOUBLE nextRoot (DOUBLE x);

  // Return nearest root to x or INFINITY if none.
  DOUBLE nearestRoot (DOUBLE x);

  // Return the actual calculated point of intersection.
  Point getRoot (DOUBLE x);

  // Flip root list between p and q.
  void flip (Point p, Point q);

  static int numRootLists;

  // f, g
  Function *func[2];
  // next root list for f, g
#ifdef LINKED_LIST_ROOTLISTS  
  RootList *next[2];
#else
#ifdef SETP_ROOTLISTS
  bool lessThan (const RootList *that) const {
    return this->func[1] < that->func[1];
  }
#else
  class Compare {
  public:
    bool operator() (const RootList *a, const RootList *b) {
      if (a->func[0] == b->func[0])
	return a->func[1] < b->func[1];
      if (a->func[0] == b->func[1])
	return a->func[1] < b->func[0];
      if (a->func[1] == b->func[0])
	return a->func[0] < b->func[1];
      if (a->func[1] == b->func[1])
	return a->func[0] < b->func[0];
      assert(0);
      return false;
    }
  };      
#endif
#endif

private:
  // sense == 0 if f ``=='' g; == 1 if f < g on the left; == -1 otherwise.
  int sense;

  // r contains only relevant roots of f - g sorted.
  vector<Point> roots;

  void build ();
};

class Segment;

class Function {
public:
  static const DOUBLE FINFINITY;
  static const int IMPLICIT;
  static const int LINE;
  static const int UPPER_CIRCLE;
  static const int LOWER_CIRCLE;
  static const int TELOMERE_CURVE;

  Function ();
  virtual ~Function ();

  void incref ();
  void decref ();
  int getRef () { return refcount; }

  virtual void free () { delete this; }
  virtual int getType() const = 0;
  virtual DOUBLE getMinX () const = 0;
  virtual DOUBLE getMaxX () const = 0;
  virtual DOUBLE getY (DOUBLE x) const = 0;
  virtual DOUBLE getDY (DOUBLE x) const = 0;
  virtual RootList *makeRootList (Function *that) = 0;
  virtual void transform (const Transformation &tr, vector<Segment> &out,
			  bool ymonotone = false) = 0;
  virtual void transform (Segment &seg, const Transformation &tr,
			  vector<Segment> &out, bool ymonotone = false) = 0;
  virtual DOUBLE distance (Point p) const;
  virtual int bboxOrder (Function *g) { return 0; }
  virtual int slopeOrder (Function *g, const Point &p) { return 0; }
  virtual Point nearestPoint (Point p) const { return Point(p.x, getY(p.x)); }

  RootList *find (Function *that);

  void clearRootLists ();

  void setRootList(Function *that, RootList *rl) {}

  void getRootLists (vector<RootList *> &res);

  // Force the two functions to be ``tangent'', not intersecting,
  // by clearing the rootlist and rebuilding based on a sampled x,y.
  void makeTangent (Function *that);

  // Compare this to that at x (ler==0), x- (ler==-1), or x+ (ler==1).
  // Return -1 if this below that, 0 if equal, 1 if this above that.
  int compare (Function *that, DOUBLE x, int ler);

  // cheap_compare samples the functions if the root list is not available.
  int cheap_compare (Function *that, DOUBLE x, int ler);

  // Return the nearest root > x or FINFINITY.
  DOUBLE nextRoot (Function *that, DOUBLE x);

  // Return the nearest root to x or FINFINITY.
  DOUBLE nearestRoot (Function *that, DOUBLE x);

  // Return the actual calculated point of intersection.
  Point getRoot (Function *that, DOUBLE x);

  // Flip the root list interval between p and q.
  void flip (Function *that, Point p, Point q) {
    getRootList(that)->flip(p, q);
  }

  bool overlaps (Function *that) const {
    return (this->getMinX() < that->getMaxX() &&
	    that->getMinX() < this->getMaxX());
  }

  void check (Point p);

  RootList *getRootList (Function *that);


#ifdef LINKED_LIST_ROOTLISTS  
  RootList *head[2];
#else
#ifdef SETP_ROOTLISTS
  SetP<RootList> lists;
#else
  SetPC<RootList, RootList::Compare> lists;
#endif
#endif

private:
  int refcount;
};

class Segment {
public:
  Segment (const Point &tail, const Point &head, Function *f);
  Segment () : f(0) {}
  Segment (const Segment &that) {
    tail = that.tail;
    head = that.head;
    f = that.f;
    if (f != 0)
      f->incref();
  }
  void operator= (const Segment &that) {
    tail = that.tail;
    head = that.head;
    if (that.f != 0)
      that.f->incref();
    if (f != 0)
      f->decref();
    f = that.f;
  }
  ~Segment () { if (f != 0) f->decref(); }

  Point getTail () const { return tail; }
  Point getHead () const { return head; }
  Function *getF () const { return f; }

  Segment reverse () { return Segment(head, tail, f); }
  DOUBLE getMinX () { return tail.x < head.x ? tail.x : head.x; }
  DOUBLE getMaxX () { return tail.x > head.x ? tail.x : head.x; }
  Point getMin () const {
    if (tail.x < head.x || (tail.x == head.x && tail.y < head.y))
      return tail;
    else
      return head;
  }
  Point getMax () const {
    if (tail.x < head.x || (tail.x == head.x && tail.y < head.y))
      return head;
    else
      return tail;
  }
  Point nearestPoint (Point p) const {
    Point q = f->nearestPoint(p);
    if (q.x <= tail.x)
      return tail;
    if (q.x >= head.x)
      return head;
    return q;
  }

  void transform (const Transformation &tr, vector<Segment> &out);

private:
  Point tail, head;
  Function *f;
};

class IP {
public:
  IP (Point p) : x(int(rint(p.x))), y(int(rint(p.y))) {}
  int x, y;
};

inline ostream &operator<< (ostream &out, const IP &p) {
  return out << p.x << " " << p.y;
}

void approximate (Function *f, Point tail, Point head, vector<Point> &points);
void segdump (const char *filename, const vector<Segment> &segs);

namespace Timing {
  extern int arrange, rootlist, dump, transform, find, y, compare;
  extern int numRootList, numLink, numExtraRoots, newroot, N;
  extern int numInsertByRoot, degree;
  extern int rootCount[10];
  extern int numFG, timeFG, numEvil, timeXY, numXY;
  extern int rootHist[1024];
  extern int pointInside;
  extern int numrealroots, timerealroots;
  extern int fit, sum, sumloops, sumarrange, adapt;
  extern int ncells, nbubbles;
  extern int embeddedEndpoints, embeddingViolations, splitAtInconsistencies;

  inline clock_t utime () {
    struct tms buf;
    clock_t t = times(&buf);
    return buf.tms_utime;
  }    

  inline void begin (int &counter) {
    counter -= utime();
  }

  inline void end (int &counter) {
    counter += utime();
  }
}    

#endif
