#ifndef Sets_h
#define Sets_h

#include <assert.h>
#include <ext/hash_map>
#include <set>

using std::set;
using __gnu_cxx::hash_map;

const size_t RND = 3608578767u;

inline size_t hashpn (const void *p, int n) {
  assert(n % 4 == 0);
  n /= 4;
  const size_t *s = (const size_t *) p;
  size_t sum = 0;
  for (int i = 0; i < n; i++)
    sum = RND * sum + s[i];
  return sum;
}

template<class T>
struct Hash {
  size_t operator() (const T &a) const { return hashpn(&a, sizeof(T)); }
};

template<class K, class T>
class HashMap : public hash_map<K, T, Hash<K> > {};

template<class T>
class LessThanP {
public:
  bool operator() (const T *a, const T *b) const {
    return a->lessThan(b);
  }
};

template<class T>
class SetP {
public:
  int size () { return setP.size(); }

  T *insert (T *t) {
    current = setP.insert(t).first;
    return *current;
  }

  T *find (T *t) {
    current = setP.find(t);
    return current == setP.end() ? 0 : *current;
  }

  void remove () {
    setP.erase(current);
  }

  T *first () {
    current = setP.begin();
    return curr();
  }

  void operator++ (int i) { current++; }

  void operator-- (int i) { current--; }

  T *curr () { return current == setP.end() ? 0 : *current; }

  T *next () {
    typename set<T *, LessThanP<T> >::iterator it = current;
    it++;
    return it == setP.end() ? 0 : *it;
  }

  T *prev () {
    if (current == setP.begin())
      return 0;
    typename set<T *, LessThanP<T> >::iterator it = current;
    it--;
    return *it;
  }

  void clear () { setP.clear(); }

  set<T *, LessThanP<T> > setP;
  typename set<T *, LessThanP<T> >::iterator current;
};

template<class T, class C>
class SetPC {
public:
  typedef set<T *, C> Set;
  typedef typename Set::iterator Iterator;
  Set setP;
  Iterator current;

  int size () { return setP.size(); }

  bool empty () { return setP.empty(); }

  T *insert (T *t) {
    current = setP.insert(t).first;
    return *current;
  }

  T *insert (T *t, bool check) {
    std::pair<Iterator, bool> p = setP.insert(t);
    assert(p.second);
    current = p.first;
    return t;
  }

  T *find (T *t) {
    current = setP.find(t);
    return current == setP.end() ? 0 : *current;
  }

  T *find (T *t, bool check) {
    current = setP.find(t);
    assert(current != setP.end());
    assert(*current == t);
    return t;
  }

  T *upper_bound (T *t) {
    current = setP.upper_bound(t);
    return current == setP.end() ? 0 : *current;
  }

  T *lower_bound (T *t) {
    current = setP.lower_bound(t);
    return current == setP.end() ? 0 : *current;
  }

  void remove () {
    assert(current != setP.end());
    Iterator previous = current;
    ++current;
    setP.erase(previous);
  }

  void remove (T *t) {
    find(t);
    if (curr() == t)
      remove();
  }

  void remove (T *t, bool check) {
    find(t, true);
    remove();
  }

  T *first () {
    current = setP.begin();
    return curr();
  }

  T *last () {
    current = setP.end();
    --current;
    return curr();
  }

  T *operator++ () {
    if (current == setP.end())
      current = setP.begin();
    else
      ++current;
    return curr();
  }

  T *operator-- () {
    if (current == setP.begin())
      current = setP.end();
    else
      --current;
    return curr();
  }

  T *operator++ (int i) {
    T *temp = curr();
    if (current == setP.end())
      current = setP.begin();
    else
      ++current;
    return temp;
  }

  T *operator-- (int i) {
    T *temp = curr();
    if (current == setP.begin())
      current = setP.end();
    else
      --current;
    return temp;
  }

  T *curr () { return current == setP.end() ? 0 : *current; }

  T *next () {
    Iterator it = current;
    it++;
    return it == setP.end() ? 0 : *it;
  }

  T *prev () {
    if (current == setP.begin())
      return 0;
    Iterator it = current;
    it--;
    return *it;
  }
  
  void clear () { setP.clear(); }

};


class RefCount {
  int refcount;
public:
  RefCount () { refcount = 0; }
  virtual ~RefCount () {}

  void incref () { refcount++; }
  void decref () {
    if (--refcount == 0)
      delete this;
  }
};

template<class C>
class Ptr {
private:
  C *ptr;
public:
  Ptr () : ptr(0) {}
  Ptr (C *ptr) : ptr(ptr) { if (ptr != 0) ptr->incref(); }
  Ptr (const Ptr<C> &that) : ptr(that.ptr) { if (ptr != 0) ptr->incref(); }
  ~Ptr () { if (ptr != 0) ptr->decref(); }
  void operator= (const Ptr<C> &that) {
    if (that.ptr != 0) that.ptr->incref();
    if (ptr != 0) ptr->decref();
    ptr = that.ptr;
  }
  operator C* () const { return ptr; }
  C* operator-> () { return ptr; }
  const C* operator-> () const { return ptr; }
};

#endif
