char copyright3[] = "copyright (c) 2008 Victor J. Milenkovic, not for commercial use without permission";

#include "Embedding.h"
#include "Circle.h"
#include <map>
using namespace std;

//#define BUBBLE_TEST

class Embedding::Chain {
public:
  // dprev is last non-vertical segment in previous chain.
  // dnext is first (non-vertical) segment in this chain.
  Chain (Embedding::DEdge &dprev, Embedding::DEdge &dnext, int index, bool km);
  
  // Helper for constructor: add a segment with head h, radius r,
  // and center c.
  void addSegment (Point h, double r, Point c);
  
  char km;	// kinetic or monotonic
  char sw;    // All non-vertical DEdges have sw
  char endIsConcave[2 /* tail/head */];
  int index, loop;
  
  // kinetic:
  // For upper chain, angles.front() is 0/PI for convex/concave
  // turning point, and vice versa for angles.back().
  // For lower chain, angles.front() is -PI/0 for convex/concave
  // turning point, and vice versa for angles.back().
  // monotonic:
  // Front and back always 0.  Segments split at PI/2 or -PI/2
  // Turn from -PI/2 to PI/2 or vice versa is null.
  // For upper chain, PI/2 to PI represented by -PI/2 to 0,
  // and horizontal segment has PI/2.
  // For lower chain, -PI to -PI/2 represented by 0 to PI/2,
  // and horizontal segment has -PI/2.
  vector<double> angles;
  double getA (int i) { return angles[i]; }
  static bool compare (double *a, double *b) { return *a < *b; }
  static bool compareM (double a, double b) {
    if ((a >= 0) == (b >= 0))
      return a < b;
    if (a >= 0)
      return a != M_PI;
    return b == M_PI;
  }
  static bool compareMP (double *a, double *b) { return compareM(*a, *b); }
  vector<double *> sortedAngles;
  
  static double minM (double a, double b) { return compareM(a, b) ? a : b; }
  static double maxM (double a, double b) { return compareM(a, b) ? b : a; }
  static double sinM (double a) { return a >= 0 ? sin(a) : -sin(a); }
  static double cosM (double a) { return a >= 0 ? cos(a) : -cos(a); }

  // angles[2*i] and angles[2*i+1] are the alpha and beta
  // for a pure turn at points[i], i=0,...,n
  vector<Point> points;
  Point getP (int i) { return points[i/2]; }
  
  // angles[2*i+1] and angles[2*i+2] are alpha and beta
  // for an arc with center centers[i] and radius radii[i], i=0,...,n-1
  // radius=infinity for line segment
  vector<Point> centers;
  Point getC (int i) { return i % 2 == 0 ? getP(i) : centers[i/2]; }
  
  vector<double> radii;
  double getR (int i) { return i % 2 == 0 ? 0 : radii[i/2]; }
  
  bool isUpper () { return sw; }
  Point tail () { return points.front(); }
  Point head () { return points.back(); }
  Point end (int th) { return th ? points.back() : points.front(); }
  double minX () { return isUpper() ? head().x : tail().x; }
  double maxX () { return isUpper() ? tail().x : head().x; }
  
  // Calculate the angle of a normal vector.
  // pm == 0 / 1 means it should be >= 0 / <= 0
  double calcAngle (Point v, int pm=-1);
  double calcAngleM (Point v);
  bool lt (int thisI, Chain *that, int thatI);
  
  void sumK (Chain *that, Embedding *convK);
  void sumM (Chain *that, Embedding *convM, int thatChainSize  );
  void sumM (Point ctp, int pairID, Embedding *convM);
  
  void overlapS (Chain *that, vector<int> &pairs);
  void overlap (Chain *that, vector<int> &pairs);
  Point getP (int i, double angle, bool swap=false);
  Point getP (int i, double angle, double &xError);
  static Point getP (Point a, double alpha, Point b, double beta,
		     double theta, double &xError);
  Ptr<Function> sumF (Chain *that, int thisI, int thatI,
		      Point tail, Point head);
  
  // Accurate floating point sum of points plus one x-roundoff.
  // If axError==0/bxError==0, then bxError/axError is the roundoff
  // error in the calculation of b.x/a.x.
  // x-coordinate is consistent rounding of a.x + b.x + error.
  // (It's rounded down for convenient.)
  static Point sum (Point a, double axError, Point b, double bxError);
  
  // Exact floating point sum.
  // Returns sum of a and b with rounding error in error.
  static double sum (double a, double b, double &error) {
    double s = a + b;
    if (fabs(a) > fabs(b)) {
      double bb = s - a;
      // s == a + bb
      error = b - bb;
    }
    else {
      double aa = s - b;
      error = a - aa;
    }
    return s;
  }

  static bool isConvex (Embedding::DEdge dprev, Embedding::DEdge dnext);
};

class Embedding::PairIDSet /* : public IntOp */ {
  // For v = *i2set[i], v[0] = -i-1, and v[1...] is the set in sorted order.
  // *i2set[0] is the empty set {} = <-1>
  vector<vector<int>*> i2set;
  
  class Compare {
  public:
    bool operator() (vector<int> *a, vector<int> *b) {
      for (int i = 1; i < b->size(); i++)
	if (i == a->size())
	  return true;
	else if ((*a)[i] != (*b)[i])
	  return (*a)[i] < (*b)[i];
      return false;
    }
  };
  
  set<vector<int> *, Compare> sets;
  
public:
  PairIDSet () {
    // add null set
    i2set.push_back(new vector<int>(1));
    (*i2set[0])[0] = -1;
    sets.insert(i2set[0]);
  }
  
  ~PairIDSet () {
    for (int i = 0; i < i2set.size(); i++)
      delete i2set[i];
  }
  
  // Combine two sets of pair ids.
  // a >= 0 represents the set {a}
  // a < 0 represents *i2set[-a-1]
  int op (int a, int b) {
    // idempotent
    if (a == b)
      return -1;
    if (a == -1)
      return b;
    if (b == -1)
      return a;
    
    vector<int> v;
    v.push_back(-i2set.size()-1);
    if (a >= 0)
      v.push_back(a);
    else {
      vector<int> &va = *i2set.at(-a-1);
      for (int i = 1; i < va.size(); i++)
	v.push_back(va[i]);
    }
    if (b >= 0)
      v.push_back(b);
    else {
      vector<int> &vb = *i2set.at(-b-1);
      for (int i = 1; i < vb.size(); i++)
	v.push_back(vb[i]);
    }
    sort(v.begin(), v.end());
    int n = 1;
    for (int i = 1; i < v.size(); i++)
      if (v[i] == v[n-1])
	n--;
      else
	v[n++] = v[i];
    v.resize(n);
    
    if (v.size() == 1)
      return -1;
    if (v.size() == 2)
      return v[1];
    
    set<vector<int> *, Compare>::iterator it = sets.find(&v);
    if (it != sets.end())
      return (**it)[0];
    
    i2set.push_back(new vector<int>(v));
    sets.insert(i2set.back());
    return v[0];
  }
  
  void getSet(int a, vector<int> &v) {
    v.clear();
    if (a >= 0)
      v.push_back(a);
    else {
      vector<int> &av = *i2set[-a-1];
      for (int i = 1; i < av.size(); i++)
	v.push_back(av[i]);
    }
  }

  virtual bool isNull (int a) {
    if (a == -1)
      return true;
    if (a >= 0)
      return false;
    vector<int> &av = *i2set[-a-1];
    for (int i = 2; i < av.size(); i += 2)
      if (av[i-1] / 5 != av[i] / 5)
	return false;
    return true;
  }
};

Embedding::~Embedding () {
  // cout << "~Embedding " << this << endl;
  
  for (Verti it = verts.begin(); it != verts.end(); it++)
    delete *it;

  for (int i = 0; i < edges.size(); i++)
    delete edges[i];

  for (int i = 0; i < edgesV.size(); i++)
    delete edgesV[i];

  for (int i = 0; i < chainsK.size(); i++)
    delete chainsK[i];
  for (int i = 0; i < chainsM.size(); i++)
    delete chainsM[i];
}

bool Embedding::Chain::isConvex (Embedding::DEdge dprev, Embedding::DEdge dnext) {
  assert(dprev.down() != dnext.down());
  if (dprev.edge()->isVertical())
    return true;
  if (dnext.edge()->isVertical())
    return false;
  return dnext.down() == (dprev.edge()->height < dnext.edge()->height);
}   

Embedding::Chain::Chain (Embedding::DEdge &dprev, Embedding::DEdge &dnext,
			 int index, bool km)
  : loop(-1), index(index), sw(dnext.down()), km(km) {
  assert(dprev.down() != dnext.down());
  endIsConcave[0] = !isConvex(dprev, dnext);

  if (km == 0) {
    if (isConvex(dprev, dnext))
      angles.push_back(isUpper() ? 0 : -M_PI);
    else
      angles.push_back(isUpper() ? M_PI : 0);
    
    do {
      points.push_back(dnext.tail()->p);
      Edge *edge = dnext.edge();
      int type = edge->f == 0 ? Function::LINE : edge->f->getType();
      if (type == Function::LINE) {
	double angle = calcAngle((dnext.tail()->p - dnext.head()->p).rot90(),
				 !isUpper());
	angles.push_back(angle);
	angles.push_back(angle);
	centers.push_back(Point(-Function::FINFINITY, -Function::FINFINITY));
	radii.push_back(Function::FINFINITY);
      }
      else if (type == Function::LOWER_CIRCLE ||
	       type == Function::UPPER_CIRCLE) {
	int sgn = (type == Function::LOWER_CIRCLE) == isUpper() ? -1 : 1;
	Circle *circle = (dynamic_cast<Cbranch *>((Function *) edge->f))->circle;
	Point c = circle->c;
	centers.push_back(c);
	radii.push_back(circle->r * sgn);
	double alpha = calcAngle((dnext.tail()->p - c) * sgn, !isUpper());
	double beta = calcAngle((dnext.head()->p - c) * sgn, !isUpper());
	angles.push_back(alpha);
	angles.push_back(beta);
	
#ifdef BLEEN
	// assert(sgn *  (beta - alpha) > 0);
	if (alpha == beta) {
	  cout << "arc is straight!!" << endl;
	  centers.back() = Point(-Function::FINFINITY, -Function::FINFINITY);
	  radii.back() = Function::FINFINITY;
	}
#endif

	if (sgn * (beta - alpha) <= 0) {
	  double angle = calcAngle((dnext.tail()->p - dnext.head()->p).rot90(),
				   !isUpper());
	  angles[angles.size()-2] = angle;
	  angles[angles.size()-1] = angle;
	  centers.back() = Point(-Function::FINFINITY, -Function::FINFINITY);
	  radii.back() = Function::FINFINITY;
	}
      }
      
      dprev = dnext;
      ++dnext;
    } while (dprev.down() == dnext.down());
    
    if (isConvex(dprev, dnext))
      angles.push_back(isUpper() ? M_PI : 0);
    else
      angles.push_back(isUpper() ? 0 : -M_PI);
    points.push_back(dnext.tail()->p);
  }
  else {
    angles.push_back(0);
    points.push_back(dnext.tail()->p);

    do {
      Point head = dnext.head()->p;
      Edge *edge = dnext.edge();
      edge->info = index;
      if (edge->f == 0 || edge->f->getType() == Function::LINE)
	addSegment(head, Function::FINFINITY,
		   Point(-Function::FINFINITY, -Function::FINFINITY));
      else {
	bool isLowerCircle = edge->f->getType() == Function::LOWER_CIRCLE;
	Circle *circle =
	  (dynamic_cast<Cbranch *>((Function *) edge->f))->circle;
	double r = isLowerCircle == isUpper() ? -circle->r : circle->r;
	addSegment(head, r, circle->c);
      }
      
      if (edge->f != 0)
	dprev = dnext;
      ++dnext;
    } while (dnext.edge()->f == 0 || dprev.down() == dnext.down());
    
    angles.push_back(M_PI);
  }
  endIsConcave[1] = !isConvex(dprev, dnext);
    
  sortedAngles.resize(angles.size());
  for (int i = 0; i < angles.size(); i++)
    sortedAngles[i] = &angles[i];
  if (km == 0)
    sort(sortedAngles.begin(), sortedAngles.end(), compare);
  else
    sort(sortedAngles.begin(), sortedAngles.end(), compareMP);

#ifdef BLEEN
  cout << (endIsConcave[0] ? "concave" : "convex") << endl;
  for (int i = 0; i < angles.size() - 1; i++)
    cout << "angle " << getA(i) << " "
	 << "center " << getC(i) << " "
	 << "radius " << getR(i) << endl;
  cout << "angle " << angles.back() << endl;
  cout << (endIsConcave[1] ? "concave" : "convex") << endl << endl;
#endif
}

#ifdef BLEEN
void Embedding::Chain::addSegment (Point head, double r, Point c) {
  Point tail = points.back();
  double theta = isUpper() ?
    calcAngleM((tail - head).rot90()) :
    calcAngleM((head - tail).rot90());
  int durSgn = isUpper() == (r > 0) ? 1 : -1;

  double alpha, beta, at, smaller;
  if (r == Function::FINFINITY) {
    alpha = theta;
    beta = theta;
    at = 0;
    smaller = alpha;
  }
  else if (theta >= M_PI / 4 || theta <= -M_PI / 4) {
    alpha = calcAngleM((tail - c) * durSgn);
    beta = calcAngleM((head - c) * durSgn);
    At = (theta + (theta < 0) * M_PI) - (alpha + (alpha < 0) * M_PI);
    smaller = alpha;
  }
  else if (isUpper() ^ (r > 0) ^ (tail.y < head.y)) {
    alpha = calcAngleM((tail - c) * durSgn);
    beta = alpha == M_PI ? 2 * theta : 2 * theta - alpha;
    at = beta - theta;
    smaller = alpha;
  }
  else {
    beta = calcAngleM((head - c) * durSgn);
    alpha = beta == M_PI ? 2 * theta : 2 * theta - beta;
    at = theta - alpha;
    smaller = beta;
  }

  angles.push_back(alpha);
  if (at == 0) {
    radii.push_back(r);
    centers.push_back(c);
  }
  else {
    double rNew = tail.distance(head) / 2 / sin(at);
    int udSgn = 2 * isUpper() - 1;
    Point cNew = smaller == alpha ?
      tail - Point(cosM(alpha), sinM(alpha)) * (rNew * udSgn):
      head - Point(cosM(beta), sinM(beta)) * (rNew * udSgn);

    static double rNewMax;
    double rDiff = fabs(rNew - r);
    if (rNewMax < rDiff) {
      rNewMax = rDiff;
      cout << "rNewMax " << rNewMax << endl;
    }

    static double cNewMax;
    double cDiff = c.distance(cNew);
    if (cNewMax < cDiff) {
      cNewMax = cDiff;
      cout << "cNewMax " << cNewMax << endl;
    }

    radii.push_back(rNew);
    centers.push_back(cNew);
  }
  angles.push_back(beta);
  points.push_back(head);
}
#endif

void Embedding::Chain::addSegment (Point head, double r, Point c) {
  Point tail = points.back();
  Point perp = (tail - head).rot90();
  double theta = isUpper() ? calcAngleM(perp) : calcAngleM(-perp);
  
  if (r == Function::FINFINITY) {
    angles.push_back(theta);
    radii.push_back(r);
    centers.push_back(c);
    angles.push_back(theta);
    points.push_back(head);
    return;
  }
  
  double sinPhi = perp.length() / 2 / fabs(r);
  assert(sinPhi <= 1);
  double phi = asin(sinPhi);
  double alpha = theta - phi;
  double beta = theta + phi;
  
  if (alpha <= -M_PI/2) {
    alpha += M_PI;
    if (alpha < 0)
      alpha = 0;
  }
  else if (theta > 0 && alpha < 0)
    alpha = 0;
  
  if (beta > M_PI/2) {
    beta -= M_PI;
    if (beta >= 0)
      beta = M_PI;
  }
  else if (theta < 0 && beta >= 0)
    beta = M_PI;
  
  if (r < 0) {
    double temp = alpha;
    alpha = beta;
    beta = temp;
  }
  
  int durSgn = isUpper() == (r > 0) ? 1 : -1;

  double alpha2 = calcAngleM((tail - c) * durSgn);
  static double alphaMax;
  double diff = alpha - alpha2;
  if (alpha == M_PI && alpha2 < 0)
    diff = -alpha2;
  else if (alpha2 == M_PI && alpha < 0)
    diff = -alpha;
  if (diff > alphaMax) {
    alphaMax = diff;
    cout << "alphaMax " << alphaMax << endl;
    if (alphaMax > 1e-3)
      cout << "this is it" << endl;
  }

  double beta2 = calcAngleM((head - c) * durSgn);
  static double betaMax;
  diff = beta - beta2;
  if (beta == M_PI && beta2 < 0)
    diff = -beta2;
  else if (beta2 == M_PI && beta < 0)
    diff = -beta;
  if (diff > betaMax) {
    betaMax = diff;
    cout << "betaMax " << betaMax << endl;
    if (betaMax > 1e-3)
      cout << "this is it" << endl;
  }

  // b = sqrt(r^2 - a^2) = r * sqrt(1 - (a/r)^2) * perp / a
  double cosPhi = sqrt(1 / (sinPhi * sinPhi) - 1) / 2;
  if (r > 0) cosPhi = -cosPhi;
  Point cNew = (tail + head) / 2 + perp * cosPhi;
  static double cNewMax;
  double cDiff = c.distance(cNew);
  if (cNewMax < cDiff) {
    cNewMax = cDiff;
    cout << "cNewMax " << cNewMax << endl;
  }

  angles.push_back(alpha);
  radii.push_back(r);
  centers.push_back(cNew);
  angles.push_back(beta);
  points.push_back(head);
}

double Embedding::Chain::calcAngleM (Point v) {
  if (v.y < 0)
    v.y = 0;
  if (v.y == 0)
    return v.x > 0 ? 0 : M_PI;
  if (v.x >= 0)
    return atan2(v.y, v.x);
  else {
    double angle = atan2(-v.y, -v.x);
    assert(angle < 0);
    return angle;
  }
}

double Embedding::Chain::calcAngle (Point v, int pm) {
  if (pm == 0 && v.y < 0 || pm == 1 && v.y > 0)
    v.y = 0;
  assert(v.x != 0 || v.y != 0);
  if (v.y == 0)
    if (v.x > 0)
      return 0;
    else if (pm == 0)
      return M_PI;
    else
      return -M_PI;
  else
    return v.angle();
}

void Embedding::buildChains (bool km) {
  for (int i = 0; i < loops.size(); i++) {
    Loop &loop = loops[i];
    if (cells[loop.cell].state == OUTSIDE) {
      loop.chain(km) = -1;
      continue;
    }
    assert(cells[loop.cell].state == INSIDE);

    Embedding::DEdge dprev = loop.d;
    while (dprev.edge()->f == 0)
      ++dprev;
    Embedding::DEdge dnext = dprev.next();
    if (km == 0) {
      while (dprev.down() == dnext.down()) {
	dprev = dnext;
	++dnext;
      }
    }
    else {
      while (dnext.edge()->f == 0 || dprev.down() == dnext.down()) {
	if (dnext.edge()->f != 0)
	  dprev = dnext;
	++dnext;
      }
    }

    static int count;
    count++;
    // if (count == 17)
    // cout << "this is it" << endl;

    vector<Chain *> &chains = km == 0 ? chainsK : chainsM;
    loop.chain(km) = chains.size();
    Embedding::DEdge dstart = dnext;
    do {
      Chain *chain = new Chain(dprev, dnext, chains.size(), km);
      assert(chain->index == chains.size());
      chain->loop = i;
      chains.push_back(chain);
    } while (dnext != dstart);
  }
}

//#define SLOW_OVERLAP
//#ifdef SLOW_OVERLAP
void Embedding::Chain::overlapS (Chain *that, vector<int> &pairs) {
  for (int thisI = 1; thisI < this->angles.size(); thisI++) {
    double thisAlpha = this->getA(thisI-1);
    double thisBeta = this->getA(thisI);
    double thisMin =
      km ? minM(thisAlpha, thisBeta) : min(thisAlpha, thisBeta);
    double thisMax =
      km ? maxM(thisAlpha, thisBeta) : max(thisAlpha, thisBeta);

    for (int thatI = 1; thatI < that->angles.size(); thatI++) {
      double thatAlpha = that->getA(thatI-1);
      double thatBeta = that->getA(thatI);
      double thatMin =
	km ? minM(thatAlpha, thatBeta) : min(thatAlpha, thatBeta);
      double thatMax =
	km ? maxM(thatAlpha, thatBeta) : max(thatAlpha, thatBeta);
  
      if (!km && thisMax >= thatMin && thatMax >= thisMin ||
	  km && !compareM(thisMax, thatMin) && !compareM(thatMax, thisMin)) {
	pairs.push_back(thisI - 1);
	pairs.push_back(thatI - 1);
      }
    }
  }
}
//#else
void Embedding::Chain::overlap (Chain *that, vector<int> &pairs) {
  vector<double *>::iterator it;
#ifdef BLEEN
  vector<int> pairsS;
  overlapS(that, pairsS);
#endif

  if (km == 0) {
    for (int thisI = 1; thisI < this->angles.size(); thisI++) {
      double thisAlpha = this->getA(thisI-1);
      double thisBeta = this->getA(thisI);
      double thisMin = min(thisAlpha, thisBeta);
      double thisMax = max(thisAlpha, thisBeta);

      for (it = lower_bound(that->sortedAngles.begin(),
			    that->sortedAngles.end(), &thisMin, compare);
	   it != that->sortedAngles.end() && **it <= thisMax; ++it) {
	int endI = *it - &that->angles[0];

	for (int thatI = endI;  thatI <= endI + 1; thatI++) {
	  if (thatI == 0 || thatI == that->angles.size())
	    continue;
	  double thatAlpha = that->getA(thatI-1);
	  double thatBeta = that->getA(thatI);
	  double thatMin = min(thatAlpha, thatBeta);
	  double thatMax = max(thatAlpha, thatBeta);
	
	  // If both thatAlpha and thatBeta are in [thisMin,thisMax],
	  // then only consider the pair when thatI == endI.
	  if (thatI == endI + 1 &&
	      thisMin <= thatBeta && thatBeta <= thisMax)
	    continue;

	  pairs.push_back(thisI - 1);
	  pairs.push_back(thatI - 1);

#ifdef BLEEN
	  bool found = false;
	  for (int i = 0; !found && i < pairsS.size(); i += 2)
	    found = pairsS[i] == thisI - 1 && pairsS[i+1] == thatI - 1;
	  assert(found);
#endif
	}
      }
    }

    for (int thatI = 1; thatI < that->angles.size(); thatI++) {
      double thatAlpha = that->getA(thatI-1);
      double thatBeta = that->getA(thatI);
      double thatMin = min(thatAlpha, thatBeta);
      double thatMax = max(thatAlpha, thatBeta);

      for (it = upper_bound(this->sortedAngles.begin(),
			    this->sortedAngles.end(), &thatMin, compare);
	   it != this->sortedAngles.end() && **it < thatMax; ++it) {
	int endI = *it - &this->angles[0];

	for (int thisI = endI;  thisI <= endI + 1; thisI++) {
	  if (thisI == 0 || thisI == this->angles.size())
	    continue;
	  double thisAlpha = this->getA(thisI-1);
	  double thisBeta = this->getA(thisI);
	  double thisMin = min(thisAlpha, thisBeta);
	  double thisMax = max(thisAlpha, thisBeta);
	
	  // This pair has been found already.
	  if (thisMin <= thatAlpha && thatAlpha <= thisMax ||
	      thisMin <= thatBeta && thatBeta <= thisMax)
	    continue;

	  // If both thisAlpha and thisBeta are in (thatMin,thatMax),
	  // then only consider the pair when thisI == endI.
	  if (thisI == endI + 1 &&
	      thatMin < thisBeta && thisBeta < thatMax)
	    continue;

	  pairs.push_back(thisI - 1);
	  pairs.push_back(thatI - 1);

#ifdef BLEEN
	  bool found = false;
	  for (int i = 0; !found && i < pairsS.size(); i += 2)
	    found = pairsS[i] == thisI - 1 && pairsS[i+1] == thatI - 1;
	  assert(found);
#endif
	}
      }
    }

#ifdef BLEEN
    assert(pairs.size() == pairsS.size());
#endif
  }
  else if (km == 1) {
    for (int thisI = 1; thisI < this->angles.size(); thisI++) {
      double thisAlpha = this->getA(thisI-1);
      double thisBeta = this->getA(thisI);
      double thisMin = minM(thisAlpha, thisBeta);
      double thisMax = maxM(thisAlpha, thisBeta);

      for (it = lower_bound(that->sortedAngles.begin(),
			    that->sortedAngles.end(), &thisMin, compareMP);
	   it != that->sortedAngles.end() && !compareM(thisMax, **it); ++it) {
	int endI = *it - &that->angles[0];

	for (int thatI = endI;  thatI <= endI + 1; thatI++) {
	  if (thatI == 0 || thatI == that->angles.size())
	    continue;
	  double thatAlpha = that->getA(thatI-1);
	  double thatBeta = that->getA(thatI);
	  double thatMin = minM(thatAlpha, thatBeta);
	  double thatMax = maxM(thatAlpha, thatBeta);
	
	  // If both thatAlpha and thatBeta are in [thisMin,thisMax],
	  // then only consider the pair when thatI == endI.
	  if (thatI == endI + 1 &&
	      !compareM(thatBeta, thisMin) &&
	      !compareM(thisMax, thatBeta))
	    // thisMin <= thatBeta && thatBeta <= thisMax)
	    continue;

	  pairs.push_back(thisI - 1);
	  pairs.push_back(thatI - 1);

#ifdef BLEEN
	  bool found = false;
	  for (int i = 0; !found && i < pairsS.size(); i += 2)
	    found = pairsS[i] == thisI - 1 && pairsS[i+1] == thatI - 1;
	  assert(found);
#endif
	}
      }
    }

    for (int thatI = 1; thatI < that->angles.size(); thatI++) {
      double thatAlpha = that->getA(thatI-1);
      double thatBeta = that->getA(thatI);
      double thatMin = minM(thatAlpha, thatBeta);
      double thatMax = maxM(thatAlpha, thatBeta);

      for (it = upper_bound(this->sortedAngles.begin(),
			    this->sortedAngles.end(), &thatMin, compareMP);
	   it != this->sortedAngles.end() && !compareM(thatMax, **it); ++it) {
	int endI = *it - &this->angles[0];

	for (int thisI = endI;  thisI <= endI + 1; thisI++) {
	  if (thisI == 0 || thisI == this->angles.size())
	    continue;
	  double thisAlpha = this->getA(thisI-1);
	  double thisBeta = this->getA(thisI);
	  double thisMin = minM(thisAlpha, thisBeta);
	  double thisMax = maxM(thisAlpha, thisBeta);
	
	  // This pair has been found already.
	  if (!compareM(thatAlpha, thisMin) &&
	      !compareM(thisMax, thatAlpha) ||
	      !compareM(thatBeta, thisMin) &&
	      !compareM(thisMax, thatBeta))
	    // if (thisMin <= thatAlpha && thatAlpha <= thisMax ||
	    // thisMin <= thatBeta && thatBeta <= thisMax)
	    continue;

	  // If both thisAlpha and thisBeta are in (thatMin,thatMax),
	  // then only consider the pair when thisI == endI.
	  if (thisI == endI + 1 &&
	      compareM(thatMin, thisBeta) &&
	      compareM(thisBeta, thatMax))
	    // thatMin < thisBeta && thisBeta < thatMax)
	    continue;

	  pairs.push_back(thisI - 1);
	  pairs.push_back(thatI - 1);

#ifdef BLEEN
	  bool found = false;
	  for (int i = 0; !found && i < pairsS.size(); i += 2)
	    found = pairsS[i] == thisI - 1 && pairsS[i+1] == thatI - 1;
	  assert(found);
#endif
	}
      }
    }

#ifdef BLEEN
    if (pairs.size() != pairsS.size())
      for (int i = 0; i < pairsS.size(); i += 2) {
	int thisI = pairsS[i];
	int thatI = pairsS[i+1];
	bool found = false;
	for (int j = 0; !found && j < pairs.size(); j += 2)
	  found = pairsS[j] == thisI && pairsS[j+1] == thatI;
	assert(found);
      }
    assert(pairs.size() == pairsS.size());
#endif
  }
}
//#endif

bool Embedding::Chain::lt (int thisI, Embedding::Chain *that, int thatI) {
  int thisBack = this->angles.size() - 1;
  int thatBack = that->angles.size() - 1;

  if ((thisI == 0 || thisI == thisBack) &&
      (thatI == 0 || thatI == thatBack))
    return this->angles[thisI] < that->angles[thatI];

  double minAngle, maxAngle;
  if (isUpper()) {
    minAngle = 0;
    maxAngle = M_PI;
  }
  else {
    minAngle = -M_PI;
    maxAngle = 0;
  }

  if ((thisI == 0 || thisI == thisBack) && this->angles[thisI] == minAngle ||
      (thatI == 0 || thatI == thatBack) && that->angles[thatI] == maxAngle)
    return true;

  if ((thisI == 0 || thisI == thisBack) && this->angles[thisI] == maxAngle ||
      (thatI == 0 || thatI == thatBack) && that->angles[thatI] == minAngle)
    return false;

  return this->angles[thisI] <= that->angles[thatI];
}

void Embedding::Chain::sumK (Chain *that, Embedding *convK) {
  vector<int> pairs;
  overlap(that, pairs);

  int thisBack2 = this->angles.size() - 2;
  int thatBack2 = that->angles.size() - 2;
  double minAngle, maxAngle;
  if (isUpper()) {
    minAngle = 0;
    maxAngle = M_PI;
  }
  else {
    minAngle = -M_PI;
    maxAngle = 0;
  }

  for (int i = 0; i < pairs.size(); i += 2) {
    int thisI = pairs[i];
    int thatI = pairs[i+1];

    double thisAlpha = this->angles[thisI];
    double thisBeta = this->angles[thisI + 1];
    double thisMin = min(thisAlpha, thisBeta);
    double thisMax = max(thisAlpha, thisBeta);
    int thisMinI, thisMaxI;
    if (thisAlpha < thisBeta ||
	thisAlpha == thisBeta &&
	(thisI != 0 && thisI != thisBack2 ||
	 thisI == 0 && thisAlpha == minAngle ||
	 thisI == thisBack2 && thisBeta == maxAngle)) {
      thisMinI = thisI;
      thisMaxI = thisI + 1;
    }
    else {
      thisMinI = thisI + 1;
      thisMaxI = thisI;
    }

    double thatAlpha = that->angles[thatI];
    double thatBeta = that->angles[thatI + 1];
    double thatMin = min(thatAlpha, thatBeta);
    double thatMax = max(thatAlpha, thatBeta);
    int thatMinI, thatMaxI;
    if (thatAlpha < thatBeta ||
	thatAlpha == thatBeta &&
	(thatI != 0 && thatI != thatBack2 ||
	 thatI == 0 && thatAlpha == minAngle ||
	 thatI == thatBack2 && thatBeta == maxAngle)) {
      thatMinI = thatI;
      thatMaxI = thatI + 1;
    }
    else {
      thatMinI = thatI + 1;
      thatMaxI = thatI;
    }

    if (lt(thisMaxI, that, thatMinI) || !lt(thisMinI, that, thatMaxI))
      continue;
    
    bool swap = ((thisMinI == thisI) != (thatMinI == thatI));
    double alpha, beta;
    if (swap) {
      alpha = min(thisMax, thatMax);
      beta = max(thisMin, thatMin);
    }
    else {
      alpha = max(thisMin, thatMin);
      beta = min(thisMax, thatMax);
    }

    static int count;
    count++;

    Point thisTail = this->getP(thisI, alpha, swap);
    Point thisHead = this->getP(thisI, beta, !swap);
    Point thatTail = that->getP(thatI, alpha, swap);
    Point thatHead = that->getP(thatI, beta, !swap);

    Point tail = thisTail + thatTail;
    Point head = thisHead + thatHead;
    if (tail == head)
      continue;

    if (tail.x == head.x) {
      convK->addEdge(tail, head, 0, 0, 1, 0);
      continue;
    }

    Ptr<Function> f;
    double r = this->getR(thisI) + that->getR(thatI);
    bool isInconsistent = false;
    if (r == 0 || ((r > 0) ^ (tail.x < head.x) ^ swap ^ isUpper()) != 0) {
      cout << "inconsistent sum order" << endl;
      f = Line::alloc(tail, head);
      isInconsistent = true;
    }
    else
      f = sumF(that, thisI, thatI, tail, head);

    static int info2 = 100;
    int info = (tail.x > head.x) ^ isUpper();
    if (false && info == 1 && !isInconsistent)
      convK->addEdge(tail, head, f, 0, 1, ++info2);
    else
      convK->addEdge(tail, head, f, 0, 1, 0);
  }
}

inline bool onSegment (double x, vector<Segment> &segs, int i) {
  return i < segs.size() &&
    segs[i].getTail().x <= x && x <= segs[i].getHead().x;
}

void calcEnvelope (vector<Segment> &sums, bool upper, vector<Segment> &out) {
  if (sums.size() == 0)
    return;
  if (sums.size() == 1) {
    out.push_back(sums[0]);
    return;
  }

  int n1 = sums.size() / 2;
  int n2 = sums.size() - n1;
  vector<Segment> sums1;
  vector<Segment> sums2;
  for (int i = 0; i < n1; i++)
    sums1.push_back(sums[i]);
  for (int i = n1; i < sums.size(); i++)
    sums2.push_back(sums[i]);
  
  vector<Segment> out1;
  vector<Segment> out2;
  calcEnvelope(sums1, upper, out1);
  calcEnvelope(sums2, upper, out2);

  static int count;
  count++;
  // if (count == 15938)
  // cout << "this is it" << endl;

  Point tail;
  double x;
  Function *f = 0;
  int i1 = 0, i2 = 0;
  int sgn = 2 * upper - 1;

  while (i1 < out1.size() || i2 < out2.size()) {
    if (f == 0) {
      if (i2 == out2.size() ||
	  (i1 < out1.size() &&
	   (out1[i1].getTail().x < out2[i2].getTail().x ||
	    (out1[i1].getTail().x == out2[i2].getTail().x &&
	     out1[i1].getF()->compare(out2[i2].getF(),
				      out1[i1].getTail().x, 1) * sgn > 0)))) {
	tail = out1[i1].getTail();
	x = tail.x;
	f = out1[i1].getF();
      }
      else {
	tail = out2[i2].getTail();
	x = tail.x;
	f = out2[i2].getF();
      }
    }
  
    bool on1 = onSegment(x, out1, i1);
    bool on2 = onSegment(x, out2, i2);
    double x1 = Function::FINFINITY;
    double x2 = Function::FINFINITY;
    double x12 = Function::FINFINITY;
    if (on1)
      x1 = out1[i1].getHead().x;
    else if (i1 < out1.size())
      x1 = out1[i1].getTail().x;
    if (on2)
      x2 = out2[i2].getHead().x;
    else if (i2 < out2.size())
      x2 = out2[i2].getTail().x;
    if (on1 && on2) {
      if (f == out1[i1].getF())
	assert(out1[i1].getF()->compare(out2[i2].getF(), x, 1) * sgn >= 0);
      else if (f == out2[i2].getF())
	assert(out1[i1].getF()->compare(out2[i2].getF(), x, 1) * sgn <= 0);
      else
	assert(0);
      x12 = out1[i1].getF()->nextRoot(out2[i2].getF(), x);
    }
      
    Point head(min(min(x1, x2), x12), -Function::FINFINITY);
    if (head.x == x12)
      head.y = out1[i1].getF()->getRoot(out2[i2].getF(), head.x).y;
    else if (on1 && out1[i1].getF() == f) {
      if (head.x == x1)
	head.y = out1[i1].getHead().y;
      else
	head.y = f->getY(head.x);
    }
    else if (on2 && out2[i2].getF() == f) {
      if (head.x == x2)
	head.y = out2[i2].getHead().y;
      else
	head.y = f->getY(head.x);
    }
    else
      assert(0);

    if (on1 && head.x == x1)
      i1++;
    if (on2 && head.x == x2)
      i2++;

    bool newOn1 = onSegment(head.x, out1, i1);
    bool newOn2 = onSegment(head.x, out2, i2);
    Function *newF = 0;
    if (newOn1 && !newOn2)
      newF = out1[i1].getF();
    else if (!newOn1 && newOn2)
      newF = out2[i2].getF();
    else if (newOn1 && newOn2) {
      if (out1[i1].getF()->compare(out2[i2].getF(), head.x, 1) * sgn > 0)
	newF = out1[i1].getF();
      else
	newF = out2[i2].getF();
    }

    if (newF != f) {
      out.push_back(Segment(tail, head, f));
      if (newF != 0) {
	tail.x = head.x;
	if (tail.x == x12 && !(on1 && head.x == x1) && !(on2 && head.x == x2))
	  tail.y = head.y;
	else if (newOn1 && out1[i1].getF() == newF) {
	  if (tail.x == out1[i1].getTail().x)
	    tail.y = out1[i1].getTail().y;
	  else
	    tail.y = newF->getY(tail.x);
	}
	else if (newOn2 && out2[i2].getF() == newF) {
	  if (tail.x == out2[i2].getTail().x)
	    tail.y = out2[i2].getTail().y;
	  else
	    tail.y = newF->getY(tail.x);
	}
	else
	  assert(0);
      }
      f = newF;
    }

    x = head.x;
  }
}

void Embedding::Chain::sumM (Chain *that, Embedding *convM,
			     int thatChainSize) {
  vector<int> pairs;
  overlap(that, pairs);

  // double minXX = this->minX() + that->minX();
  vector<Segment> segs;
  for (int i = 0; i < pairs.size(); i += 2) {
    int thisI = pairs[i];
    double thisAlpha = this->getA(thisI);
    double thisBeta = this->getA(thisI + 1);
    double thisMin = minM(thisAlpha, thisBeta);
    double thisMax = maxM(thisAlpha, thisBeta);

    int thatI = pairs[i+1];
    double thatAlpha = that->getA(thatI);
    double thatBeta = that->getA(thatI + 1);
    double thatMin = minM(thatAlpha, thatBeta);
    double thatMax = maxM(thatAlpha, thatBeta);

    if (thisAlpha == thisBeta && compareM(thatBeta, thatAlpha))
      // thatAlpha > thatBeta)
      continue;

    if (compareM(thisBeta, thisAlpha) && thatAlpha == thatBeta)
      // thisAlpha > thisBeta 
      continue;

    // if (thisMax < thatMin || thatMax < thisMin)
    // continue;

    if (thisAlpha == thisBeta && thatAlpha == thatBeta)
      continue;

    // if (thisAlpha == 0 && thisBeta == 0)
    // continue;

    // if (thatAlpha == 0 && thatBeta == 0)
    // continue;

    if (thisAlpha == thisBeta || thatAlpha == thatBeta) {
      Point tail, head;
      if (compareM(thatAlpha, thatBeta)) {
	double xError;
	Point p = that->getP(thatI, thisAlpha, xError);
	Timing::begin(Timing::adapt);
	tail = sum(this->getP(thisI), 0, p, xError);
	head = sum(this->getP(thisI+1), 0, p, xError);
	Timing::end(Timing::adapt);
      }
      else if (compareM(thisAlpha, thisBeta)) {
	double xError;
	Point p = this->getP(thisI, thatAlpha, xError);
	Timing::begin(Timing::adapt);
	tail = sum(that->getP(thatI), 0, p, xError);
	head = sum(that->getP(thatI+1), 0, p, xError);
	Timing::end(Timing::adapt);
      }
      else
	assert(0);
      if (tail.x == head.x)
	continue;
      assert((tail.x > head.x) == sw);

      if (isUpper())
	segs.push_back(Segment(head, tail, Line::alloc(tail, head)));
      else
	segs.push_back(Segment(tail, head, Line::alloc(tail, head)));
      continue;
    }

    if (compareM(thisBeta, thisAlpha) && compareM(thatBeta, thatAlpha))
      // thisAlpha > thisBeta && thatAlpha > thatBeta)
      continue;

    // bool swap = (thisAlpha > thisBeta || thatAlpha > thatBeta);
    bool swap = compareM(thisBeta, thisAlpha) || compareM(thatBeta, thatAlpha);
    double alpha, beta;
    if (swap) {
      alpha = minM(thisMax, thatMax);
      beta = maxM(thisMin, thatMin);
    }
    else {
      alpha = maxM(thisMin, thatMin);
      beta = minM(thisMax, thatMax);
    }

    static int count;
    count++;
    // if (count == 28410)
    // cout << "this is it" << endl;

    double thisTailError, thisHeadError;
    Point thisTail = this->getP(thisI, alpha, thisTailError);
    Point thisHead = this->getP(thisI, beta, thisHeadError);
    double thatTailError, thatHeadError;
    Point thatTail = that->getP(thatI, alpha, thatTailError);
    Point thatHead = that->getP(thatI, beta, thatHeadError);
    Timing::begin(Timing::adapt);
    Point tail = sum(thisTail, thisTailError, thatTail, thatTailError);
    Point head = sum(thisHead, thisHeadError, thatHead, thatHeadError);
    Timing::end(Timing::adapt);

#ifdef BLEEN
    if (isUpper()) {
      if (head.x < minXX)
	head.x = minXX;
    }
    else if (tail.x < minXX)
      tail.x = minXX;
#endif

    if (tail.x == head.x || (tail.x < head.x) == isUpper())
      continue;

    Ptr<Function> f;
    double r = this->getR(thisI) + that->getR(thatI);
    bool isInconsistent = false;
    if (r == 0 || ((r > 0) ^ (tail.x < head.x) ^ swap ^ isUpper()) != 0) {
      cout << "inconsistent sum order" << endl;
      f = Line::alloc(tail, head);
      isInconsistent = true;
    }
    else
      f = sumF(that, thisI, thatI, tail, head);

    if (isUpper())
      segs.push_back(Segment(head, tail, f));
    else
      segs.push_back(Segment(tail, head, f));
  }

  vector<Segment> envelope;
  // segdump("segs.fig", segs);
  calcEnvelope(segs, isUpper(), envelope);
  // segdump("envelope.fig", envelope);
  int mult = 1 - 2 * isUpper();
  Point prev =
    isUpper() ? this->head() + that->head() : this->tail() + that->tail();

  int pairID = (this->index * thatChainSize + that->index) * 5;

  for (int i = 0; i < envelope.size(); i++) {
    Segment &s = envelope[i];
    Point tail = s.getTail();
    Point head = s.getHead();
    if (tail != prev) {
      assert(tail.x == prev.x);
      convM->addEdge(prev, tail, 0, 0, mult, 0);
    }
    convM->addEdge(tail, head, s.getF(), 0, mult, pairID);
    prev = head;
  }
  Point next =
    isUpper() ? this->tail() + that->tail() : this->head() + that->head();
  if (prev != next) {
    assert(prev.x == next.x);
    convM->addEdge(prev, next, 0, 0, mult, 0);
  }

  Chain *tt[2] = { this, that };
  for (int ia = 0; ia <= 1; ia++)
    for (int th = 0; th <= 1; th++)
      if (tt[ia]->endIsConcave[th])
	tt[!ia]->sumM(tt[ia]->end(th), pairID + 1 + 2 * ia + th, convM);
}

Point Embedding::Chain::sum (Point a, double aError, Point b, double bError) {
  if (aError == 0 && bError == 0)
    return a + b;
  double error1 = aError + bError, error2 = 0;
  if (aError == 0)
    a.x = sum(a.x, error1, error1);
  else if (bError == 0)
    b.x = sum(b.x, error1, error1);
  else
    assert(0);
  Point p(sum(a.x, b.x, error2), a.y + b.y);

  if (error2 > 0) {
    if (error1 > 0) {
      double px = nextafter(p.x, Function::FINFINITY);
      if (error2 == (px - p.x) / 2)
	p.x = px;
    }
  }
  else if (error2 < 0) {
    if (error1 < 0) {
      double px = nextafter(p.x, -Function::FINFINITY);
      if (error2 == (px - p.x) / 2)
	p.x = px;
    }
  }

  //  if (error2 < 0 || error2 == 0 && error1 < 0)
  // p.x = nextafter(p.x, -Function::FINFINITY);
  return p;
}

void Embedding::Chain::sumM (Point p, int pairID, Embedding *convM) {
  for (int i = 1; i < angles.size() - 2; i += 2) {
    Point tail = getP(i) + p;
    Point head = getP(i+1) + p;
    if (tail == head)
      continue;
    if (tail.x == head.x) {
      convM->addEdge(head, tail, 0);
      continue;
    }

    double r = getR(i);
    if (r == Function::FINFINITY) {
      convM->addEdge(head, tail, Line::alloc(tail, head), 0, 1, pairID);
      continue;
    }
    assert(r != 0);

    Function *f = 0;
    Point c = getC(i) + p;
    if (c.x - r == c.x + r)
      f = Line::alloc(tail, head);
    else {
      Circle *circle = Circle::alloc(c, fabs(r));
      f = circle->getBranch(isUpper() ^ (r < 0));
    }
    convM->addEdge(head, tail, f, 0, 1, pairID);
  }
}

Point Embedding::Chain::getP (int i, double angle, bool swap) {
  if (i % 2 == 0)
    return getP(i);
  if (angles[i] == angles[i+1])
    return getP(i + swap);
  if (angle == angles[i])
    return getP(i);
  if (angle == angles[i+1])
    return getP(i + 1);
  return getC(i) + Point(getR(i), angle, true);
}

Point Embedding::Chain::getP (int i, double angle, double &xError) {
  xError = 0;
  if (i % 2 == 0)
    return getP(i);
  double alpha = angles[i];
  double beta = angles[i+1];
  Point a = getP(i);
  Point b = getP(i+1);
  if (angle == alpha)
    return a;
  if (angle == beta)
    return b;

  if (compareM(alpha, beta) == !compareM(M_PI/2, angle))
    return getP(a, alpha, b, beta, angle, xError);
  else
    return getP(b, beta, a, alpha, angle, xError);
}
  
Point Embedding::Chain::getP
(Point a, double alpha, Point b, double beta, double theta, double &xError) {
  double alpha0 = alpha == M_PI ? 0 : alpha;
  double beta0 = beta == M_PI ? 0 : beta;
  double ab = beta0 - alpha0;
  // accurate value of cos(beta) - cos(alpha):
  double cab = (alpha >= 0 && alpha != M_PI) != (beta >= 0 && beta != M_PI) ?
    cosM(beta) - cosM(alpha) :
    -(sinM(alpha) + sinM(beta)) * sin(ab) / (1 + cos(ab));
  double abx = b.x - a.x;

  double at = theta - alpha0;
  // accurate value of cos(theta) - cos(alpha):
  double cat = -(sinM(alpha) + sinM(theta)) * sin(at) / (1 + cos(at));
  double catDcab = cat / cab;
  double apx = catDcab * abx;
  double apy = -apx / tan((alpha0 + theta) / 2);

  return Point(sum(a.x, apx, xError), a.y + apy);
}

Ptr<Function> Embedding::Chain::sumF (Chain *that, int thisI, int thatI,
				      Point tail, Point head) {
  if (tail.x == head.x)
    return 0;

  if (this->angles[thisI] == this->angles[thisI+1] ||
      that->angles[thatI] == that->angles[thatI+1])
    return Line::alloc(tail, head);

  double thisR = this->getR(thisI);
  Point thisC = this->getC(thisI);
  double thatR = that->getR(thatI);
  Point thatC = that->getC(thatI);

  Point c = thisC + thatC;

#ifndef SHORT_ARCS
  Point th = head - tail;
  if ((c - tail) * th < 0 ||
      (c - head) * th > 0) {
    double l = th.length();
    static double maxShortArc;
    if (maxShortArc < l) {
      maxShortArc = l;
      cout << "maxShortArc " << maxShortArc << endl;
    }
    return Line::alloc(tail, head);
  }
#endif

  double r = thisR + thatR;
  if (r == 0 || c.x - r == c.x + r)
    return Line::alloc(tail, head);
  Circle *circle = Circle::alloc(c, fabs(r));
  return circle->getBranch(isUpper() == (r > 0));
}

Point Embedding::minmaxX (int iLoop) {
  Point minmax(Function::FINFINITY, -Function::FINFINITY);
  int km = (loops[iLoop].chainM != -1);
  assert(loops[iLoop].chain(km) != -1);
  vector<Chain *> &chains = km == 0 ? chainsK : chainsM;
  for (int i = loops[iLoop].chain(km);
       i < chains.size() && chains[i]->loop == iLoop; i++) {
    Chain *chain = chains[i];
    double minX = chain->minX();
    if (minmax.x > minX)
      minmax.x = minX;
    double maxX = chain->maxX();
    if (minmax.y < maxX)
      minmax.y = maxX;
  }
  return minmax;
}

void reportBubble () {
  cout << "found bubble!" << endl;
}

double calcOverlapArea (Embedding *aa, int aLoop, Embedding *bb, int bLoop,
			Point t, double r) {
  static int count;
  count++;
  static double maxOverlapArea;
  extern double maxXY;
  Embedding *a = aa->loop(aLoop);
  Embedding *b = bb->loop(bLoop);
  Transformation tr(t, Point(-1, 0));
  Embedding t_b(b, tr);
  vector<Embedding *> sets(2);
  sets[0] = a;
  sets[1] = &t_b;
  vector<int> cn(2);
  cn[0] = a->cells.at(0).state;
  cn[1] = t_b.cells.at(0).state;
  if (count == 609) {
    a->edgedraw("a.fig");
    b->edgedraw("b.fig");
    t_b.edgedraw("t-b.fig");
  }
  Embedding ant_b(sets, cn, Embedding::Intersection());
  if (ant_b.isNull()) {
    if (r / maxXY < 1e-13 && maxOverlapArea < r / maxXY) {
      maxOverlapArea = r / maxXY;
      cout << "maxOverlapArea2 " << maxOverlapArea << endl;
    }

    delete a;
    delete b;
    return 0;
  }
  // double overlapArea = fabs(ant_b.area() / maxXY / maxXY);
  double overlapArea = ant_b.size() / maxXY;
  if (overlapArea > 1e-12)
    Timing::nbubbles++;
  // cout << "overlapArea " << overlapArea << " overlapR " << r << endl;
  if (r == 0 && maxOverlapArea < overlapArea) {
    maxOverlapArea = overlapArea;
    cout << "maxOverlapArea " << maxOverlapArea << endl;
    a->edgedraw("a.fig");
    b->edgedraw("b.fig");
    t_b.edgedraw("t-b.fig");
    ant_b.edgedraw("ant-b.fig");
  }
  static double maxOverlapR;
  if (maxOverlapR < r) {
    maxOverlapR = r;
    extern double maxXY;
    cout << "maxOverlapR " << r/maxXY << " overlapArea " << overlapArea << endl;
  }
  delete a;
  delete b;
  return overlapArea;
}

Embedding *Embedding::loop (int iLoop) {
  Embedding *emb = new Embedding();
  DEdge &d = loops[iLoop].d;
  DEdge dedge = d;
  do {
    emb->addEdge(dedge.tail()->p, dedge.head()->p, dedge.edge()->f)->height =
      dedge.edge()->height;
  } while(++dedge != d);
  emb->build();
  CellState even = isHole(iLoop) ? INSIDE : OUTSIDE;
  CellState odd = isHole(iLoop) ? OUTSIDE : INSIDE;
  for (int i = 0; i < emb->cells.size(); i++)
    emb->cells[i].state = emb->cells[i].depth % 2 == 0 ? even : odd;
  return emb;
}

Embedding *Embedding::sumLoops
(Embedding *that, int thisLoop, int thatLoop, bool km) {
  static int count;
  count++;
  // if (count == 187)
  // cout << "this is it" << endl;

  //#define DEBUG_FIGS
#ifdef DEBUG_FIGS
  this->loops[thisLoop].edgedraw("thisLoop.fig");
  that->loops[thatLoop].edgedraw("thatLoop.fig");
#endif

  Embedding *sum = new Embedding();
  bool thisIsHole = this->isHole(thisLoop);
  bool thatIsHole = that->isHole(thatLoop);
  if (thisIsHole && thatIsHole) {
    sum->build();
    sum->cells.at(0).state = INSIDE;
    return sum;
  }

  if (thisIsHole || thatIsHole) {
    Point thisMinMax = this->minmaxX(thisLoop);
    Point thatMinMax = that->minmaxX(thatLoop);
    double minX = thisMinMax.x + thatMinMax.y;
    double maxX = thisMinMax.y + thatMinMax.x;
    if (thisIsHole && minX >= maxX ||
	thatIsHole && minX <= maxX) {
      sum->build();
      sum->cells.at(0).state = INSIDE;
      return sum;
    }
  }

  Timing::begin(Timing::sumloops);

  vector<Chain *> &thisChains = km == 0 ? this->chainsK : this->chainsM;
  vector<Chain *> &thatChains = km == 0 ? that->chainsK : that->chainsM;
  for (int thisI = this->loops[thisLoop].chain(km);
       thisI < thisChains.size() && thisChains[thisI]->loop == thisLoop;
       thisI++) {
    Chain *thisChain = thisChains[thisI];

    for (int thatI = that->loops[thatLoop].chain(km);
	 thatI < thatChains.size() && thatChains[thatI]->loop == thatLoop;
	 thatI++) {
      Chain *thatChain = thatChains[thatI];

      if (thisChain->isUpper() == thatChain->isUpper())
	if (km == 0)
	  thisChain->sumK(thatChain, sum);
	else
	  thisChain->sumM(thatChain, sum, thatChains.size());
    }
  }

  sum->checkMults();

  // Make all horizontal edges with the same y have the same function.
  double minX = Function::FINFINITY, maxX = -Function::FINFINITY;
  for (int i = 0; i < sum->edges.size(); i++) {
    Edge *edge = sum->edges[i];
    if (minX > edge->tail()->p.x)
      minX = edge->tail()->p.x;
    if (maxX < edge->head()->p.x)
      maxX = edge->head()->p.x;
  }
  for (int i = 0; i < sum->edges.size(); i++) {
    Edge *edge = sum->edges[i];
    double y = edge->tail()->p.y;
    if (edge->head()->p.y != y)
      continue;
    if (edge->f->getType() == Function::UPPER_CIRCLE ||
	edge->f->getType() == Function::LOWER_CIRCLE) {
      Circle *circle = (dynamic_cast<Cbranch *>((Function *) edge->f))->circle;
      Point th = edge->head()->p - edge->tail()->p;
      if ((edge->tail()->p - circle->c) * th < 0 &&
	  (edge->head()->p - circle->c) * th > 0)
	continue;
    }
    else
      assert(edge->f->getType() == Function::LINE);
    edge->f = Line::alloc(Point(minX, y), Point(maxX, y));
  }
  
  if (km == 0) {
    sum->arrange1d(sum->edges);
    sum->arrange1d(sum->edgesV);

Timing::end(Timing::sumloops);
Timing::begin(Timing::sumarrange);

#ifdef DEBUG_FIGS
    extern char convName[];
    char buf[64];
    sprintf(buf, "%sK.fig", convName);
    sum->edgedraw(buf);
#endif  

    sum->arrange();
    sum->splitAtNearVerts(1e-13);
    Timing::ncells += sum->cells.size();
    bool unbounded = (thisIsHole || thatIsHole);
    vector<int> cn;
    cn.push_back(unbounded);
    NonZero test;
    
    sum->setCellState(cn, test);
    
#ifdef BLEEN
    for (int i = 0; i < sum->edges.size(); i++) {
      Edge *edge = sum->edges[i];
      if (edge->info == 0)
	continue;
      Cell &cell = sum->cells[sum->loops[edge->loop[edge->mult == 1]].cell];
    }
#endif

#ifdef BUBBLE_AREA
    double minOutsideArea = 1e9;
    for (int i = 1; i < sum->cells.size(); i++) {
      Cell &cell = sum->cells[i];
      if (cell.state == OUTSIDE && cell.loops.size() > 0) {
	// double area = fabs(sum->loops[edge->loop[edge->mult == 1]].area());
	double area = fabs(sum->loops[cell.loops[0]].area());
	if (area < 2e-12) {
	  cout << "found bubble! " << area << endl;
	  Timing::nbubbles++;
	  cell.state = INSIDE;
	}
	else if (area < minOutsideArea) {
	  cout << "outside cell " << area << endl;
	  minOutsideArea = area;
	}
      }
    }
#else
    Timing::begin(Timing::adapt);
    double minOutsideR = 1e-5;
    Point ps[sum->cells.size()];
    double rs[sum->cells.size()];
    sum->pointInside(ps, rs);
    for (int i = 1; i < sum->cells.size(); i++) {
      Cell &cell = sum->cells[i];
      if (cell.state == OUTSIDE) {
#ifdef BUBBLE_TEST
	double overlapArea = calcOverlapArea(this, thisLoop, that, thatLoop, ps[i], rs[i] > 0 ? rs[i] : nextafter(0, 1));
	// if (overlapArea > 1e-12)
	// cell.state = INSIDE;
#endif
	double r = rs[i];
	extern double maxXY;
	if (r / maxXY < 1e-13) {
	  // #ifndef BUBBLE_TEST
	  // cout << "found bubble! " << r << endl;
	  Timing::nbubbles++;
	  cell.state = INSIDE;
	  // #endif
	}
	else if (r < minOutsideR) {
	  // cout << "outside cell " << r << endl;
	  minOutsideR = r;
	}
      }
    }
    Timing::end(Timing::adapt);
#endif    
  }
  else {
    sum->sumInfo = true;
    sum->emb[0] = this;
    sum->emb[1] = that;
#ifdef DEBUG_FIGS
    sum->edgedraw("tf.fig");
#endif

    for (int i = 0; i < sum->edges.size(); i++) {
      Edge *edge = sum->edges[i];
      edge->id = edge->info / 5;
      edge->info = (1 << (edge->info % 5));
    }

    sum->arrange1d(sum->edges);
    sum->arrange1d(sum->edgesV);

Timing::end(Timing::sumloops);
Timing::begin(Timing::sumarrange);

#ifdef DEBUG_FIGS
    extern char convName[];
    char buf[64];
    sprintf(buf, "%sM.fig", convName);
    sum->edgedraw(buf);
#endif  

    for (int i = 0; i < sum->edges.size(); i++) {
      Edge *edge = sum->edges[i];
      if ((edge->info & 1) != 0 && edge->info != 1)
	for (int j = 1; j <= 4; j++)
	  if ((edge->info & (1 << j))) {
	    edge->info ^= (1 << j);
	    break;
	  }
      if (edge->info != 1) {
	assert(edge->info != 0);
	int j = 1;
	while ((edge->info & (1 << j)) == 0) j++;
	int ne = 1;
	for (int k = j+1; k <= 4; k++)
	  if ((edge->info & (1 << k)) != 0)
	    ne++;
	assert(edge->mult == ne || edge->mult == -ne);
	edge->mult /= ne;
	for (int k = j+1; k <= 4; k++)
	  if ((edge->info & (1 << k)) != 0)
	    sum->addEdge(edge->tail()->p, edge->head()->p, edge->f,
			 edge->id, edge->mult, (1 << k));
	edge->info = (1 << j);
	edge->id = edge->id * 5 + j;
      }
      else
	edge->id *= 5;
    }

    sum->checkMults();

    sum->arrange();

    Timing::ncells += sum->cells.size();

    bool unbounded = (thisIsHole || thatIsHole);
    vector<int> cn;
    cn.push_back(unbounded);
    NonZero test;
    // sum->fillMSlits (this, that, &pairIDSet);
    sum->setCellState(cn, test);
#ifdef T_CURVE_HACK
    for (int i = 0; i < sum->cells.size(); i++) {
      Cell &cell = sum->cells[i];
      if (cell.state != OUTSIDE || cell.loops.size() != 1)
	continue;

      DEdge start = sum->loops[cell.loops[0]].d, curr = start;
      while (curr.edge()->info % 5 != 0 && ++curr != start);
      if (curr.edge()->info % 5 != 0) {
	cout << "fill hole!! "
	     << start.edge()->tail()->p << " "
	     << start.edge()->head()->p.x - start.edge()->tail()->p.x
	     << endl;
	cell.state = INSIDE;
      }
    }
#endif

    sum->sumInfo = false;
  }

  // if (count == 19)
  // cout << "this is it" << endl;
  sum->getBoundary();

#ifdef DEBUG_FIGS
  sum->edgedraw("sum.fig");
#endif

#ifdef BUBBLE_TEST
  if (km == 0 || km == 1) {
    Timing::begin(Timing::timeFG);

    for (Verti it = sum->verts.begin(); it != sum->verts.end(); ++it) {
      Point p = (*it)->p;
      double oa = calcOverlapArea(this, thisLoop, that, thatLoop, p, 0);
      if (oa > 1e-3) {
	sum->edgedraw("sum.fig");
	cout << "p " << p << endl;
      }
    }
    
    for (int i = 0; i < sum->edges.size(); i++) {
      Edge *edge = sum->edges[i];
      double x = (edge->tail()->p.x + edge->head()->p.x) / 2;
      double y = edge->f->getY(x);
      Point p(x, y);
      double oa = calcOverlapArea(this, thisLoop, that, thatLoop, p, 0);
    }
    
    for (int i = 0; i < sum->edgesV.size(); i++) {
      Edge *edge = sum->edgesV[i];
      double x = edge->tail()->p.x;
      double y = (edge->tail()->p.y + edge->head()->p.y) / 2;
      Point p(x, y);
      double oa = calcOverlapArea(this, thisLoop, that, thatLoop, p, 0);
    }

    Point ps[sum->cells.size()];
    double rs[sum->cells.size()];
    sum->pointInside(ps, rs);
    //cout << "cell sizes";
    for (int i = 1; i < sum->cells.size(); i++)
      if (sum->cells[i].state == OUTSIDE) {
	calcOverlapArea(this, thisLoop, that, thatLoop, ps[i], rs[i]);
#ifdef BLEEN
	cout << " " << i << " " << rs[i] << endl;
	if (rs[i] < 1e-6) {
	  cout << sum->loops[sum->cells[i].loops[0]].d.tail()->p << endl;
	  cout << sum->loops[sum->cells[i].loops[0]].area() << endl;
	  cout << "bubble!!!" << endl;
	}
#endif
      }
    // cout << endl;
    Timing::end(Timing::timeFG);
  }
#endif

  //#define COMPARE_KM
#ifdef COMPARE_KM
  if (km == 0) {
    if (this->chainsM.size() == 0)
      this->buildChains(1);
    if (that->chainsM.size() == 0)
      that->buildChains(1);
    
    Embedding *sumM = sumLoops(that, thisLoop, thatLoop, 1);
    sumM->edgedraw("sumM.fig");
    vector<Embedding *> sets(2);
    sets[0] = sum;
    sets[1] = sumM;
    vector<int> cn(2);
    cn[0] = sum->cells.at(0).state;
    cn[1] = sumM->cells.at(0).state;
    Embedding diff(sets, cn, Embedding::XOR());
    diff.edgedraw("diff.fig");
    static double maxArea;
    double area = diff.area();
    int nOcellsK = 0;
    for (int i = 0; i < sum->cells.size(); i++)
      if (sum->cells[i].state == OUTSIDE)
	nOcellsK++;
    int nOcellsM = 0;
    for (int i = 0; i < sumM->cells.size(); i++)
      if (sumM->cells[i].state == OUTSIDE)
	nOcellsM++;
    if (nOcellsK != nOcellsM)
      cout << "num cells " << nOcellsK << " " << nOcellsM << endl;
    if (maxArea < area) {
      maxArea = area;
      cout << "maxArea " << maxArea << endl;
      Embedding *sumK = sumLoops(that, thisLoop, thatLoop, 0);
    }
    delete sumM;
  }
#endif

#ifdef DEBUG_FIGS
  double area1 = sum->area();
#endif
  for (int i = 0; i < sum->edges.size(); i++)
    sum->edges[i]->id = sum->edges[i]->info = 0;
  sum->mergeEdges();
#ifdef DEBUG_FIGS
  sum->edgedraw("sumME.fig");
  double area2 = sum->area();
  double areaD = fabs(area1 - area2);
  static double areaDMax;
  if (areaDMax < areaD) {
    areaDMax = areaD;
    cout << "areaDMax " << areaDMax << endl;
  }
#endif

  Timing::end(Timing::sumarrange);

  return sum;
}

#ifdef BLEEN
void Embedding::fillMBubbles (Embedding *a, Embedding *b,
			      PairIDSet *pairIDSet) {
  Embedding *emb[2] = { a, b };

  // Current active edge for a chain.
  map<int, Edge *> chains;
  vector<int> cset;

  SweepNode *sweep = 0;

  // Vectors of outgoing and incoming edges.
  vector<Edge *> edgesTH[2];

  // Related chains 
  vector<int> other[2 /* below / above */];

  Verti jv;
  for (Verti iv = verts.begin(); iv != verts.end(); iv = jv) {
    // Determine range of vertices with same x.
    double x = (*iv)->p.x;
    for (jv = iv; jv != verts.end() && (*jv)->p.x == x; ++jv);
    
    // Get incoming and outgoing edges for this x.
    for (Verti iiv = iv; iiv != jv; ++iiv) {
      Vert *vert = *iiv;
      for (int th = 0; th <= 1; th++) {
	for (Edge *edge = vert->edge[th][0];
	     edge != 0 && edge->getVert(th) == vert;
	     edge = edge->edge[th][1])
	  edgesTH[th].push_back(edge);
      }
    }

    // Remove incoming edges from sweep list.
    for (int i = 0; i < edgesTH[1].size(); i++) {
      Edge *edge = edgesTH[1][i];
      sweep = edge->node->remove();
      delete edge->node;
      pairIDSet->getSet(edge->info, cset);
      for (int j = 0; j < cset.size(); j++) {
	assert(chains[cset[j]] == edge);
	chains[cset[j]] = 0;
      }
#ifdef BLEEN
      cout << "chains " << edge->info << " " << 0 << " "
	   << edge->head()->p.x << endl;
#endif
    }

    // Insert outgoing edges into the sweep list.
    for (int i = 0; i < edgesTH[0].size(); i++) {
      SweepNode *node = new HeightNode(edgesTH[0][i]);
      if (sweep == 0)
	sweep = node;
      else {
	sweep->insert(node);
	sweep = node->rebalance();
      }
      Edge *edge = node->edge;
      pairIDSet->getSet(edge->info, cset);
      for (int j = 0; j < cset.size(); j++) {
	Edge *edge2 = chains[cset[j]];
	assert(chains[cset[j]] == 0);
	chains[cset[j]] = edge;
      }
    } 

    for (int i = 0; i < edgesTH[0].size(); i++) {
      Edge *edge = edgesTH[0][i];

      other[0].clear();
      other[1].clear();

      pairIDSet->getSet(edge->info, cset);
      for (int j = 0; j < cset.size(); j++) {
	int cct = cset[j];
	int chain[2] =  { (cct / 5) / b->chains.size(),
			  (cct / 5) % b->chains.size() };
	bool sw = a->chains[chain[0]]->sw;
	assert(sw == b->chains[chain[1]]->sw);
	int type = cct % 5;
	if (type != 0) {
	  // assert(chains[cct - type] != 0);
	  other[sw].push_back(cct - type);
	  int ab = (type - 1) / 2;
	  int th = (type - 1) % 2;
	  int aNext = a->nextChainIndex(chain[0], th);
	  int bNext = b->nextChainIndex(chain[1], th);
	  // other[sw ^ !emb[!ab]->chains[chain[!ab]]->endIsConcave[th]].push_back
	  // ((aNext * b->chains.size() + bNext) * 5 + 2 * ab + !th + 1);
	}
	else {
	  for (int th = 0; th <= 1; th++) {
	    if (!a->chains[chain[0]]->endIsConcave[th] &&
		!b->chains[chain[1]]->endIsConcave[th]) {
	      int aNext = a->nextChainIndex(chain[0], th);
	      int bNext = b->nextChainIndex(chain[1], th);
	      other[!sw].push_back((aNext * b->chains.size() + bNext) * 5);
	    }
	    for (int ab = 0; ab <= 1; ab++)
	      if (emb[ab]->chains[chain[ab]]->endIsConcave[th])
		other[!sw].push_back(cct + 2 * ab + th + 1);
	  }
	}
      }

      for (int du = 0; du <= 1; du++)
	for (int j = 0; j < other[du].size(); j++) {
	  int cct2 = other[du][j];
	  Edge *edge2 = chains[cct2];
	  if (edge2 == 0 || (edge->height < edge2->height) == du)
	    continue;
	  cout << "potential bubble!" << endl;
	  for (Edge *e = edge; e != edge2; e = e->node->stepEdge(!du)) {
	    Cell &cell = cells[loops[e->loop[du]].cell];
	    if (cell.state == OUTSIDE) {
	      cout << "filling bubble! "
		   << loops[e->loop[du]].area() << endl;
	      cell.state = INSIDE;
	    }
	  }
	}
    }

    edgesTH[0].clear();
    edgesTH[1].clear();
  }

}
#endif
  
int Embedding::nextChainIndexM (int i, int th) {
  int loop = chainsM[i]->loop;
  if (th == 0) {
    if (i-1 >= 0 && chainsM[i-1]->loop == loop)
      return chainsM[i-1]->index;
    if (chainsM.back()->loop == loop)
      return chainsM.back()->index;
    while (loops.at(++loop).chainM == -1);
    return chainsM[loops[loop].chainM - 1]->index;
  }
  else {
    if (i+1 < chainsM.size() && chainsM[i+1]->loop == loop)
      return chainsM[i+1]->index;
    return chainsM[loops[loop].chainM]->index;
  }
}

Embedding *Embedding::sumCells
(Embedding *that, int thisCellI, int thatCellI, bool km) {
  Cell &thisCell = this->cells[thisCellI];
  Cell &thatCell = that->cells[thatCellI];
  vector<Embedding *> in;
  vector<int> cn;
  int id = 0;
  for (int thisI = 0; thisI < thisCell.loops.size(); thisI++)
    for (int thatI = 0; thatI < thatCell.loops.size(); thatI++) {
      in.push_back
	(sumLoops(that, thisCell.loops[thisI], thatCell.loops[thatI], km));
      cn.push_back(in.back()->isUnbounded());
    }

  return new Embedding(in, cn, Intersection());
}

Embedding *Embedding::sum (Embedding *that, bool km) {
  Timing::begin(Timing::sum);

  if (this->chains(km).size() == 0)
    this->buildChains(km);

  if (that->chains(km).size() == 0)
    that->buildChains(km);

  vector<Embedding *> in;
  vector<int> cn;
  Union test;
  for (int thisI = 0; thisI < this->cells.size(); thisI++) {
    if (this->cells[thisI].state != INSIDE)
      continue;
    for (int thatI = 0; thatI < that->cells.size(); thatI++) {
      if (that->cells[thatI].state != INSIDE)
	continue;
      in.push_back(sumCells(that, thisI, thatI, km));
      cn.push_back(in.back()->isUnbounded());
    }
  }

  Timing::end(Timing::sum);
  return new Embedding(in, cn, Union());
}

static inline void updateInCircle
(Embedding::Edge *b, Embedding::Edge *a, Embedding::Edge *e, double xRight,
 Point &c, double &r) {
  double xLeft =
    e == a ? a->tail()->p.x : e == b ? b->tail()->p.x : e->head()->p.x;
  double dx = xRight - xLeft;
  if (dx <= r)
    return;
  
  double xMid = (xLeft + xRight) / 2;
  double yB = b->f->getY(xMid);
  double dy = a->f->getY(xMid) - yB;
  if (r > 0 && dy <= r)
    return;
  
  c.x = xMid;
  c.y = yB + dy/2;
  r = min(dx, dy);
  if (r < 0)
    r = nextafter(0, 1);
}

double Embedding::size () {
  int ncells = cells.size();
  Point ps[ncells];
  double rs[ncells];
  pointInside(ps, rs);

  double rMax = 0;

  for (int i = 0; i < ncells; i++)
    if (cells[i].state == INSIDE && rs[i] > rMax)
      rMax = rs[i];

  return rMax;
}

void Embedding::pointInside (Point *ps, double *rs) {
  Timing::begin(Timing::pointInside);
  int ncells = cells.size();
  for (int i = 0; i < ncells; i++)
    rs[i] = 0;

  int nedges = edges.size();
  int id[nedges];
  int info[nedges];
  for (int i = 0; i < nedges; i++) {
    id[i] = edges[i]->id;
    info[i] = edges[i]->info;
    edges[i]->id = i;
    edges[i]->info = -1;
  }

  Edge *below[nedges], *above[nedges], *bottom = 0;

  Verti jv;
  for (Verti iv = verts.begin(); iv != verts.end(); iv = jv) {
    // Determine range of vertices with same x.
    double x = (*iv)->p.x;
    for (jv = iv; jv != verts.end() && (*jv)->p.x == x; ++jv);
    
    // Remove incoming edges for this x.
    for (Verti iiv = iv; iiv != jv; ++iiv) {
      Vert *v = *iiv;
      for (Edge *e = v->edge[1][1];
	   e != 0 && e->getVert(1) == v;
	   e = e->edge[1][0]) {
	Edge *b = below[e->id];
	Edge *a = above[e->id];

	if (b != 0) {
	  int icell = loops[b->loop[0]].cell;
	  if (loops[e->loop[1]].cell == icell)
	    updateInCircle(b, e, edges[b->info], x, ps[icell], rs[icell]);
	  above[b->id] = a;
	}
	else
	  bottom = a;

	if (a != 0) {
	  int icell = loops[e->loop[0]].cell;
	  if (loops[a->loop[1]].cell == icell)
	    updateInCircle(e, a, edges[e->info], x, ps[icell], rs[icell]);
	  below[a->id] = b;
	}

	if (a != 0 && b != 0)
	  b->info = e->id;
      }
    }
    
    // Insert outgoing edges for this x.
    for (Verti iiv = iv; iiv != jv; ++iiv) {
      Vert *v = *iiv;
      for (Edge *e = v->edge[0][1];
	   e != 0 && e->getVert(0) == v;
	   e = e->edge[0][0]) {
	Edge *b = e->edge[0][1];
	Edge *a = b == 0 ? bottom : above[b->id];

	if (a != 0 && b != 0) {
	  int icell = loops[b->loop[0]].cell;
	  if (loops[a->loop[1]].cell == icell)
	    updateInCircle(b, a, edges[b->info], x, ps[icell], rs[icell]);
	}

	if (b != 0) {
	  b->info = e->id;
	  above[b->id] = e;
	}
	else
	  bottom = e;

	if (a != 0) {
	  e->info = e->id;
	  below[a->id] = e;
	}

	below[e->id] = b;
	above[e->id] = a;
      }
    }
  }

  for (int i = 0; i < nedges; i++) {
    edges[i]->id = id[i];
    edges[i]->info = info[i];
  }

  Timing::end(Timing::pointInside);
}

Point Embedding::pointInside (int iCell) {
  int ncells = cells.size();
  Point ps[ncells];
  double rs[ncells];
  pointInside(ps, rs);
  if (iCell != -1)
    return ps[iCell];
  int iMax = -1;
  for (int i = 0; i < ncells; i++)
    if (cells[i].state == INSIDE &&
	(iMax == -1 || rs[i] > rs[iMax]))
      iMax = i;
  assert(iMax != -1);
  if (false && /* DEBUG */ !pointInside(ps[iMax])) {
    double r = rs[iMax];
    Point p = ps[iMax];
    cout << "pointinside " << p << endl;
    edgedraw("pointinside.fig");
    pointInside(ps, rs);
    pointInside(p);
  }
  // assert(pointInside(ps[iMax]));
  return ps[iMax];
}

bool Embedding::pointInside (Point p) {
  int iMax = -1;
  for (int i = 0; i < edges.size(); i++) {
    Edge *edge = edges[i];
    if (p.x < edge->tail()->p.x || p.x > edge->head()->p.x)
      continue;
    double y = edge->f->getY(p.x);
    if (y == p.y)
      return false;
    if (y < p.y && (iMax == -1 || edges[iMax]->height < edge->height))
      iMax = i;
  }

  return iMax != -1 && edges[iMax]->mult == 1;
}

double Embedding::DEdge::area () {
  Point t = tail()->p;
  Point h = head()->p;
  DOUBLE dArea = (t % h) / 2;
  
  if (e->f != 0 &&
      (e->f->getType() == Function::LOWER_CIRCLE ||
       e->f->getType() == Function::UPPER_CIRCLE)) {
    Circle *circle = (dynamic_cast<Cbranch *>((Function *) e->f))->circle;
    Point vtail = t - circle->c;
    Point vhead = h - circle->c;
    DOUBLE angle = vhead(vtail.inverse()).angle();
    if ((angle < 0)^ (vtail.x < vhead.x) ^
	(e->f->getType() == Function::LOWER_CIRCLE))
      angle = -angle;
    DOUBLE ddArea = circle->r * circle->r * angle / 2;
    DOUBLE dCross = (vtail % vhead) / 2;
    dArea += ddArea - dCross;
  }

  return dArea;
}

double Embedding::Loop::area () {
  double total = 0;
  DEdge curr = d;
  do {
    total += curr.area();
  } while (++curr != d);
  return total;
}

double Embedding::area () {
  double total = 0;
  for (int i = 0; i < edges.size(); i++) {
    Edge *edge = edges[i];
    DEdge dedge(edge, cells[loops[edge->loop[0]].cell].state == OUTSIDE);
    total += dedge.area();
  }
  for (int i = 0; i < edgesV.size(); i++) {
    Edge *edge = edgesV[i];
    DEdge dedge(edge, cells[loops[edge->loop[0]].cell].state == OUTSIDE);
    total += dedge.area();
  }
  return total;
}

#ifdef BLEEN
void Embedding::fillMSlits (Embedding *a, Embedding *b, PairIDSet *pairIDSet) {
  Embedding *emb[2] = { a, b };

  // Vector of outgoing and incoming edges.
  vector<Edge *> edgesTH[2];

  // Current active edge for a chain.
  map<int, Edge *> chains;
  vector<int> cset;

  // Related chains 
  vector<int> other[2 /* below / above */];

  buildLists(false);

  Verti jv;
  for (Verti iv = verts.begin(); iv != verts.end(); iv = jv) {
    // Determine range of vertices with same x.
    double x = (*iv)->p.x;
    for (jv = iv; jv != verts.end() && (*jv)->p.x == x; ++jv);
    
    // Get incoming and outgoing edges for this x.
    edgesTH[0].clear();
    edgesTH[1].clear();
    for (Verti iiv = iv; iiv != jv; ++iiv) {
      Vert *vert = *iiv;
      for (int th = 0; th < 2; th++) {
	for (Edge *edge = vert->edge[th][0]; edge != 0;
	     edge = edge->edge[th][0])
	  if (edge->getVert(th) == vert)
	    edgesTH[th].push_back(edge);
      }
    }

    // Remove incoming edges from active set
    for (int i = 0; i < edgesTH[1].size(); i++) {
      Edge *edge = edgesTH[1][i];
      pairIDSet->getSet(edge->info, cset);
      for (int j = 0; j < cset.size(); j++) {
	Edge *edge2 = chains[cset[j]];
	assert(chains[cset[j]] == edge);
	chains[cset[j]] = 0;
      }
    }

    for (int i = 0; i < edgesTH[0].size(); i++) {
      Edge *edge = edgesTH[0][i];
      pairIDSet->getSet(edge->info, cset);
      for (int j = 0; j < cset.size(); j++) {
	int cct = cset[j];
	assert(chains[cct] == 0);
	chains[cct] = edge;

	other[0].clear();
	other[1].clear();
	int chain[2] =  { (cct / 5) / b->chains.size(),
			  (cct / 5) % b->chains.size() };
	bool sw = a->chains[chain[0]]->sw;
	assert(sw == b->chains[chain[1]]->sw);
	int type = cct % 5;
	if (type != 0) {
	  // rule 2
	  other[sw].push_back(cct - type);
	  int ab = (type - 1) / 2;
	  int th = (type - 1) % 2;
	  int aNext = a->nextChainIndex(chain[0], th);
	  int bNext = b->nextChainIndex(chain[1], th);
	  int nn = (aNext * b->chains.size() + bNext) * 5;
	  // rule 3
	  other[sw ^ !emb[!ab]->chains[chain[!ab]]->endIsConcave[th]].push_back
	    (nn + 2 * ab + !th + 1);

	  if (a->chains[chain[0]]->endIsConcave[th] &&
	      b->chains[chain[1]]->endIsConcave[th] &&
	      chains[cct - type] == 0 ||
	      chains[nn] == 0)
	    // rule 4
	    other[sw].push_back(nn + 2 * !ab + !th + 1);
	}
	else {
	  for (int th = 0; th <= 1; th++) {
	    if (!a->chains[chain[0]]->endIsConcave[th] &&
		!b->chains[chain[1]]->endIsConcave[th]) {
	      int aNext = a->nextChainIndex(chain[0], th);
	      int bNext = b->nextChainIndex(chain[1], th);
	      // rule 1
	      other[!sw].push_back((aNext * b->chains.size() + bNext) * 5);
	    }
	    for (int ab = 0; ab <= 1; ab++)
	      if (emb[ab]->chains[chain[ab]]->endIsConcave[th])
		// rule 2
		other[!sw].push_back(cct + 2 * ab + th + 1);
	  }
	}


	Edge *edge2 = 0;
	int cct2;
	for (int du = 0; du <= 1 && edge2 == 0; du++)
	  for (int j = 0; j < other[du].size(); j++) {
	    cct2 = other[du][j];
	    Edge *e = chains[cct2];
	    if (e != 0 && e != edge && (edge->height < e->height) != du) {
	      edge2 = e;
	      break;
	    }
	  }
	if (edge2 == 0)
	  continue;

	if (edge->tail() != edge2->tail() || edge->head() != edge2->head())
	  cout << "true slit!!" << endl;

	if (edge2->tail() != edge->tail()) {
	  if (edge2->tail()->p.x == edge->tail()->p.x)
	    addEdge(edge2->tail()->p, edge->tail()->p, 0, 0, edge2->mult);
	  else
	    addEdge(edge2->tail()->p, edge->tail()->p, edge2->f, edge2->id,
		    edge2->mult, edge2->info)->height = edge2->height;
	  edge2->setTail(edge->tail());
	}
	
	if (edge2->head() != edge->head()) {
	  Edge *e1 = edge->head()->p.x <= edge2->head()->p.x ? edge : edge2;
	  Edge *e2 = edge->head()->p.x <= edge2->head()->p.x ? edge2 : edge;
	  
	  // Remove e2 from head list.
	  Edge **e = &e2->head()->edge[1][0];
	  while (*e != e2)
	    e = &(*e)->edge[1][0];
	  *e = (*e)->edge[1][0];

	  if (e1->head()->p.x == e2->head()->p.x)
	    addEdge(e1->head()->p, e2->head()->p, 0, 0, e2->mult);
	  else {
	    Edge *e3 = addEdge(e1->head()->p, e2->head()->p, e2->f,
			       e2->id, e2->mult, e2->info);
	    e3->height = e2->height;

	    // Insert e3 into tail and head lists.
	    for (int th = 0; th <= 1; th++) {
	      e3->edge[th][0] = e3->getVert(th)->edge[th][0];
	      e3->getVert(th)->edge[th][0] = e3;
	    }
	  }

	  // Insert e2 into new head list.
	  e2->setHead(e1->head());
	  e2->edge[1][0] = e2->head()->edge[1][0];
	  e2->head()->edge[1][0] = e2;
	}
	
	edge->info = pairIDSet->op(edge->info, cct);
	chains[cct] = 0;
	edge2->info = pairIDSet->op(edge2->info, cct2);
	chains[cct2] = 0;
	int mult = 2 * ((type == 0) == sw) - 1;
	edge->mult += mult;
	edge2->mult -= mult;

	cout << "filled slit " << edge->tail()->p << " "
	     << edge->head()->p.x - edge->tail()->p.x << endl;
      }
    }
  }

  int nedges = 0;
  for (int i = 0; i < edges.size(); i++)
    if (edges[i]->mult == 0)
      delete edges[i];
    else
      edges[nedges++] = edges[i];
  edges.resize(nedges);

  arrange1d(edgesV);
  embed();
  build();
}
#endif

void Embedding::getMOrder (map<int, Edge *> &chains, int cct, double x,
			   vector<int> *other) {
  other[0].clear();
  other[1].clear();
  int chain[2] =  { (cct / 5) / emb[1]->chainsM.size(),
		    (cct / 5) % emb[1]->chainsM.size() };
  bool sw = emb[0]->chainsM[chain[0]]->sw;
  assert(sw == emb[1]->chainsM[chain[1]]->sw);
  int type = cct % 5;
  if (type != 0) {
    // rule 2
    other[sw].push_back(cct - type);
    int ab = (type - 1) / 2;
    int th = (type - 1) % 2;
    // Call this p1 + U2.  p1 must be concave.
    assert(emb[ab]->chainsM[chain[ab]]->endIsConcave[th]);
    int U1p2 = cct - type + 2 * !ab + th + 1;
    int aNext = emb[0]->nextChainIndexM(chain[0], th);
    int bNext = emb[1]->nextChainIndexM(chain[1], th);
    int nn = (aNext * emb[1]->chainsM.size() + bNext) * 5;
    int p1L2 = nn + 2 * ab + !th + 1;

    bool p2isConcave = emb[!ab]->chainsM[chain[!ab]]->endIsConcave[th];

    // rule 3
    other[sw ^ !p2isConcave].push_back(nn + 2 * ab + !th + 1);
    
    Chain *u1 = emb[ab]->chainsM[chain[ab]];
    Chain *l2 = emb[!ab]->chainsM[!ab == 0 ? aNext : bNext];
    Point p1 = emb[ab]->chainsM[chain[ab]]->end(th);
    Point p2 = emb[!ab]->chainsM[chain[!ab]]->end(th);

    if (p2isConcave &&
	(u1->minX() + p2.x <= x && x <= u1->maxX() + p2.x &&
	 chains[U1p2] == 0 ||
	(l2->minX() + p1.x <= x && x <= l2->maxX() + p1.x &&
	 chains[p1L2] == 0)))
      // rule 4
      other[sw].push_back(nn + 2 * !ab + !th + 1);

    if (!p2isConcave && p1.x + l2->minX() <= x && x <= p1.x + l2->maxX())
      // p1+U2 > p1+L2 > L1+L2
      // rule 32
      other[!sw].push_back(nn);
  }
  else {
    for (int th = 0; th <= 1; th++) {
      int aNext = emb[0]->nextChainIndexM(chain[0], th);
      int bNext = emb[1]->nextChainIndexM(chain[1], th);
      int nn = (aNext * emb[1]->chainsM.size() + bNext) * 5;

      if (!emb[0]->chainsM[chain[0]]->endIsConcave[th] &&
	  !emb[1]->chainsM[chain[1]]->endIsConcave[th])
	// rule 1
	other[!sw].push_back(nn);

      for (int ab = 0; ab <= 1; ab++)
	if (emb[ab]->chainsM[chain[ab]]->endIsConcave[th]) {
	  // U1 + U2 >= p1 + U2
	  // rule 2
	  other[!sw].push_back(cct + 2 * ab + th + 1);

	  if (!emb[!ab]->chainsM[chain[!ab]]->endIsConcave[th]) {
	    Point p1 = emb[ab]->chainsM[chain[ab]]->end(th);
	    Chain *u2 = emb[!ab]->chainsM[chain[!ab]];
	    if (p1.x + u2->minX() <= x && x <= p1.x + u2->maxX())
	      // p2 is convex
	      // U1 + U2 >= p1 + U2 > p1 + L2
	      // rule 23
	      other[!sw].push_back(nn + 2 * ab + !th + 1);
	  }
	}
    }
  }
}  

void Embedding::arrangeM (vector<Event *> &events) {
  if (verts.size() == 0)
    return;

  // arrange1d(edges);

  // create telomeres
  for (Verti it = verts.begin(); it != verts.end(); it++)
    (*it)->edge[1][0] = 0;

  set<Vert *, Vert::CompareYX> tails;
  for (int i = 0; i < edges.size(); i++)
    tails.insert(edges[i]->tail());

  vector<Edge *> telos;
  double minX = (*verts.begin())->p.x;
  double maxX = (*verts.rbegin())->p.x;
  Vert *prev = 0;

  for (set<Vert *, Vert::CompareYX>::iterator it = tails.begin();
       it != tails.end(); ++it) {
    Vert *head = *it;
    double x = head->p.x - TELOMERE_LENGTH, y = head->p.y;
    Vert *tail = 0;
    if (prev != 0 && prev->p.y == y && prev->p.x >= x)
      tail = prev;
    else
      tail = addVert(Point(x, y));
    prev = head;
    Function *line = 0;
    line = Line::alloc(Point(minX - TELOMERE_LENGTH, y), Point(maxX, y));
    Edge *telo = new Edge(tail, head, line, -1, 0);
    telos.push_back(telo);
    head->edge[1][0] = telo;
  }

  Treap *sweep = 0;
  Queue queue;

  for (int i = 0; i < edges.size(); i++) {
    Edge *edge = edges[i];
    assert(edge->tail()->edge[1][0] != 0);
    queue.push(new Event(Event::CONTINUE, edge->tail()->p, edge, 0, 0));
    queue.push(new Event(Event::REMOVE, edge->head()->p, edge, 0, 0));
  }

  // Current active edge for a chain.
  map<int, Edge *> chains;
  vector<int> cset;

  for (int i = 0; i < telos.size(); i++) {
    Edge *edge = telos[i];
    if (edge->tail()->edge[1][0] == 0)
      queue.push(new Event(Event::INSERT, edge->tail()->p, edge, 0, 0));
    else
      queue.push(new Event(Event::CONTINUE, edge->tail()->p, edge, 0, 0));
    queue.push(new Event(Event::TREMOVE, edge->head()->p, edge, 0, 0));
  }

  vector<int> other[2];
  
  double lastCONTINUEx = -Function::FINFINITY;
  double lastREMOVEx = -Function::FINFINITY;
  vector<Event *> saveEvents;

  while (!queue.empty()) {
    Event *event = queue.top();
    queue.pop();

    Edge *a = event->a;
    Edge *b = event->b;
    Edge *c = event->c;
    Point p = event->p;
    Event::Type type = event->type;

    if (type == Event::INSERT) {
      assert(c->id == -1);
      SweepNode *node = new SweepNode(c);
      if (sweep == 0)
	sweep = node;
      else {
	sweep->insert(node);
	sweep = node->rebalance();
	Edge *next = node->nextEdge();
	Edge *prev = node->prevEdge();
	if (next != 0)
	  queue.push(new Event(Event::CHECK, p, 0, c, next));
	if (prev != 0)
	  queue.push(new Event(Event::CHECK, p, 0, prev, c));

#ifdef BLEEN
	if (isShallow(c->f, p.x)) {
	  while (prev != 0 && isShallow(prev->f, p.x))
	    prev = prev->node->prevEdge();
	  assert(prev == 0 || prev->tail()->p.y < c->tail()->p.y + 1e-6);
	  while (next != 0 && isShallow(next->f, p.x))
	    next = next->node->nextEdge();
	  assert(next == 0 || next->tail()->p.y > c->tail()->p.y - 1e-6);
	}
#endif
      }

      delete event;
    }
    else if (type == Event::CONTINUE) {
      // if (c->ssn == 1149053)
      // cout << "this is it" << endl;

      if (c->id != -1 && lastCONTINUEx < c->tail()->p.x) {
	assert(chains[c->id] == 0);
	chains[c->id] = c;
	lastCONTINUEx = c->tail()->p.x;
	saveEvents.clear();
	while (!queue.empty() && queue.top()->p.x == lastCONTINUEx) {
	  saveEvents.push_back(queue.top());
	  queue.pop();
	}
	for (int i = 0; i < saveEvents.size(); i++) {
	  Event *event2 = saveEvents[i];
	  queue.push(event2);
	  if (event2->type == Event::CONTINUE && event2->c->id != -1) {
	    assert(chains[event2->c->id] == 0);
	    chains[event2->c->id] = event2->c;
	  }
	}
      }

      Edge *prevID = 0, *nextID = 0;
      if (c->id != -1) {
	int cct = c->id;
	Timing::begin(Timing::adapt);
	getMOrder(chains, cct, p.x, other);
	Timing::end(Timing::adapt);

	for (int j = 0; j < other[0].size(); j++) {
	  int cct2 = other[0][j];
	  Edge *e = chains[cct2];
	  assert(e != c);
	  if (e != 0 && e->node != 0 &&
	      (prevID == 0 || prevID->node->compareTree(e->node) <= 0))
	    prevID = e;
	}

	for (int j = 0; j < other[1].size(); j++) {
	  int cct2 = other[1][j];
	  Edge *e = chains[cct2];
	  assert(e != c);
	  if (e != 0 && e->node != 0 &&
	      (nextID == 0 || nextID->node->compareTree(e->node) >= 0))
	    nextID = e;
	}
      }

      if (prevID != 0 && nextID != 0)
	assert(prevID->node->compareTree(nextID->node) <= 0);

      SweepNode *node = new SweepNode(c);
      Edge *next = 0, *prev = 0;
      if (sweep == 0)
	sweep = node;
      else {
	Edge *telo = c->tail()->edge[1][0];
	next = telo->edge[1][0];
	prev = telo->edge[1][1];

	// next < prevID
	if (next != 0 && prevID != 0 &&
	    next->node->compareTree(prevID->node) <= 0) {
	  prev = prevID;
	  next = prevID->node->nextEdge();
	}
	// prev > nextID
	else if (prev != 0 && nextID != 0 &&
		 prev->node->compareTree(nextID->node) >= 0) {
	  next = nextID;
	  prev = nextID->node->prevEdge();
	}
	else {
	  // prevID > prev
	  if (prevID != 0 &&
	      (prev == 0 || prevID->node->compareTree(prev->node) > 0))
	    prev = prevID;
	  // nextID < next
	  if (nextID != 0 &&
	      (next == 0 || nextID->node->compareTree(next->node) < 0))
	    next = nextID;
	}

	if (next != 0)
	  next->node->setMax();
	if (prev != 0)
	  prev->node->setMin();

	sweep->insert(node);

	if (next != 0)
	  next->node->clearMinMax();
	if (prev != 0)
	  prev->node->clearMinMax();

	sweep = node->rebalance();

	next = node->nextEdge();
	prev = node->prevEdge();
	if (next != 0)
	  queue.push(new Event(Event::CHECK, p, 0, c, next));
	if (prev != 0)
	  queue.push(new Event(Event::CHECK, p, 0, prev, c));

#ifdef BLEEN
	Edge *savePrev = prev, *saveNext = next;
	if (isShallow(c->f, p.x)) {
	  while (prev != 0 && !isShallow(prev->f, p.x))
	    prev = prev->node->prevEdge();
	  assert(prev == 0 || prev->tail()->p.y < c->tail()->p.y + 1e-6);
	  while (next != 0 && !isShallow(next->f, p.x))
	    next = next->node->nextEdge();
	  assert(next == 0 || next->tail()->p.y > c->tail()->p.y - 1e-6);
	}
	prev = savePrev;
	next = saveNext;
#endif
      }

      if (c->id != -1) {
	while (prev != 0 && prev->id == -1)
	  prev = prev->node->prevEdge();
	while (next != 0 && next->id == -1)
	  next = next->node->nextEdge();
	event->a = prev;
	event->b = next;
	event->type = Event::INSERT;
	events.push_back(event);
      }
      else
	delete event;
    }
    else if (type == Event::REMOVE || type == Event::TREMOVE) {
      if (c->id != -1 && lastREMOVEx < c->head()->p.x) {
	assert(chains[c->id] == c);
	chains[c->id] = 0;
	lastREMOVEx = c->head()->p.x;
	saveEvents.clear();
	while (!queue.empty() && queue.top()->p.x == lastREMOVEx) {
	  saveEvents.push_back(queue.top());
	  queue.pop();
	}
	for (int i = 0; i < saveEvents.size(); i++) {
	  Event *event2 = saveEvents[i];
	  queue.push(event2);
	  if (event2->type == Event::REMOVE && event2->c->id != -1) {
	    assert(chains[event2->c->id] == event2->c);
	    chains[event2->c->id] = 0;
	  }
	}
      }

      SweepNode *node = c->node;
      Edge *prev = node->prevEdge();
      Edge *next =  node->nextEdge();
      if (prev != 0 && next != 0) {
#ifdef BLEEN
	// debug
	double prevY = prev->f->getY(p.x);
	double nextY = next->f->getY(p.x);
	double cY = c->f->getY(p.x);
#endif
	queue.push(new Event(Event::CHECK, p, 0, prev, next));
      }
#ifdef BLEEN
      while (prev != 0 && prev->id == -1)
	prev = prev->node->prevEdge();
      while (next != 0 && next->id == -1)
	next = next->node->nextEdge();
#endif
      c->edge[1][0] = next;
      c->edge[1][1] = prev;
      sweep = node->remove();
      delete node;

      if (c->id != -1) {
	event->a = prev;
	event->b = next;
	events.push_back(event);
      }
      else
	delete event;
    }
    else if (type == Event::SWAP) {
      if (a->node->nextEdge() != b) {
	delete event;
	continue;
      }

      if (a->id != -1 && b->id != -1) {
	Timing::begin(Timing::adapt);
	getMOrder(chains, a->id, p.x, other);
	Timing::end(Timing::adapt);
	bool skipIt = false;
	for (int i = 0; i < other[1].size(); i++)
	  if (other[1][i] == b->id)
	    skipIt = true;
	if (skipIt) {
	  delete event;
	  continue;
	}
      }

      if (p.x != a->f->nearestRoot(b->f, p.x))
	p.y = a->bestY(b, p.x);
      event->p = p;
      SweepNode *nodeA = a->node;
      SweepNode *nodeB = b->node;
      a->node = nodeB;
      b->node = nodeA;
      nodeA->edge = b;
      nodeB->edge = a;
      Edge *prev = nodeA->prevEdge();
      Edge *next = nodeB->nextEdge();
      queue.push(new Event(Event::CHECK, p, 0, b, a));
      if (prev != 0)
	queue.push(new Event(Event::CHECK, p, 0, prev, b));
      if (next != 0)
	queue.push(new Event(Event::CHECK, p, 0, a, next));

      if (a->id != -1 && b->id != -1)
	events.push_back(event);
      else
	delete event;
    }
    else if (type == Event::CHECK) {
      // if (a->ssn == 110493 && b->ssn == 110394)
      // cout << "this is it" << endl;

      if (a->node == 0 || a->node->nextEdge() != b) {
	delete event;
	continue;
      }

#ifdef BLEEN
      // debug
      double maxX = max(a->tail()->p.x, b->tail()->p.x);
      double ay = a->f->getY(maxX);
      double by = b->f->getY(maxX);
#endif

      p = a->nextRoot(b, p.x);
      if (p.x != Function::FINFINITY)
	queue.push(new Event(Event::SWAP, p, 0, a, b));
      
      delete event;
    }
  }

  for (int i = 0; i < telos.size(); i++)
    delete telos[i];
}

double Embedding::Vert::height () {
  return edge[0][0] != 0 ? edge[0][0]->height : edge[1][0]->height;
}

Embedding::Edge *Embedding::Vert::edgeAbove (int hilo) {
  if (edge[2][hilo] != 0)
    return 0;

  Edge *e = edge[0][hilo] != 0 ?
    edge[0][hilo]->edge[0][hilo] : edge[1][hilo]->edge[1][hilo];

  if (e == 0)
    return 0;

  if (e->tail()->p.x < p.x && p.x < e->head()->p.x)
    return e;

  return 0;
}

void Embedding::splitAtNearVerts (double epsilon) {
  static int count;
  count++;
  // if (count == 83)
  // cout << "this is it";

  Vert *prev = 0;

  for (Verti iv = verts.begin(); iv != verts.end();) {
    Vert *v = *iv++;
    
    // if (11.17 < v->p.x && v->p.x < 11.18 &&
    // -4.22 > v->p.y && v->p.y > -4.23)
    // cout << "this is it" << endl;

    if (iv != verts.end() && v->edge[2][0] == 0) {
      Vert *w = *iv;
      if (v->p.x == w->p.x) {
	assert(w->edge[2][1] == 0);
	Edge *ev = v->edgeAbove(0);
	if (ev == 0 || ev->height > w->height()) {
	  // if (prev != 0 && v->p.x == prev->p.x)
	  // cout << "prev is it" << endl;
	  Edge *e = addEdge(v->p, w->p, 0, 0, 0, 0);
	  v->edge[2][0] = e;
	  w->edge[2][1] = e;
	}
      }
    }

    for (int hilo = 0; hilo < 2; hilo++) {
      if (v->edge[2][hilo] != 0)
	continue;
      Edge *e = v->edgeAbove(hilo);
      if (e == 0)
	continue;
      double dist = e->f->distance(v->p);
      if (dist >= epsilon)
	continue;
      
      double y = e->f->getY(v->p.x);
      Vert *w = 0;
      if (hilo == 0 && iv != verts.end() && (*iv)->p.x == v->p.x)
	w = *iv;
      if (hilo == 1) {
	Verti jv = iv;
	jv--;
	if (jv != verts.begin()) {
	  jv--;
	  if ((*jv)->p.x == v->p.x)
	    w = *jv;
	}
      }
      if (hilo == 0 && v->p.y < y && (w == 0 || y < w->p.y) ||
	  hilo == 1 && v->p.y > y && (w == 0 || y > w->p.y)) {
	// if (prev != 0 && v->p.x == prev->p.x)
	// cout << "prev is it" << endl;
	Point q(v->p.x, y);
	Edge *l = addEdge(e->tail()->p, q, e->f, e->id, e->mult, e->info);
	l->height = e->height;
	e->setTail(l->head());
	prev = l->head();
	Edge *u = addEdge(v->p, q, 0, 0, 0, 0);
	v->edge[2][hilo] = u;
	e->tail()->edge[2][!hilo] = u;
      }
      else {
	// if (prev != 0 && v->p.x == prev->p.x)
	// cout << "prev is it" << endl;
	addEdge(e->tail()->p, v->p, e->f, e->id, e->mult, e->info)->height
	  = e->height;
	e->setTail(v);
      }
    }
  }

  build(true);
}
