#include "Treap.h"
#include <assert.h>
#include <stdlib.h>

Treap::Treap () {
  parent = left = right = 0;
  goRight = goLeft = false;
  priority = random();
}

void Treap::pushDownRoot () {
  if (parent != 0)
    parent->pushDownRoot();
  pushDown();
}    

void Treap::pullUpRoot () {
  for (Treap *node = this; node != 0; node = node->parent)
    node->pullUp();
}

void Treap::setMin () {
  for (Treap *up = this, *dn = right; up != 0; dn = up, up = up->parent)
    if (dn == up->right)
      up->goRight = true;
}

void Treap::setMax () {
  for (Treap *up = this, *dn = left; up != 0; dn = up, up = up->parent)
    if (dn == up->left)
      up->goLeft = true;
}

void Treap::clearMinMax () {
  for (Treap *node = this; node != 0; node = node->parent)
    node->goRight = node->goLeft = false;
}

void Treap::insert (Treap *node) {
  int cmp = goRight ? -1 : goLeft ? 1 : compare(node);
  if (cmp == 0)
    cmp = compare(node);
  assert(cmp != 0);

  if (cmp > 0) {
    if (left == 0) {
      left = node;
      node->parent = this;
      node->left = node->right = 0;
    }
    else
      left->insert(node);
  }
  else {
    if (right == 0) {
      right = node;
      node->parent = this;
      node->left = node->right = 0;
    }
    else
      right->insert(node);
  }
}

void Treap::locate (Treap *node, Treap *&prev, Treap *&next) {
  prev = next = 0;
  if (this == 0)
    return;
  Treap *root = this;
  while (root != 0) {
    int cmp = root->goRight ? -1 : root->goLeft ? 1 : root->compare(node);
    assert(cmp != 0);
    if (cmp > 0) {
      next = root;
      root = root->left;
    }
    else {
      prev = root;
      root = root->right;
    }
  }
}

Treap *Treap::rebalance () {
  pushDownRoot();
  pullUpRoot();

  while (parent != 0 && priority < parent->priority)
    if (this == parent->left)
      rotateRight();
    else
      rotateLeft();

  for (Treap *node = this;; node = node->parent)
    if (node->parent == 0)
      return node;
}

void Treap::rotateRight () {
  assert(parent->left == this);
  parent->left = right;
  if (right != 0)
    right->parent = parent;
  right = parent;
  parent = right->parent;
  right->parent = this;
  if (parent != 0)
    if (parent->left == right)
      parent->left = this;
    else {
      assert(parent->right == right);
      parent->right = this;
    }
  right->pullUp();
  pullUp();
}

void Treap::rotateLeft () {
  assert(parent->right == this);
  parent->right = left;
  if (left != 0)
    left->parent = parent;
  left = parent;
  parent = left->parent;
  left->parent = this;
  if (parent != 0)
    if (parent->right == left)
      parent->right = this;
    else {
      assert(parent->left == left);
      parent->left = this;
    }
  left->pullUp();
  pullUp();
}

Treap *Treap::remove () {
  pushDownRoot();

  while (left != 0 && right != 0)
    if (left->priority < right->priority) {
      left->pushDown();
      left->rotateRight();
    }
    else {
      right->pushDown();
      right->rotateLeft();
    }

  Treap *node = left != 0 ? left : right;
  if (node != 0)
    node->parent = parent;
  if (parent != 0)
    if (parent->left == this)
      parent->left = node;
    else {
      assert(parent->right == this);
      parent->right = node;
    }

  if (node == 0)
    node = parent;

  left = right = parent = 0;

  for (; node != 0; node = node->parent) {
    node->pullUp();
    if (node->parent == 0)
      return node;
  }

  return 0;
}

// Root of tree.
Treap *Treap::root () {
  Treap *node = this;
  while (node->parent != 0)
    node = node->parent;
  return node;
}

// First node in tree.
Treap *Treap::first () {
  Treap *node = root();
  while (node->left != 0)
    node = node->left;
  return node;
}

// Last node in tree.
Treap *Treap::last () {
  Treap *node = root();
  while (node->right != 0)
    node = node->right;
  return node;
}

// Next node in tree.
Treap *Treap::next () {
  if (right != 0) {
    Treap *node = right;
    while (node->left != 0)
      node = node->left;
    return node;
  }

  for (Treap *node = this; node->parent != 0; node = node->parent)
    if (node == node->parent->left)
      return node->parent;

  return 0;
}

// Previous node in tree.
Treap *Treap::prev () {
  if (left != 0) {
    Treap *node = left;
    while (node->right != 0)
      node = node->right;
    return node;
  }

  for (Treap *node = this; node->parent != 0; node = node->parent)
    if (node == node->parent->right)
      return node->parent;

  return 0;
}

// -1/0/1 if this is left/equal/right of that in tree.
// Assumes no setMin or setMax active.
int Treap::compareTree (Treap *that) {
  if (this == that)
    return 0;
  setMin();
  int ret = -1;
  for (Treap *up = that, *dn = up->left; up != 0; dn = up, up = up->parent)
    if (dn == up->left && up->goRight)
      ret = 1;
  clearMinMax();
  return ret;
}

