#include "poly.h"
#include "curve.h"

int Abranch::BranchId;

// planar algebraic curves

Acurve::Acurve (Point &tail, Point &head, double *ibbox)
{
  double a = tail.x - head.x, b = head.y - tail.y, c = - a*tail.x - b*tail.y;
  double x[3] = {c, a, b};
  f = forig = linear_polynomial(3, x);
  fx = f.der(1);
  fy = f.der(2);
  for (int i = 0; i < 4; ++i)
    bbox[i] = ibbox[i];
}

Acurve::Acurve (const Polynomial &fi, double *ibbox)
  : f(fi), forig(fi)
{
  f.normalize_coeffs();
  f.val_normalize();
  fx = f.der(1);
  fx.normalize_coeffs();
  fx.val_normalize();
  fy = f.der(2);
  fy.normalize_coeffs();
  fy.val_normalize();
  for (int i = 0; i < 4; ++i)
    bbox[i] = ibbox[i];
}

Acurve::~Acurve ()
{
  for (int i = 0; i < bifs.size(); ++i)
    delete bifs[i];
  for (int i = 0; i < branches.size(); ++i)
    delete branches[i];
}

bool Acurve::form_branches (bool ymonotone)
{
  bifurcations();
  Abranches sweep;
  int i = 0, j, n = bifs.size();
  while (i < n && bifs[i]->type == Left)
    update_left(bifs[i++], sweep);
  int nb = sweep.size();
  while (i < n && bifs[i]->type != Right) {
    for (j = i; j < n && bifs[j]->point.x == bifs[i]->point.x; ++j)
      switch (bifs[j]->type) {
      case Bottom:
	nb = nb + (update_bottom(bifs[j], sweep) ? 1 : -1);
	break;
      case Top:
	nb = nb + (update_top(bifs[j], sweep) ? 1 : -1);
	break;
      case Split:
	nb += 2;
	update_split(bifs[j], sweep);
	break;
      case Merge:
	nb -= 2;
	update_merge(bifs[j], sweep);
	break;
      case Min: case Max:
	update_minmax(bifs[j], sweep, ymonotone);
	break;
      case Singular:
	update_singular(bifs[j], sweep);
	break;
      }
    double xa, r[f.degree];
    xa = 0.5*(bifs[i]->point.x + (j < n ? bifs[j]->point.x : bbox[1]));
    if (nb != yvals(xa, r))
      return false;
    i = j;
  }
  if (nb != n - i) return false;
  while (i < n)
    update_right(bifs[i++], sweep);
  formZones();
  return true;
}

void Acurve::bifurcations ()
{
  double ibbox[4], dmin = 1e-12;
  for (int i = 0; i < 4; ++i)
    ibbox[i] = bbox[i];
  bbox[0] -= 1.0;
  bbox[1] += 1.0;
  bbox[2] -= 1.0;
  bbox[3] += 1.0;
  x_turn_points();
  y_turn_points();
  adjust_bbox(ibbox);
  bbox_points();
  sort(bifs.begin(), bifs.end(), bif_order);
  for (int i = 1; i < bifs.size(); ++i)
    if (bifs[i]->type == bifs[i-1]->type && 
	fabs(bifs[i]->point.x - bifs[i-1]->point.x) < dmin &&
	fabs(bifs[i]->point.y - bifs[i-1]->point.y) < dmin) {
      delete bifs[i];
      for (int j = i; j < bifs.size(); ++j)
	bifs[j] = bifs[j+1];
      bifs.pop_back();
    }
}

void Acurve::x_turn_points ()
{
  Polynomial fyy = fy.der(2);
  Psystem psys(f, fy);
  double *tp = new double[2*psys.degree()];
  int n = psys.realroots(tp);
  for (int i = 0; i < n; ++i) {
    double *p = tp + 2*i;
    if (in_bbox(p[0], p[1], bbox)) {
      double vx = fx.val(p), vyy = fyy.val(p);
      Biftype btype;
      if (fabs(vx) < SINGULAR)
	btype = Singular;
      else
	btype = vx*vyy > 0.0 ? Merge : Split;
      Bifurcation *b = new Bifurcation(btype, p[0], p[1]);
      bifs.push_back(b);
    }
  }
  delete [] tp;
}

void Acurve::y_turn_points ()
{
  Polynomial fxx = fx.der(1);
  Psystem psys(f, fx);
  double *tp = new double [2*psys.degree()];
  int n = psys.realroots(tp);
  for (int i = 0; i < n; ++i) {
    double *p = tp + 2*i;
    if (in_bbox(p[0], p[1], bbox)) {
      double vy = fy.val(p), vxx = fxx.val(p);
      if (fabs(vy) < SINGULAR)
	continue;
      Biftype btype = vy*vxx > 0.0 ? Max : Min;
      Bifurcation *b = new Bifurcation(btype, p[0], p[1]);
      bifs.push_back(b);
    }
  }
  delete [] tp;
}

void Acurve::adjust_bbox (double *ibbox)
{
  for (int i = 0; i < bifs.size(); ++i) {
    const Point &p = bifs[i]->point;
    ibbox[0] = min(ibbox[0], p.x);
    ibbox[1] = max(ibbox[1], p.x);
    ibbox[2] = min(ibbox[2], p.y);
    ibbox[3] = max(ibbox[3], p.y);
  }
  for (int i = 0; i < 4; ++i)
    bbox[i] = 0.5*(bbox[i] + ibbox[i]);
}

void Acurve::bbox_points ()
{
  bbox_points_lr(Left);
  bbox_points_lr(Right);
  bbox_points_bt(Bottom);
  bbox_points_bt(Top);
}

void Acurve::bbox_points_lr (Biftype btype)
{
  double x, roots[2*f.degree];
  x = btype == Left ? bbox[0] : bbox[1];
  Polynomial g = f.reduce2(1, x);
  Psystem psys(g);
  int n = psys.realroots(roots);
  for (int i = 0; i < n; ++i)
    if (roots[i] > bbox[2] - Transformation::TINY &&
	roots[i] < bbox[3] + Transformation::TINY) {
      Bifurcation *b = new Bifurcation(btype, x, roots[i]);
      bifs.push_back(b);
    }
}

void Acurve::bbox_points_bt (Biftype btype)
{
  double y, roots[2*f.degree];
  y = btype == Bottom ? bbox[2] : bbox[3];
  Polynomial g = f.reduce2(2, y);
  Psystem psys(g);
  int n = psys.realroots(roots);
  for (int i = 0; i < n; ++i)
    if (roots[i] > bbox[0] + Transformation::TINY &&
	roots[i] < bbox[1] - Transformation::TINY) {
      Bifurcation *b = new Bifurcation(btype, roots[i], y);
      bifs.push_back(b);
    }
}

bool bif_order (Bifurcation *a, Bifurcation *b)
{
  return (a->point.x < b->point.x ||
	  a->point.x == b->point.x && a->point.y < b->point.y);
}

void Acurve::update_left (Bifurcation *bif, Abranches &sweep)
{
  double h = (double) sweep.size();
  Abranch *b = new Abranch(bif, h, this);
  bif->branches.push_back(b);
  branches.push_back(b);
  sweep.push_back(b);
}

void Acurve::update_right (Bifurcation *bif, Abranches &sweep)
{
  Abranch *b = sweep[0];
  b->set_head(bif);
  bif->branches.push_back(b);
  Abranches::iterator iter = remove(sweep.begin(), sweep.end(), b);
  sweep.erase(iter, sweep.end());
}

bool Acurve::update_bottom (Bifurcation *bif, Abranches &sweep)
{
  Point v = der(bif->point);
  if (v.y > 0.0) {
    double h = sweep.empty() ? 0.0 : sweep[0]->h - 1.0;
    Abranch *b = new Abranch(bif, h, this);
    bif->branches.push_back(b);
    branches.push_back(b);
    sweep.insert(sweep.begin(), b);
    return true;
  }
  if (sweep.empty())
    return false;
  Abranch *b = sweep[0];
  b->set_head(bif);
  bif->branches.push_back(b);
  Abranches::iterator iter = remove(sweep.begin(), sweep.end(), b);
  sweep.erase(iter, sweep.end());
  return false;
}

bool Acurve::update_top (Bifurcation *bif, Abranches &sweep)
{
  Point v = der(bif->point);
  int n = sweep.size() - 1;
  if (v.y < 0.0) {
    double h = sweep.empty() ? 0.0 : sweep[n]->h + 1.0;
    Abranch *b = new Abranch(bif, h, this);
    bif->branches.push_back(b);
    branches.push_back(b);
    sweep.push_back(b);
    return true;
  }
  if (sweep.empty())
    return false;
  Abranch *b = sweep[n];
  b->set_head(bif);
  bif->branches.push_back(b);
  Abranches::iterator iter = remove(sweep.begin(), sweep.end(), b);
  sweep.erase(iter, sweep.end());
  return false;
}

void Acurve::update_split (Bifurcation *bif, Abranches &sweep)
{
  int i, n = sweep.size();
  double h1, h2;
  Point &p = bif->point;
  if (n == 0) {
    i = 0;
    h1 = 0.0;
    h2 = 1.0;
  }
  else {
    double y[f.degree];
    yvals(bif, y);
    i = 0;
    while (i < n && y[i] < p.y)
      ++i;
    if (i == 0) {
      h1 = sweep[0]->h - 2.0;
      h2 = sweep[0]->h - 1.0;
    }
    else if (i == n) {
      h1 = sweep[n-1]->h + 1.0;
      h2 = sweep[n-1]->h + 2.0;
    }
    else {
      h1 = 2.0*sweep[i-1]->h/3.0 + sweep[i]->h/3.0;
      h2 = sweep[i-1]->h/3.0 + 2.0*sweep[i]->h/3.0;
    }
  }
  Abranch *b1 = new Abranch(bif, h1, this);
  Abranch *b2 = new Abranch(bif, h2, this);
  bif->branches.push_back(b1);
  bif->branches.push_back(b2);
  branches.push_back(b1);
  branches.push_back(b2);
  sweep.insert(sweep.begin() + i, b2);
  sweep.insert(sweep.begin() + i, b1);
}

void Acurve::update_merge (Bifurcation *bif, Abranches &sweep)
{
  int n = find_branch(bif, sweep);
  if (n > sweep.size() - 2)
    return;
  Abranch *b1 = sweep[n], *b2 = sweep[n+1];
  b1->set_head(bif);
  b2->set_head(bif);
  bif->branches.push_back(b1);
  bif->branches.push_back(b2);
  Abranches::iterator iter = remove(sweep.begin(), sweep.end(), b1);
  sweep.erase(iter, sweep.end());
  iter = remove(sweep.begin(), sweep.end(), b2);
  sweep.erase(iter, sweep.end());
}

void Acurve::update_minmax (Bifurcation *bif, Abranches &sweep, bool ymonotone)
{
  int n = find_branch(bif, sweep);
  if (n > sweep.size() - 1)
    return;
  Abranch *b = sweep[n];
  if (bif->point.y < b->ymin)
    b->ymin = bif->point.y;
  else if (bif->point.y > b->ymax)
    b->ymax = bif->point.y;
  if (!ymonotone) 
    return;
  b->set_head(bif);
  bif->branches.push_back(b);
  Abranch *c = new Abranch(bif, b->h, this);
  bif->branches.push_back(c);
  branches.push_back(c);
  sweep[n] = c;
}

void Acurve::update_singular (Bifurcation *bif, Abranches &sweep)
{
  int n = find_branch(bif, sweep); 
  if (n > sweep.size() - 2)
    return;
  for (int i = 0; i < 2; ++i) {
    sweep[n+i]->set_head(bif);
    bif->branches.push_back(sweep[n+i]);
  }
  for (int i = 0; i < 2; ++i) {
    Abranch *b = new Abranch(bif, sweep[n+i]->h, this);
    bif->branches.push_back(b);
    branches.push_back(b);
    sweep[n+i] = b;
  }
}

int Acurve::find_branch (Bifurcation *bif, Abranches &sweep)
{
  double y[f.degree];
  int i = 0, n = yvals(bif, y);
  while (i < n && y[i] < bif->point.y)
    ++i;
  return i;
}

int Acurve::hvals (double x, double *res) const
{
  int nh = 0;
  for (int i = 0; i < branches.size(); i++) {
    Abranch *branch = branches[i];
    if (x >= branch->getMinX() && x <= branch->getMaxX())
      res[nh++] = branch->h;
  }
  if (nh == 0)
    return 0;
  sort(res, res + nh);
  int savenh = nh;
  nh = 1;
  for (int i = 1; i < savenh; i++)
    if (res[i] != res[i - 1])
      res[nh++] = res[i];
  assert(nh == savenh); // eps: want to remove above loop
  return nh;
}

int Acurve::yvals (Bifurcation *bif, double *res) const
{
  Timing::numXY++;
  Timing::begin(Timing::timeXY);
  Polynomial g = f.reduce2(1, bif->point.x);
  int nr = 0;
  double y[g.degree];
  if (bif->turn()) {
    g = g.deflate2(bif->point.y);
    y[0] = y[1] = bif->point.y;
    nr = 2;
  }
  Psystem psys(g);
  nr += psys.realroots(y + nr);
  sort(y, y + nr);
  int i = 0, j;
  while (i < nr && y[i] < bbox[2])
    ++i;
  for (j = i; j < nr && y[j] <= bbox[3]; ++j)
    res[j - i] = y[j];
  Timing::end(Timing::timeXY);
  return j - i;
}

int Acurve::yvals (double x, double *res, int nh) const
{
  Timing::numXY++;
  Timing::begin(Timing::timeXY);
  Polynomial g = f.reduce2(1, x);
  double y[g.degree];
  Psystem psys(g);
  int nr = psys.realroots(y);
  sort(y, y + nr);
  int i = 0, j;
  while (i < nr && y[i] < bbox[2])
    ++i;
  for (j = i; j < nr && y[j] <= bbox[3]; ++j)
    res[j - i] = y[j];
  int ny = j - i;
  if (nh != -1 && nh != ny) {
    double dBifmin = fabs(x - bifs[0]->point.x), d;
    int iBifmin = 0;
    for (int i = 1; i < bifs.size(); ++i)
      if (dBifmin > (d = fabs(x - bifs[i]->point.x))) {
	dBifmin = d;
	iBifmin = i;
      }
    double yBifmin = bifs[iBifmin]->point.y;

    while (ny < nh)
      res[ny++] = yBifmin;
    while (ny > nh) {
      int jmin = 0;
      double dmin = fabs(yBifmin - res[0]), d;
      for (int j = 1; j < ny; j++)
	if (dmin > (d = fabs(yBifmin - res[j]))) {
	  dmin = d;
	  jmin = j;
	}
      res[jmin] = res[--ny];
    }
    sort(res, res + ny);
  }
  Timing::end(Timing::timeXY);
  return ny;
}

int Acurve::xvals (double y, double *res) const
{
  Timing::numXY++;
  Timing::begin(Timing::timeXY);
  Polynomial g = f.reduce2(2, y);
  Psystem psxs(g);
  int nx = psxs.realroots(res);
  sort(res, res + nx);
  Timing::end(Timing::timeXY);
  return nx;
}

Point Acurve::der (const Point &p)
{
  Point res;
  double xy[2] = {p.x, p.y};
  double vx = fx.val(xy),  vy = fy.val(xy);
  if (vy > 0.0 || vy == 0 && vx > 0.0) {
    res.x = vy;
    res.y = - vx;
  }
  else {
    res.x = - vy;
    res.y = vx;
  }
  return res;
}

Point Acurve::grad (const Point &p)
{
  Point res;
  double xy[2] = {p.x, p.y};
  double vx = fx.val(xy), vy = fy.val(xy);
  res.x = vx;
  res.y = vy;
  return res;
}

double Acurve::distance (const Point &p)
{
  double xy[2] = {p.x, p.y};
  double v = f.val(xy), vx = fx.val(xy), vy = fy.val(xy);
  return v/sqrt(vx*vx + vy*vy);
}

void Acurve::segments
(Segments &res, double minmin, double maxmin, double minmax, double maxmax)
{
  for (int i = 0; i < branches.size(); ++i) {
    Abranch *branch = branches[i];
    double minX = random_number(minmin, maxmin);
    double maxX = random_number(minmax, maxmax);
    if (branch->head != 0 && branch->head->point.x <= minX ||
	branch->tail != 0 && branch->tail->point.x >= maxX)
      continue;
    Point tail = branch->tail != 0 ?
      tail = branch->tail->point :
      tail = Point(minX, branch->getY(minX));
    Point head = branch->head != 0 ?
      head = branch->head->point :
      head = Point(maxX, branch->getY(maxX));
    res.push_back(Segment(tail, head, branch));
  }
}

int Acurve::closestBranchOrBif (Point p) {
  double h[branches.size()], y[f.degree];
  int nh = hvals(p.x, h), ny = yvals(p.x, y, nh);
  int iBranchMin = -1;
  double dymin = Function::FINFINITY;
  if (nh != 0) {
    assert(ny != 0);
    int imin = 0;
    double dy;
    for (int i = 0; i < ny; ++i)
      if (dymin > (dy = fabs(p.y - y[i]))) {
	dymin = dy;
	imin = i;
      }
    double hmin = h[imin];
    for (int i = 0; i < branches.size(); i++) {
      Abranch *branch = branches[i];
      if (branch->h == hmin && p.x > branch->getMinX() && p.x < branch->getMaxX()) {
	iBranchMin = i;
	break;
      }
    }
  }

  double db, dbmin = Function::FINFINITY;
  int iBifMin = -1;
  for (int i = 0; i < bifs.size(); i++)
    if (dbmin > (db = (p - bifs[i]->point).norm())) {
      dbmin = db;
      iBifMin = i;
    }
  dbmin = sqrt(dbmin);
    
  if (iBranchMin != -1 && dymin < dbmin)
    return iBranchMin + 1;
  if (iBifMin != -1)
    return -(iBifMin + 1);
  assert(0);
  return 0;
}

int Acurve::closestBif (Point p) {
  double db, dbmin = Function::FINFINITY;
  int iBifMin = -1;
  for (int i = 0; i < bifs.size(); i++)
    if (dbmin > (db = (p - bifs[i]->point).norm())) {
      dbmin = db;
      iBifMin = i;
    }
  dbmin = sqrt(dbmin);
  assert(iBifMin != -1);
  return iBifMin;
}

int Acurve::closestZone (Point p, int *zone, int &nzone) {
  int ip = closestBranchOrBif(p);
  assert(ip > 0);
  int n = branches.size() + bifs.size();
  int bb0[n], bb1[n], *bb[2] = { bb0, bb1 }, ibb[2] = { -1, -1 };

  for (int i = 0; i < 2; i++) {
    int last = ip;
    do {
      bb[i][++ibb[i]] = last;
      if (last > 0) {
	Abranch *branch = branches[last - 1];
	Bifurcation *bif = i == 0 ? branch->tail : branch->head;
	if (bif == 0)
	  last = i == 0 ? bifs.size() : bifs.size() + 1;
	else
	  for (last = 0; bifs.at(last) != bif; last++);
	last = -last - 1;
	if (ibb[i] >= 1 && bb[i][ibb[i] - 1] == last) {
	  bif = i == 1 ? branch->tail : branch->head;
	  for (last = 0; bifs.at(last) != bif; last++);
	  last = -last - 1;
	}
      }
      else {
	if (-last - 1 >= bifs.size())
	  break;
	Bifurcation *bif = bifs.at(-last - 1);
	if (bif->type != Split && bif->type != Merge)
	  break;
	assert(bif->branches.size() == 2);
	last = bif->branches[i]->index() + 1;
	if (ibb[i] >= 1 && bb[i][ibb[i] - 1] == last)
	  last = bif->branches[!i]->index() + 1;
      }
    } while (last != ip);
  }
  assert(ibb[0] < n && ibb[1] < n && (ibb[0] + ibb[1] + 1) <= 2 * n);

  nzone = 0;
  int backward = 0;
  int j = backward;
  for (int i = ibb[j]; i > 0; i--)
    zone[nzone++] = bb[j][i];
  for (int i = 0; i <= ibb[!j]; i++)
    zone[nzone++] = bb[!j][i];

  for (int i = 1; i < nzone; i++)
    if (zone[i] == zone[0]) {
      nzone = i + 1;
      break;
    }

  assert(nzone > 1);

  for (int i = 0; i < nzone; i++)
    if (zone[i] == ip)
      return i;
  assert(0);
  return -1;
}

int Acurve::closestBranchOrBif (Point p, int *zone, int nzone) {
  int ibmin = -1;
  double dymin = Function::FINFINITY, dy;

  for (int i = 0; i < nzone; i++)
    if (zone[i] < 0 && p.x == bifs.at(-zone[i] - 1)->point.x)
      return zone[i];
    else if (zone[i] > 0) {
      Abranch *branch = branches[zone[i] - 1];
      if (p.x > branch->getMinX() && p.x < branch->getMaxX() &&
	  dymin > (dy = fabs(p.y - branch->y(p.x)))) {
	dymin = dy;
	ibmin = zone[i];
      }
    }

  static double maxDYmin;
  if (maxDYmin < dymin) {
    maxDYmin = dymin;
    cout << "maxDYmin " << maxDYmin << endl;
  }
  assert(ibmin > 0);
  return ibmin;
}

void Acurve::segments (Point tail, Point head, int *zone, int nzone,
		       Segments &res) {
  int ibMin = -1, ibMax = -1;
  for (int i = 0; i < nzone; i++)
    if (zone[i] < 0) {
      int ib = -zone[i] - 1;
      if (ibMin == -1 || bifs.at(ib)->point.x < bifs.at(ibMin)->point.x)
	ibMin = ib;
      if (ibMax == -1 || bifs.at(ib)->point.x > bifs.at(ibMax)->point.x)
	ibMax = ib;
    }
  if (tail.x < bifs.at(ibMin)->point.x)
    tail.x = bifs.at(ibMin)->point.x;
  if (tail.x > bifs.at(ibMax)->point.x)
    tail.x = bifs.at(ibMax)->point.x;
  if (head.x < bifs.at(ibMin)->point.x)
    head.x = bifs.at(ibMin)->point.x;
  if (head.x > bifs.at(ibMax)->point.x)
    head.x = bifs.at(ibMax)->point.x;

  bool isCycle = zone[0] == zone[nzone - 1];
  if (isCycle) nzone--;
  int iTail = closestBranchOrBif(tail, zone, nzone);
  int iHead = closestBranchOrBif(head, zone, nzone);
  if (iTail < 0) tail = bifs.at(-iTail - 1)->point;
  if (iHead < 0) head = bifs.at(-iHead - 1)->point;
  int jTail, jHead;
  for (jTail = 0; jTail < nzone && zone[jTail] != iTail; jTail++);
  for (jHead = 0; jHead < nzone && zone[jHead] != iHead; jHead++);
  assert(jTail < nzone && zone[jTail] == iTail);
  assert(jHead < nzone && zone[jHead] == iHead);

  if (iTail == iHead) {
    int j = (iTail > 0 ? jTail :
	     jTail == 0 ? 1 : jTail == nzone - 1 ? nzone - 2 : jTail + 1);
    assert(j >= 0 && j < nzone);
    res.push_back(Segment(tail, head, branches[zone[j] - 1]));
    return;
  }

  assert(!(!isCycle && jHead < jTail && fabs(tail.x - head.x) > 1e-6));
  int d = 1;
  if (!isCycle && jHead < jTail)
    d = -1;
  if (isCycle && fabs(tail.x - head.x) < 1e-6) {
    assert(jTail != jHead);
    int nturns[2] = { 0, 0 };
    for (int id = 0; id < 2; id++) {
      int d = 1 - 2 * id;
      for (int j = jTail + d; j != jHead; j += d) {
	if (j < 0) j = nzone - 1;
	if (j >= nzone) j = 0;
	if (j == jHead) break;
	if (zone[j] == -ibMin - 1 || zone[j] == -ibMax - 1)
	  nturns[id]++;
      }
    }
    d = nturns[1] < nturns[0] ? -1 : 1;
  }

  if (iTail < 0) {
    jTail += d;
    if (jTail >= nzone) jTail = 0;
    if (jTail < 0) jTail = nzone - 1;
  }
  if (iHead < 0) {
    jHead -= d;
    if (jHead >= nzone) jHead = 0;
    if (jHead < 0) jHead = nzone - 1;
  }

  while (jTail != jHead) {
    Abranch *branch = branches[zone[jTail] - 1];
    jTail += d;
    if (jTail >= nzone) jTail = 0;
    if (jTail < 0) jTail = nzone - 1;
    Point mid = bifs.at(-zone[jTail] - 1)->point;
    jTail += d;
    if (jTail >= nzone) jTail = 0;
    if (jTail < 0) jTail = nzone - 1;
    res.push_back(Segment(tail, mid, branch));
    tail = mid;
  }

  assert(zone[jTail] > 0);
  res.push_back(Segment(tail, head, branches[zone[jTail] - 1]));
}

void Acurve::segments (Segments &res)
{
  for (int i = 0; i < branches.size(); ++i) {
    Abranch *branch = branches[i];
    Point tail = branch->tail->point;
    Point head = branch->head->point;
    res.push_back(Segment(tail, head, branch));
  }
}

Acurve * Acurve::transform (const Transformation &tr, bool ymonotone)
{
  if (tr.getT() == Point(0, 0) && tr.getR() == Point(1, 0))
    return this;
  Acurve *c = curves[tr];
  if (c == 0) {
    Point thi = tr.getR().conjugate();
    Point oi = -thi(tr.getT());
    Polynomial g = f.linsub2(thi.x, - thi.y, oi.x, thi.y, thi.x, oi.y);
    c = new Acurve(g, bbox);
    c->form_branches(ymonotone);
    curves[tr] = c;
    matchZones(c, tr);
  }
  return c;
}

void Abranch::transform (const Transformation &tr, Segments &out, bool ymonotone) 
{
  Acurve *curveT = curve->transform(tr, ymonotone);
  for (int i = 0; i < curveT->branchesTR.size(); i++)
    if (curveT->branchesTR[i] == this) {
      vector<Segment> &segs = *curveT->segsTR[i];
      for (int j = 0; j < segs.size(); j++)
	out.push_back(segs[j]);
      return;
    }

  int oldSize = out.size();
  double minX = getMinX(), maxX = getMaxX();
  double mx = 0;
  if (minX == -Function::FINFINITY && maxX == Function::FINFINITY)
    mx = 0;
  else if (minX == -Function::FINFINITY)
    mx = maxX - 1;
  else if (maxX == Function::FINFINITY)
    mx = minX + 1;
  else
    mx = 0.5 * (minX + maxX);
  double my = y(mx);
  Point mid(mx, my);
  Point midT = tr(mid);
  int z = 2 * (curveT->branches.size() + curveT->bifs.size());
  int zone[z], nzone;
  int midI = curveT->closestZone(midT, zone, nzone);
  assert(zone[midI] > 0);
  int midIsave = midI;

  if (zone[0] == zone[nzone - 1]) {
    Abranch *that = curveT->branches[zone[midI] - 1];

    Bifurcation *thisB = this->tail;
    if (thisB->type != Split && thisB->type != Merge)
      thisB = this->head;
    assert(thisB->type == Split || thisB->type == Merge);
    Abranch *thisN = this->nextBranch(thisB == this->head);

    Bifurcation *thatB = that->tail;
    if (thatB->type != Split && thatB->type != Merge)
      thatB = that->head;
    assert(thatB->type == Split || thatB->type == Merge);
    Abranch *thatN = that->nextBranch(thatB == that->head);

    if ((thisN->h < this->h) ^ (thatN->h < that->h)) {
      int n2 = nzone / 2;
      for (int i = 0; i < n2; i++) {
	int temp = zone[i];
	zone[i] = zone[nzone - 1 - i];
	zone[nzone - 1 - i] = temp;
      }
      midI = nzone - 1 - midI;
    }
  }

  Point tailT = tr(tail->point);
  int tailI = curveT->find(tr, tailT, tail->type, zone, nzone);
  Point headT = tr(head->point);
  int headI = curveT->find(tr, headT, head->type, zone, nzone);
  if (headI == 0 && head->type == Singular && zone[nzone - 1] == zone[0])
    headI = nzone - 1;

  if (zone[0] != zone[nzone - 1] && tailI > headI) {
    int n2 = nzone / 2;
    for (int i = 0; i < n2; i++) {
      int temp = zone[i];
      zone[i] = zone[nzone - 1 - i];
      zone[nzone - 1 - i] = temp;
    }
    midI = nzone - 1 - midI;
    tailI = nzone - 1 - tailI;
    headI = nzone - 1 - headI;
  }

  if (zone[0] != zone[nzone - 1] ||
      (zone[0] == zone[nzone - 1] && zone[0] < 0 &&
       curveT->bifs.at(-zone[0] - 1)->type == Singular))
    assert(tailI <= headI);
  else {
    nzone--;
    assert(tailI < nzone && headI < nzone);
  }

  int tailIsave = tailI, headIsave = headI;
  if (zone[tailI] <= 0) {
    assert(tailI != headI);
    tailI = (tailI + 1) % nzone;
  }
  if (zone[headI] <= 0)
    headI = headI > 0 ? headI - 1 : nzone - 1;

  int currI = tailI;
  Point currT = tailT;
  Abranch *currF = curveT->branches[zone[currI] - 1];
  bool foundMid = (midI == currI);

  while (currI != headI) {
    int nextI = (currI + 1) % nzone;
    assert(zone[nextI] < 0);
    Point nextT = curveT->bifs.at(-zone[nextI] - 1)->point;
    out.push_back(Segment(currT, nextT, currF));
    currI = (nextI + 1) % nzone;
    currT = nextT;
    currF = curveT->branches[zone[currI] - 1];
    foundMid = (foundMid || midI == currI);
  }
  out.push_back(Segment(currT, headT, currF));
  assert(foundMid);

  curveT->branchesTR.push_back(this);
  vector<Segment> &segs = *new vector<Segment>;
  for (int i = oldSize; i < out.size(); i++)
    segs.push_back(out[i]);
  curveT->segsTR.push_back(&segs);
}  

int Acurve::find (const Transformation &tr, Point &p, Biftype type,
		  int *zone, int nzone) {
  int iNear = -1;
  double minDist2 = Function::FINFINITY;
  for (int i = (zone[0] > 0); i < nzone; i += 2) {
    assert(zone[i] < 0);
    Bifurcation *bif = bifs.at(-zone[i] - 1);
    double dist2 = (p - bif->point).norm();
    if (dist2 < minDist2) {
	iNear = i;
	minDist2 = dist2;
    }
  }
  assert(iNear != -1);

  Bifurcation *bif = bifs.at(-zone[iNear] - 1);
  if (type == Singular)
    assert(bif->type == Singular);
  else if (tr.getR().y == 0 && (type == Split || type == Merge)) {
    Biftype tTypes[2] = { Split, Merge };
    assert(bif->type == tTypes[(tr.getR().x < 0) ^ (type == Merge)]);
  }
  if (type == Singular ||
      tr.getR().y == 0 && (type == Split || type == Merge)) {
    double dist = sqrt(minDist2);
    static double maxBif;
    if (maxBif < dist) {
      maxBif = dist;
      cout << "maxBif " << maxBif << endl;
    }
    return iNear;
  }

#ifdef BLEEN
  if (typeT != Left) {
    assert(iNear != -1);

    if (type == Singular || tr.getR().y == 0) {
      p = bifs.at(-zone[iNear) - 1]->point;
      return iNear;
    }

    Bifurcation *bNear = bifs.at(-zone[iNear) - 1];
    int iPrev = iNear > 0 ? iNear - 1 : nzone - 1;
    int iNext = (iNear + 1) % nzone;
    Abranch *bPrev = branches[zone[iPrev] - 1];
    Abranch *bNext = branches[zone[iNext] - 1];
    bool prevnext = bPrev->h < bNext->h;

    if (typeT == Split)
      assert(bPrev->tail == bNear && bNext->tail == bNear);
    else
      assert(bPrev->head == bNear && bNext->head == bNear);

    if (prevnext ^ (typeT == Split) ^ (tr.getR().y > 0))
      return iPrev;
    else
      return iNext;
  }
#endif

  int z = closestBranchOrBif(p);
  for (int i = 0; i < nzone; i++)
    if (zone[i] == z)
      return i;
  assert(0);
  return(-1);
}

Abranch *Abranch::nextBranch (bool tailhead) {
  Bifurcation *bif = tailhead ? head : tail;
  assert(bif->type == Split || bif->type == Merge);
  for (int i = 0; i < curve->branches.size(); i++) {
    Abranch *that = curve->branches[i];
    if (that != this && (that->tail == bif || that->head == bif)) {
      assert(bif == (tailhead ? that->head : that->tail));
      return that;
    }
  }
  assert(0);
  return 0;
}

double Abranch::getY (double x) const {
  Abranch *thisnc = (Abranch *) this;
#ifdef USE_HASH
  if (yCache.find(x) == thisnc->yCache.end())
    return thisnc->yCache[x] = getYnocache(x);
  else
    return thisnc->yCache[x];
#else
  Point p(x, 0);
  set<Point, PointXCompare>::iterator it = thisnc->yCache.find(p);
  if (it == yCache.end()) {
    p.y = getYnocache(p.x);
    thisnc->yCache.insert(p);
  }
  else
    p = *it;
  return p.y;
#endif
}

double Abranch::y (double x) const
{
  if (x <= getMinX())
    return tail->point.y;
  else if (x >= getMaxX())
    return head->point.y;
  double hvals[curve->branches.size()];
  double yvals[curve->f.degree];
  int nh = curve->hvals(x, hvals);
  int ny = curve->yvals(x, yvals, nh);
  int i;
  for (i = 0; i < nh && hvals[i] != h; i++)
    ;
  return yvals[i];
}

double Abranch::dy (double x) const
{
  Point p(x, getY(x)), q = curve->der(p);
  return q.y/q.x;
}

int Abranch::index ()
{
  for (int i = 0; i < curve->branches.size(); ++i)
    if (curve->branches[i] == this)
      return i;
}

void Abranch::set_head (Bifurcation *bif)
{
  head = bif;
  if (bif->point.y < ymin)
    ymin = bif->point.y;
  else if (bif->point.y > ymax)
    ymax = bif->point.y;
}

int Abranch::bboxOrder (Function *g)
{
  Abranch *ga = (Abranch *) g;
  if (ymax < ga->ymin)
    return -1;
  if (ymin > ga->ymax)
    return 1;
  return 0;
}

void Abranch::polishRoots (Abranch *b, Points &roots)
{
  double safesep = 0.001;
  if (roots.size() < 2) return;
  sort(roots.begin(), roots.end());
  int k = roots.size();
  double x1 = 0.5*(max(getMinX(), b->getMinX()) + roots[0].x);
  for (int i = 0; i < k; ++i) {
    double xn = i == k - 1 ? min(getMaxX(), b->getMaxX()) : roots[i+1].x;
    double x2 = 0.5*(roots[i].x + xn);
    if (i > 0 && roots[i].x - roots[i-1].x < safesep ||
	i < k - 1 && roots[i+1].x - roots[i].x < safesep) {
      double r = safeRoot(this, b, roots[i].x, x1, x2);
      if (r != roots[i].x) {
	roots[i].x = r;
	roots[i].y = 0.5*(y(r) + b->y(r));
      }
    }
    x1 = x2;
  }
}

/* prior version
void Abranch::polishRoots (Abranch *b, Points &roots)
{
  if (roots.empty()) return;
  sort(roots.begin(), roots.end());
  int k = roots.size();
  double x1 = 0.5*(max(getMinX(), b->getMinX()) + roots[0].x);
  for (int i = 0; i < k; ++i) {
    double xn = i == k - 1 ? min(getMaxX(), b->getMaxX()) : roots[i+1].x;
    double x2 = 0.5*(roots[i].x + xn);
    double r = safeRoot(this, b, roots[i].x, x1, x2);
    if (r != roots[i].x) {
      roots[i].x = r;
      roots[i].y = 0.5*(y(r) + b->y(r));
    }
    x1 = x2;
  }
}
*/

// zones

int Acurve::index (Abranch *branch) {
  for (int i = 0; i < branches.size(); i++)
    if (branch == branches[i])
      return i;
  assert(0);
  return -1;
}

int Acurve::index (Bifurcation *bif) {
  for (int i = 0; i < bifs.size(); i++)
    if (bif == bifs[i])
      return i;
  assert(0);
  return -1;
}

int Acurve::findZone (int bb) {
  for (int i = 0; i < zones.size(); i++) {
    vector<int> &zone = zones[i];
    for (int j = 0; j < zone.size(); j++)
      if (zone[j] == bb)
	return i;
  }
  return -1;
}

bool Acurve::extend (vector<int> &zone, int dir) {
  if (zone.size() > 1 && zone.front() == zone.back())
    return false;
  int ind = dir == -1 ? 0 : zone.size() - 1;
  Abranch *branch = branches.at(zone[ind] - 1);

  Bifurcation *bif = 0;
  if (zone.size() > 1)
    bif = bifs.at(-zone[ind - dir] - 1);

  Bifurcation *bif2 = branch->tail == bif ? branch->head : branch->tail;
  if (bif2 == 0)
    return false;
  if (dir == -1)
    zone.insert(zone.begin(), -index(bif2) - 1);
  else
    zone.push_back(-index(bif2) - 1);
  if (!(bif2->type == Split || bif2->type == Merge))
    return false;

  Abranch *branch2 = 0;
  if (bif2->branches[0] == branch)
    branch2 = bif2->branches[1];
  else if (bif2->branches[1] == branch)
    branch2 = bif2->branches[0];
  else
    assert(0);
  if (dir == -1)
    zone.insert(zone.begin(), index(branch2) + 1);
  else
    zone.push_back(index(branch2) + 1);
  return true;
}  

void Acurve::formZones () {
  for (int i = 0; i < branches.size(); i++) {
    int z = findZone(i + 1);
    if (z != -1)
      continue;

    zones.push_back(vector<int>());
    vector<int> &zone = zones.back();
    zone.push_back(i + 1);
    while (extend(zone, -1));
    while (extend(zone, 1));

    if (!(zone.size() > 1 && zone.front() == zone.back()))
      continue;

    // reverse clockwise zone
    int iSplit = -1, iMerge = -1;
    for (int i = 0; i < zone.size(); i++) {
      if (zone[i] > 0)
	continue;
      Bifurcation *bif = bifs.at(-zone[i] - 1);
      if (bif->type == Split) {
	if (iSplit == -1 ||
	    bif->point.x < bifs.at(-zone[iSplit] - 1)->point.x)
	  iSplit = i;
      }
      else if (bif->type == Merge) {
	if (iMerge == -1 ||
	    bif->point.x > bifs.at(-zone[iMerge] - 1)->point.x)
	  iMerge = i;
      }
    }

    bool reverseIt = false;
    if (iSplit != -1) {
      Abranch *prev = branches.at(zone.at(iSplit - 1) - 1);
      Abranch *next = branches.at(zone.at(iSplit + 1) - 1);
      reverseIt = prev->h < next->h;
    }
    else if (iMerge != -1) {
      Abranch *prev = branches.at(zone.at(iMerge - 1) - 1);
      Abranch *next = branches.at(zone.at(iMerge + 1) - 1);
      reverseIt = prev->h > next->h;
    }
    else
      assert(0);
    if (reverseIt)
      reverse(zone.begin(), zone.end());
  }
}

void Acurve::matchZones (Acurve *curveT, const Transformation &tr) {
  for (int i = 0; i < zones.size(); i++) {
    vector<int> &zone = zones[i];
    Abranch *branch =
      zone[0] > 0 ? branches.at(zone[0] - 1) : branches.at(zone[1] - 1);
    double minX = branch->getMinX(), maxX = branch->getMaxX();
    double mx = 0;
    if (minX == -Function::FINFINITY && maxX == Function::FINFINITY)
      mx = 0;
    else if (minX == -Function::FINFINITY)
      mx = maxX - 1;
    else if (maxX == Function::FINFINITY)
      mx = minX + 1;
    else
      mx = 0.5 * (minX + maxX);
    double my = branch->getY(mx);
    Point mid(mx, my);
    Point midT = tr(mid);
    int bb = curveT->closestBranchOrBif(midT);
    assert(bb > 0);
    int j = curveT->findZone(bb);
    assert(j >= i);
    vector<int> zoneTR = curveT->zones[j];
    assert((zone.size() > 1 && zone.front() == zone.back()) ==
	   (zoneTR.size() > 1 && zoneTR.front() == zoneTR.back()));
    assert(((zone.front() > 0) + (zone.back() > 0)) ==
	   ((zoneTR.front() > 0) + (zoneTR.back() > 0)));
    if (j > i) {
      curveT->zones[j] = curveT->zones[i];
      curveT->zones[i] = zoneTR;
    }
  }
}

int Acurve::findInZone (Point p, int z, int nearSingular) {
  vector<int> &zone = zones[z];
  if (nearSingular != 0) {
    assert(nearSingular == 1 || nearSingular == 2);
    assert(zone.front() == zone.back());
    Bifurcation *bif = bifs.at(-zone[0] - 1);
    assert(bif->type == Singular);
    Abranch *next = branches.at(zone[1] - 1);
    Abranch *prev = branches.at(zone[zone.size()-2] - 1);
    double dx = fabs(p.x - bif->point.x);
    if (dx < 1e-3 * (prev->getMaxX() - prev->getMinX()) &&
	dx < 1e-3 * (next->getMaxX() - next->getMinX()))
      return nearSingular == 1 ? 1 : zone.size() - 2;
  }

  int imin = -1;
  double distmin = Function::FINFINITY;

  for (int i = 0; i < zone.size(); i++) {
    double dist = distmin;
    if (zone[i] < 0) {
      Bifurcation *bif = bifs.at(-zone[i] - 1);
      dist = p.distance(bif->point);
    }
    else {
      Abranch *branch = branches.at(zone[i] - 1);
      if (p.x > branch->getMinX() && p.x < branch->getMaxX())
	dist = fabs(p.y - branch->getY(p.x));
    }
    if (dist < distmin) {
      distmin = dist;
      imin = i;
    }
  }

  assert(imin != -1);

  // vjm 2007/03 fix for embedding
  if (zone[imin] < 0) {
    // cout << "old imin " << imin << " distmin " << distmin << endl;
    Bifurcation *bif = bifs.at(-zone[imin] - 1);
    assert(bif->type == Merge || bif->type == Split);
    distmin = fabs(p.x - bif->point.x);
    Abranch *branch = bif->branches[p.y > bif->point.y];
    for (imin = 0; imin < zone.size(); imin++)
      if (zone[imin] > 0 && branches.at(zone[imin] - 1) == branch)
	break;
    assert(imin < zone.size());
  }

  Abranch *branch = branches.at(zone[imin] - 1);
  if ((branch->tail->type == Split || branch->tail->type == Merge) &&
      fabs(p.x - branch->tail->point.x) < 1e-6) {
    if (branch != branch->tail->branches[p.y > branch->tail->point.y]) {
      branch = branch->tail->branches[p.y > branch->tail->point.y];
      for (imin = 0; imin < zone.size(); imin++)
	if (zone[imin] > 0 && branches.at(zone[imin] - 1) == branch)
	  break;
      assert(imin < zone.size());
    }
    if (distmin > fabs(p.x - branch->tail->point.x))
      distmin = fabs(p.x - branch->tail->point.x);
  }
  else if ((branch->head->type == Split || branch->head->type == Merge) &&
	   fabs(p.x - branch->head->point.x) < 1e-6) {
    if (branch != branch->head->branches[p.y > branch->head->point.y]) {
      branch = branch->head->branches[p.y > branch->head->point.y];
      for (imin = 0; imin < zone.size(); imin++)
	if (zone[imin] > 0 && branches.at(zone[imin] - 1) == branch)
	  break;
      assert(imin < zone.size());
    }
    if (distmin > fabs(p.x - branch->head->point.x))
      distmin = fabs(p.x - branch->head->point.x);
  }
	
  static double maxDmin;
  if (maxDmin < distmin) {
    maxDmin = distmin;
    cout << "maxDmin " << maxDmin << endl;
    double cdist = distance(p);
    cout << "cdist " << cdist << endl;
  }

  // vjm 2006/12 fix for embedding
  /*
  if (zone[imin] < 0) {
    Bifurcation *bif = bifs.at(-zone[imin] - 1);
    assert(bif->type == Singular);
    assert((zone.front() == zone[imin]) != (zone.back() == zone[imin]));
    if (imin == 0)
      imin++;
    else if (imin == zone.size()-1)
      imin--;
    else
      assert(0);
  }
  */

  return imin;
}

void Abranch::transform (Segment &seg, const Transformation &tr,
			 Segments &out, bool ymonotone) {
  if (tr == Transformation()) {
    out.push_back(seg);
    return;
  }
  assert(seg.getF() == this);
  Point tailTR = tr(seg.getTail());
  Point headTR = tr(seg.getHead());
  Acurve *curveT = curve->transform(tr, ymonotone);
  int z = curve->findZone(index() + 1);
  assert(z != -1);
  const vector<int> &zone = curve->zones[z];
  const vector<int> &zoneTR = curveT->zones[z];
  int thisBB = index() + 1;

  int nearSingularity = 0;
  if (zone.front() == zone.back() && zone.front() < 0) {
    assert(curve->bifs.at(-zone.front() - 1)->type == Singular);
    if (zone[1] == thisBB)
      nearSingularity = 1;
    else if (zone[zone.size()-2] == thisBB)
      nearSingularity = 2;
  }
  int iTail = curveT->findInZone(tailTR, z, nearSingularity);
  int iHead = curveT->findInZone(headTR, z, nearSingularity);

  if (iTail == iHead) {
    int iTH = iTail;
    if (zoneTR[iTH] < 0) {
      if (iTH == 0)
	iTH++;
      else
	iTH--;
    }
    out.push_back(Segment(tailTR, headTR,
			  curveT->branches.at(zoneTR[iTH] - 1)));
    return;
  }

  int seq[zone.size()], nseq = 0;
  int dir = iTail < iHead ? 1 : -1;
  if (zone.size() > 1 && zone.front() == zone.back() && zone.front() > 0) {
    int iBranch = 0;
    for (; iBranch < zone.size() && zone[iBranch] != thisBB; iBranch++);
    assert(iBranch < zone.size());
    int iPrev = iBranch == 0 ? zone.size() - 2 : iBranch - 1;
    int iNext = iBranch + 1;
    Bifurcation *prev = curve->bifs.at(-zone[iPrev] - 1);
    Bifurcation *next = curve->bifs.at(-zone[iNext] - 1);
    int dir = 0;
    if (prev == tail && next == head)
      dir = 1;
    else if (prev == head && next == tail)
      dir = -1;
    else
      assert(0);
  }

  // vjm 2006/12 fix for embedding
  if (zoneTR[iTail] < 0)
    iTail += dir;
  if (zoneTR[iHead] < 0)
    iHead -= dir;

  if (iTail == iHead) {
    out.push_back(Segment(tailTR, headTR,
			  curveT->branches.at(zoneTR[iTail] - 1)));
    return;
  }

  for (int i = iTail; zoneTR[i] != zoneTR[iHead];) {
    seq[nseq++] = zoneTR[i];
    i += dir;
    if (i >= zoneTR.size()) i = 1;
    if (i < 0) i = zoneTR.size() - 2;
  }
  seq[nseq++] = zoneTR[iHead];

    
  out.push_back(Segment(tailTR, curveT->bifs.at(-seq[1] - 1)->point,
			curveT->branches.at(seq[0] - 1)));
  for (int i = 2; i < nseq - 2; i += 2)
    out.push_back(Segment(curveT->bifs.at(-seq[i-1] - 1)->point,
			  curveT->bifs.at(-seq[i+1] - 1)->point,
			  curveT->branches.at(seq[i] - 1)));
  out.push_back(Segment(curveT->bifs.at(-seq[nseq - 2] - 1)->point, headTR,
			curveT->branches.at(seq[nseq - 1] - 1)));
}

//post root order
int Abranch::slopeOrder (Function *g, const Point &p)
{
  Abranch *ga = (Abranch *) g;
  Point df = curve->der(p).unit(), dg = ga->curve->der(p).unit();
  double s = df.x*dg.y - df.y*dg.x;
  if (s > SAFESLOPE)
    return -1;
  else if (s < - SAFESLOPE)
    return 1;
  return 0;
}

void Abranch::trace (double x1, double x2, double dx, Points &traj)
{
  x1 = max(x1, getMinX());
  x2 = min(x2, getMaxX());
  if (x1 > x2) return;
  while (x1 < x2) {
    traj.push_back(Point(x1, y(x1)));
    x1 += dx;
  }
  traj.push_back(Point(x2, y(x2)));
}

RootList *Abranch::makeRootList (Function *that_) {
  assert(that_->getType() == IMPLICIT);
  Abranch *that = (Abranch *) that_;
  curve->form_rootlists(that->curve);
  return getRootList(that);
}  

// rootlists

#define HORIZONTAL_INTERSECTIONS

void Acurve::form_rootlists (Acurve *b)
{
  int na = branches.size(), nb = b->branches.size();
  Points points[na][nb];

  // Special case:  curve has no interior intersections with itself.
  if (this != b) {
#ifdef HORIZONTAL_INTERSECTIONS
    Acurve *that = b;
    if (this->forig.isHorizontal() || that->forig.isHorizontal()) {
      Acurve *horiz, *other;
      if (this->forig.isHorizontal()) {
	horiz = this;
	other = that;
      }
      else {
	horiz = that;
	other = this;
      }
      if (!other->forig.isHorizontal()) {
	double y = -horiz->forig.terms[0].a;

	// cout << "horizontal " << horiz << " " << y << endl;

	int n = 2 * other->f.degree + 2;
	double xs[n];
	int nx = other->xvals(y, xs);
	assert(nx <= n);
	for (int i = 0; i < nx; ++i) {
	  Point p(xs[i], y);
	  if (interior_point(p) && b->interior_point(p)) {
	    int ma = closest_branch(p)->index();
	    int mb = b->closest_branch(p)->index();
	    points[ma][mb].push_back(p);
	  }
	}
      }
    }
    else {
#endif
      Timing::numFG++;
      Timing::begin(Timing::timeFG);

      Polynomial *bf = &b->f;
      Polynomial &g = b->f;
      Polynomial h(f);

#ifdef BLEEN
      double distance = 0;
      for (int i = 0; i < f.terms.size(); i++) {
	double fcoeff = f.terms[i].a;
	double gcoeff = g.coefficient(f.terms[i].exps);
	double diff = fcoeff - gcoeff;
	distance += diff * diff;
      }
      for (int i = 0; i < g.terms.size(); i++) {
	double gcoeff = g.terms[i].a;
	double fcoeff = f.coefficient(g.terms[i].exps);
	double diff = fcoeff - gcoeff;
	distance += diff * diff;
      }
      distance = sqrt(distance);
#else
      double distance = f.distance(g);
#endif

      if (distance > 0 && distance < 1e-3) {
	Timing::numEvil++;
	int iMax = 0;
	for (int i = 0; i < h.nterms; i++) {
	  double gcoeff = g.coefficient(h.terms[i].exps);
	  assert(gcoeff != 0);
	  h.terms[i].a -= gcoeff;
	  if (fabs(h.terms[i].a) > fabs(h.terms[iMax].a))
	    iMax = i;
	}
	for (int i = 0; i < g.nterms; i++)
	  assert(f.coefficient(g.terms[i].exps) != 0);
	double aMax = h.terms[iMax].a;
	for (int i = 0; i < h.nterms; i++)
	  h.terms[i].a /= aMax;
	bf = &h;
      }

      Psystem psys(f, *bf);
      double roots[4*psys.degree()];
      int nr = 0;
      if (distance != 0) {
	nr = psys.realroots(roots);
	Timing::rootHist[nr]++;
      }
      else
	Timing::rootHist[0]++;

      curvs.push_back(b);
      croots.push_back(vector<Point>(nr));
      Timing::end(Timing::timeFG);
      Timing::begin(Timing::y);
      for (int i = 0; i < nr; ++i) {
	Point p(roots[2*i], roots[2*i+1]);

	croots.back()[i] = p;

	if (interior_point(p) && b->interior_point(p)) {
	  int ma = closest_branch(p)->index();
	  int mb = b->closest_branch(p)->index();
	  points[ma][mb].push_back(p);

	  if (fabs(p.x) <= 2 && fabs(p.y) <= 2) {
	    static double maxRDist;
	    double dist1 = fabs(this->distance(p));
	    if (dist1 > maxRDist) {
	      maxRDist = dist1;
	      cout << "maxRDist " << maxRDist << endl;
	    }
	    double dist2 = fabs(b->distance(p));
	    if (dist2 > maxRDist) {
	      maxRDist = dist2;
	      cout << "maxRDist " << maxRDist << endl;
	    }

	    if (dist1 > 1e-3 || dist2 > 1e-3) {
	      cout << this->f << endl;
	      cout << b->f << endl;
	      maxRDist = 0;
	      dist1 = this->distance(p);
	      dist2 = b->distance(p);
	    }
	  }
	}
      }
      Timing::end(Timing::y);
#ifdef HORIZONTAL_INTERSECTIONS
    }
#endif
  }

  for (int i = 0; i < na; ++i) {
    Abranch *abranch = branches[i];
    for (int j = 0; j < nb; ++j) {
      if (this == b && i >= j)
	continue;
      Abranch *bbranch = b->branches[j];
      if (abranch->getMaxX() <= bbranch->getMinX() ||
	  bbranch->getMaxX() <= abranch->getMinX())
	continue;
      Abranch *aabranch = abranch < bbranch ? abranch : bbranch;
      Abranch *bbbranch = abranch > bbranch ? abranch : bbranch;
      aabranch->polishRoots(bbbranch, points[i][j]);
      RootList *rootlist = new RootList(aabranch, bbbranch, points[i][j]);
      aabranch->setRootList(bbbranch, rootlist);
    }
  }
  //double emax, eav = rootlist_error(this, b, -10.0, 10.0, .1, emax);
}

bool Acurve::interior_point (Point &p)
{
  if (!in_bbox(p.x, p.y, bbox))
    return false;
  for (int i = 0; i < branches.size(); ++i)
    if (p.x > branches[i]->getMinX() && p.x < branches[i]->getMaxX())
      return true;
  return false;
}

Abranch * Acurve::closest_branch (Point &p, bool checkBifs)
{
  double h[branches.size()], y[f.degree];
  int nh = hvals(p.x, h), ny = yvals(p.x, y, nh);
  if (nh == 0) return 0;
  int imin = 0;
  double dy, dymin = fabs(p.y - y[0]);
  for (int i = 1; i < ny; ++i)
    if ((dy = fabs(p.y - y[i])) < dymin) {
      dymin = dy;
      imin = i;
    }
  double hmin = h[imin];
  Abranch *branch = 0;
  for (int i = 0; i < branches.size(); i++) {
    branch = branches[i];
    if (branch->h == hmin &&
	p.x >= branch->getMinX() && p.x <= branch->getMaxX())
      break;
  }
  
  if (checkBifs) {
    double db, dbmin = (p - bifs[0]->point).norm();
    int ibmin = 0;
    for (int i = 1; i < bifs.size(); i++)
      if (dbmin > (db = (p - bifs[i]->point).norm())) {
	dbmin = db;
	ibmin = i;
      }
    dbmin = sqrt(dbmin);
    
    if (branch == 0 || dbmin < dymin &&
	bifs.at(ibmin)->point != branch->tail->point &&
	bifs.at(ibmin)->point != branch->head->point) {
      Bifurcation *bif = bifs.at(ibmin);
      assert(p.x < bif->point.x && bif->type == Split ||
	     p.x > bif->point.x && bif->type == Merge);
      assert(bif->branches.size() == 2);
      assert(bif->branches[0]->h < bif->branches[1]->h);
      if (p.y < bif->point.y)
	branch = bif->branches[0];
      else
	branch = bif->branches[1];
    }
  }
  return branch;
}

void Acurve::getRootLists (RootLists &res) {
  for (int i = 0; i < branches.size(); i++)
    branches[i]->getRootLists(res);
}

void Acurve::trace (double dx, Curves &curves) const
{
  for (int i = 0; i < branches.size(); ++i) {
    curves.push_back(Points());
    branches[i]->trace(bbox[0], bbox[1], dx, *(curves.end() - 1));
  }
}

void Acurve::dump (double x1, double x2, double dx)
{
  double bbox[4] = {x1, x2, 0, 0};
  cout << "(" << endl;
  for (int i = 0; i < branches.size(); ++i) {
    Points traj;
    branches[i]->trace(x1, x2, dx, traj);
    cout << "(";
    for (int j = 0; j < traj.size(); ++j)
      cout << "(" << traj[j].x << " " << traj[j].y << ") ";
    cout << ")";
  }
  cout << ")";
}

bool in_bbox (double x, double y, double *bbox)
{
  return x > bbox[0] && x < bbox[1] && y > bbox[2] && y < bbox[3];
}

double safeRoot (Abranch *br1, Abranch *br2, double x, double a, double b)
{
  double fa = br1->y(a)- br2->y(a), fb = br1->y(b) - br2->y(b);
  double f = br1->y(x) - br2->y(x);
  double e = fabs(f), tiny = 1e-12;
  double e0 = e, x0 = x;
  static double maxPolishError = 0.0;
  while (fa*fb <= 0.0) {
    double df = br1->dy(x) - br2->dy(x);
    x = df == 0.0 ? a - 1.0: x - f/df;
    f = br1->y(x) - br2->y(x);
    if (x < a || x > b || e > tiny && fabs(f) > 0.5*e) {
      x = 0.5*(a + b);
      if (x == a || x == b)
	break;
      f = br1->y(x) - br2->y(x);
    }
    else if (fabs(f) >= e)
      break;
    e = fabs(f);
    if (fa*f > 0.0) {
      a = x;
      fa = f;
    }
    else {
      b = x;
      fb = f;
    }
  }
  if (e < e0)
    return x;
  if (e - e0 > maxPolishError) {
    maxPolishError = e - e0;
    cerr << "polish error " << e - e0 << endl;
  }
  return x0;
}

// rootlist check

double rootlist_error (Acurves &curves, double xmin, double xmax, double dx, 
		       double &emax)
{
  int n = 0;
  double emc, eav = 0.0;
  emax = 0.0;
  Acurves::iterator a, b;
  for (a = curves.begin(); a < curves.end() - 1; ++a)
    for (b = a + 1; b < curves.end(); ++b) {
      cout << "rootlist " << a - curves.begin() << " " << b - curves.begin() << endl;
      double eac = rootlist_error(*a, *b, xmin, xmax, dx, emc);
      eav += eac;
      emax = max(emax, emc);
      ++n;
    }
  return n == 0 ? 0.0 : eav/n;
}      

double rootlist_error (Acurve *a, Acurve *b, double xmin, double xmax, double dx,
		       double &emax)
{
  emax = 0.0;
  double eav = 0.0;
  int nx = 0;
  for (int ia = 0; ia < a->branches.size(); ++ia)
    for (int ib = 0; ib < b->branches.size(); ++ib) {
      Abranch *bra = a->branches[ia], *brb = b->branches[ib];
      double x1 = max(bra->getMinX(), brb->getMinX());
      double x2 = min(bra->getMaxX(), brb->getMaxX());
      if (x1 < x2) {
	Rintervals rab;
	get_rintervals(bra, brb, rab);
	for (int i = 0; i < rab.size(); ++i) {
	  double xs = max(xmin, rab[i].x1), xe = min(xmax, rab[i].x2);
	  for (double x = xs; x < xe; x = min(x + dx, xe)) {
	    double dy = rab[i].flag ? bra->y(x) - brb->y(x) :
	      brb->y(x) - bra->y(x);
	    if (dy > 0.0) eav += dy;
	    if (dy > emax) emax = dy;
	    ++nx;
	  }
	}
      }
    }
  if (emax > 0.01)
    cout << "bad pair error " << emax << endl << a->f << endl << b->f << endl;
  return nx == 0 ? 0.0 : eav/nx;
}

// inconsistency counting

int inconsistent_curves (Acurves &curves, Inconsistencies &res)
{
  int n = 0;
  Acurves::iterator a, b, c;
  for (a = curves.begin(); a < curves.end() - 2; ++a)
    for (b = a + 1; b < curves.end() - 1; ++b)
      for (c = b + 1; c != curves.end(); ++c)
	n += inconsistent_branches(*a, *b, *c, res);
  return n;
}

int inconsistent_branches (Acurve *a, Acurve *b, Acurve *c, Inconsistencies &res)
{
  int n = 0;
  Abranches::iterator ia, ib, ic;
  for (ia = a->branches.begin(); ia != a->branches.end(); ++ia)
    for (ib = b->branches.begin(); ib != b->branches.end(); ++ib) {
      double x1 = max((*ia)->getMinX(), (*ib)->getMinX());
      double x2 = min((*ia)->getMaxX(), (*ib)->getMaxX());
      if (x1 < x2)
	for (ic = c->branches.begin(); ic != c->branches.end(); ++ic) {
	  double x3 = max(x1, (*ic)->getMinX()), x4 = min(x2, (*ic)->getMaxX());
	  if (x3 < x4) {
	    ++n;
	    inconsistent_triples(*ia, *ib, *ic, res);
	  }
	}
    }
  return n;
}

void inconsistent_triples (Abranch *a, Abranch *b, Abranch *c, Inconsistencies &res)
{
  Rintervals rab, rac, rbc;
  get_rintervals(a, b, rab);
  get_rintervals(a, c, rac);
  get_rintervals(b, c, rbc);
  Rintervals::iterator iab, iac, ibc;
  for (iab = rab.begin(); iab != rab.end(); ++iab)
    for (iac = rac.begin(); iac != rac.end(); ++iac) {
      double x1 = max(iab->x1, iac->x1), x2 = min(iab->x2, iac->x2);
      bool flag = (iab->flag && !iac->flag) || (!iab->flag && iac->flag);
      if (flag && x1 < x2)
	for (ibc = rbc.begin(); ibc != rbc.end(); ++ibc) {
	  double x3 = max(x1, ibc->x1), x4 = min(x2, ibc->x2);
	  bool flag2 = ((iab->flag && !iac->flag && ibc->flag) ||
			(!iab->flag && iac->flag && !ibc->flag));
	  if (x3 < x4 && flag2) {
	    Inconsistency inc(a, b, c, x3, x4);
	    res.push_back(inc);
	  }
	}
    }
}

void get_rintervals (Abranch *a, Abranch *b, Rintervals &res)
{
  double x1 = max(a->getMinX(), b->getMinX());
  bool flag = a->compare(b, x1, 1) < 0;
  while (true) {
    double x2 = a->nextRoot(b, x1);
    if (x2 == Function::FINFINITY)
      break;
    res.push_back(Rinterval(x1, x2, flag));
    flag = !flag;
    x1 = x2;
  }
  double x2 = min(a->getMaxX(), b->getMaxX());
  res.push_back(Rinterval(x1, x2, flag));
}

double measure_inconsistencies (Inconsistencies &inc, double &emax)
{
  emax = 0.0;
  double eav = 0.0;
  for (int i = 0; i < inc.size(); ++i) {
    double dx = inc[i].x2 - inc[i].x1;
    eav += dx;
    emax = max(emax, dx);
  }
  return inc.empty() ? 0.0 : eav/inc.size();
}

