// Planar algebraic curves
#ifndef CURVE_H
#define CURVE_H
#include <queue>
#include "Segment.h"

#define SINGULAR 1e-12
#define SAFESLOPE 0.15

class Abranch;

typedef vector<Abranch *> Abranches;

enum Biftype { Left, Right, Bottom, Top, Split, Merge, Min, Max, Singular };

class Bifurcation {
 public:
  Bifurcation (Biftype itype, double x, double y) : type(itype) {
    point.x = x; point.y = y;
  }
  ~Bifurcation () {}    
  bool singular () { return type == Singular; }
  bool turn () { return type == Split || type == Merge || type == Singular; }

  Biftype type;
  Point point;
  Abranches branches;
};

typedef vector<Bifurcation *> Bifurcations;

// rootlists

class AbranchPair {
public:
  AbranchPair (Abranch *f, Abranch *g) : f(f), g(g) { assert(f < g); }
  bool operator== (const AbranchPair &fp) const {
    return f == fp.f && g == fp.g;
  }

private:
  Abranch *f, *g;
};

typedef vector<RootList *> RootLists;

typedef vector<Segment> Segments;

class Acurve {
 public:
  vector<Acurve *> curvs;
  vector<vector<Point> > croots;

  Acurve (Point &tail, Point &head, double *ibbox);
  Acurve (const Polynomial &fi, double *ibbox);
  ~Acurve ();
  bool form_branches (bool ymonotone = false);
  void bifurcations ();
  void x_turn_points ();
  void y_turn_points ();
  void adjust_bbox (double *ibbox);
  void bbox_points ();
  void bbox_points_lr (Biftype btype);
  void bbox_points_bt (Biftype btype);
  void update_left (Bifurcation *bif, Abranches &sweep);
  void update_right (Bifurcation *bif, Abranches &sweep);
  bool update_bottom (Bifurcation *bif, Abranches &sweep);
  bool update_top (Bifurcation *bif, Abranches &sweep);
  void update_split (Bifurcation *bif, Abranches &sweep);
  void update_merge (Bifurcation *bif, Abranches &sweep);
  void update_minmax (Bifurcation *bif, Abranches &sweep, bool ymonotone);
  void update_singular (Bifurcation *bif, Abranches &sweep);
  int find_branch (Bifurcation *bif, Abranches &sweep);
  int hvals (double x, double *res) const;
  int yvals (Bifurcation *bif, double *res) const;
  int yvals (double x, double *res, int nh = -1) const;
  int xvals (double y, double *res) const;
  Point der (const Point &p);
  Point grad (const Point &p);
  double distance (const Point &p);
  void segments (Segments &res,
		 double minmin, double maxmin, double minmax, double maxmax);
  void segments (Point tail, Point head, int *zone, int nzone, Segments &res);
  void segments (Segments &res);
  Acurve * transform (const Transformation &tr, bool ymonotone);
  void form_rootlists (Acurve *b);
  bool interior_point (Point &p);
  Abranch * closest_branch (Point &p, bool checkBifs = false);
  void getRootLists (RootLists &res);
  int closestBranchOrBif (Point p);
  int closestBranchOrBif (Point p, int *zone, int nzone);
  int closestBif (Point p);
  int closestZone (Point p, int *zone, int &nzone);
  int find (const Transformation &tr, Point &p, Biftype type,
	    int *zone, int nzone);
  void trace (double dx, Curves &curves) const;
  void dump (double x1, double x2, double dx);

  Polynomial f, fx, fy, forig;
  double bbox[4];
  Bifurcations bifs;
  Abranches branches;
  HashMap<Transformation, Acurve*> curves;
  vector<Abranch *> branchesTR;
  vector<vector<Segment> *> segsTR;

  // A zone is actually what is sometimes called a branch: a connected
  // component of the curve after singular points are removed.  A zone
  // is an alternating list of bifurcations and branches.  Bifurcation
  // j is represented by zone[i] = -j - 1.  Branch j is represented by
  // zone[i] = j + 1.  Therefore zone[i] = 0 is an error.  A closed
  // loop zone or an open zone that begins and ends on the same
  // singular point are represented by *zone.front() == *zone.back().
  // These two types of zone are oriented counterclockwise.
  vector<vector<int> > zones;

  // Fills in zones.  Call at the end of form_branches.
  void formZones ();
  // Makes matching zones in curveT have the same index.  Call at the
  // end of transform.
  void matchZones (Acurve *curveT, const Transformation &tr);
  // bb is a bifurcation or branch, encoded as above.  Returns the
  // index of the zone containing it or -1 if none.
  int findZone (int bb);
  // Returns the index in zones[z] of the branch which is nearest to p.
  // If the zone begins and ends on the same singularity and if p is
  // expected to be on one of the two branches neighboring the
  // singularity, set nearSingular=1 if the branch is zone[1] and
  // nearSingular=2 if the branch is zone[n-2].  In case of numerical
  // uncertainty, findInZone will return the indicated branch.
  int findInZone (Point p, int z, int nearSingular=0);

  // Helper method extends a zone in direction dir = -1 or 1.
  bool extend (vector<int> &zone, int dir);

  // Return index of branch in branches.
  int index (Abranch *branch);
  // Return index of bif in bifs.
  int index (Bifurcation *bif);
};

typedef vector<Acurve *> Acurves;

bool bif_order (Bifurcation *a, Bifurcation *b);

class Abranch : public Function {
 public:
  static int BranchId;
  Abranch (Bifurcation *itail, double ih, Acurve *icurve)
    : tail(itail), head(0), h(ih), curve(icurve), id(BranchId++) 
    { ymin = tail ? tail->point.y : Function::FINFINITY;
      ymax = tail ? tail->point.y : - Function::FINFINITY; }
  double y (double x) const;
  double dy (double x) const;
  int index ();
  void trace (double x1, double x2, double dx, Points &traj);
  Abranch *nextBranch (bool tailhead);

  // Function virtual methods:
  int getType () const { return IMPLICIT; }
  double getMinX () const { return tail ? tail->point.x : - Function::FINFINITY; }
  double getMaxX () const { return head ? head->point.x : Function::FINFINITY; }
  double getYnocache (double x) const { return y(x); }
  double getY (double x) const;
  double getDY (double x) const { return dy(x); }
  double distance (Point p) const { return curve->distance(p); }
  void set_head (Bifurcation *bif);
  int bboxOrder (Function *g);
  int slopeOrder (Function *g, const Point &p);
  void polishRoots (Abranch *b, Points &roots);
  RootList *makeRootList (Function *that);
  void transform (const Transformation &tr, Segments &out, bool ymonotone = false);
  void transform (Segment &seg, const Transformation &tr, Segments &out,
		  bool ymonotone = false);
  void free () {}

  Bifurcation *tail, *head;
  double h, ymin, ymax;
  Acurve *curve;
  int id;
#ifdef USE_HASH
  HashMap<double, double> yCache;
#else
  struct PointXCompare {
    bool operator() (const Point &a, const Point &b) {
      return a.x < b.x;
    }
  };
  set<Point, PointXCompare> yCache;
#endif
};

bool in_bbox (double x, double y, double *bbox);

double safeRoot (Abranch *br1, Abranch *br2, double x, double a, double b);

// rootlist check

double rootlist_error (Acurves &curves, double xmin, double xmax, double dx,
		       double &emax);

double rootlist_error (Acurve *a, Acurve *b, double xmin, double xmax, 
		       double dx, double &emax);

// inconsistency counting

class Inconsistency {
 public:
  Inconsistency (Abranch *ia, Abranch *ib, Abranch *ic, double ix1, double ix2) :
    a(ia), b(ib), c(ic), x1(ix1), x2(ix2) {};

  Abranch *a, *b, *c;
  double x1, x2;
};

typedef vector<Inconsistency> Inconsistencies;

class Rinterval {
 public:
  Rinterval (double ix1, double ix2, bool iflag) : x1(ix1), x2(ix2), flag(iflag) {}

  double x1, x2;
  bool flag;
};

typedef vector<Rinterval> Rintervals;

int inconsistent_curves (Acurves &curves, Inconsistencies &res);

int inconsistent_branches (Acurve *a, Acurve *b, Acurve *c, Inconsistencies &res);

void inconsistent_triples (Abranch *a, Abranch *b, Abranch *c, Inconsistencies &res);

void get_rintervals (Abranch *a, Abranch *b, Rintervals &res);

double measure_inconsistencies (Inconsistencies &inc, double &emax);

#endif
