#include "poly.h"

//g++ -O2 poly.C -llapack -lblas -lF77 -lI77

Ivector::Ivector (int in) : n(in)
{
  elts = n ? new int [n] : 0;
  for (int i = 0; i < n; ++i)
    elts[i] = 0;
}

Ivector::Ivector (const Ivector &b) : n(b.n)
{
  elts = n ? new int [n] : 0;
  for (int i = 0; i < n; ++i)
    elts[i] = b.elts[i];
}

Ivector & Ivector::operator= (const Ivector &b)
{
  if (n < b.n) {
    delete [] elts; 
    elts = new int [b.n]; 
  }
  n = b.n;
  for (int i = 0; i < n; ++i)
    elts[i] = b.elts[i]; 
  return *this;
}

int Ivector::sum () const
{
  int res = 0;
  for (int i = 0; i < n; ++i)
    res += elts[i];
  return res;
}

bool Ivector::divides (const Ivector &b) const
{
  for (int i = 0; i < n; ++i)
    if (elts[i] > b.elts[i])
      return false;
  return true;
}

Ivector Ivector::operator+ (const Ivector &b) const
{
  int nr = max(n, b.n);
  Ivector res(nr);
  for (int i = 0; i < nr; ++i)
    res.elts[i] = (i < n ? elts[i] : 0) + (i < b.n ? b.elts[i] : 0);
  return res;
}

bool Ivector::operator== (const Ivector &b) const
{
  if (n != b.n) 
    return false;
  for (int i = 0; i < n; ++i)
    if (elts[i] != b.elts[i])
      return false;
  return true;
}

double Term::val (double *x) const
{
  double v = a;
  for (int i = 0; i < exps.n; ++i)
    v *= pow(x[i], exps[i]);
  return v;
}

void Term::set_v (double *x)
{
  v += a;
  for (int i = 0; i < exps.n; ++i) {
    int d = exps.elts[i] - (pred ? pred->exps.elts[i] : 0);
    for (int j = 0; j < d; ++j)
      v *= x[i];
  }
  if (pred)
    pred->v += v;
}

Term Term::der (int v) const
{
  if (exps.n == 0 || exps.elts[v] == 0)
    return Term();
  else {
    Ivector s(exps);
    --s.elts[v];
    return Term(a*exps.elts[v], s);
  }
}

Term Term::reduce (int v, double val) const
{
  int n = exps.n;
  if (n == 0)
    return *this;
  Ivector s(n-1);
  for (int i = 0; i < n; ++i)
    if (i < v)
      s.elts[i] = exps[i];
    else if (i > v)
      s.elts[i-1] = exps[i];
  return Term(a*pow(val, exps[v]), s);
}

Term Term::operator* (const Term &b) const
{
  return Term(a*b.a, exps + b.exps);
}

Polynomial::Polynomial (int invars, int idmax)
  : nvars(invars), nterms(0), degree(0), dmax(idmax), nflag(false), vflag(false)
{
  int nt = choose(nvars, nvars + dmax);
  terms = new Term [nt];
}

Polynomial::Polynomial (const Polynomial &p) 
  : nvars(p.nvars), nterms(p.nterms), degree(p.degree), dmax(p.dmax), 
    nflag(false), vflag(false)
{
  int nt = choose(nvars, nvars + dmax);
  terms = new Term [nt];
  for (int i = 0; i < p.nterms; ++i)
    terms[i] = p.terms[i];
}

Polynomial::~Polynomial () 
{ 
  delete [] terms; 
}

void Polynomial::add (double a)
{
  Ivector iv(nvars);
  add(Term(a, iv));
}

void Polynomial::add (double a, int e1)
{
  Ivector iv(1);
  iv.elts[0] = e1;
  add(Term(a, iv));
}

void Polynomial::add (double a, int e1, int e2)
{
  Ivector iv(2);
  iv.elts[0] = e1;
  iv.elts[1] = e2;
  add(Term(a, iv));
}

void Polynomial::add (double a, int e1, int e2, int e3)
{
  Ivector iv(3);
  iv.elts[0] = e1;
  iv.elts[1] = e2;
  iv.elts[2] = e3;
  add(Term(a, iv));
}

void Polynomial::add (double a, int e1, int e2, int e3, int e4)
{
  Ivector iv(4);
  iv.elts[0] = e1;
  iv.elts[1] = e2;
  iv.elts[2] = e3;
  iv.elts[3] = e4;
  add(Term(a, iv));
}

void Polynomial::add (const Term &term)
{
  if (term.a == 0.0) return;
  nflag = vflag = false;
  for (int i = 0; i < nterms; ++i)
    if (terms[i].exps == term.exps) {
      terms[i].a += term.a;
      return;
    }
  terms[nterms] = term;
  degree = max(degree, term.degree);
  ++nterms;
}

void Polynomial::normalize_coeffs ()
{
  double amax = 0.0;
  for (int i = 0; i < nterms; ++i)
    amax = max(amax, fabs(terms[i].a));
  if (amax == 0.0) return;
  double den = 1.0/amax;
  for (int i = 0; i < nterms; ++i)
    terms[i].a *= den;
}

void Polynomial::normalize ()
{
  if (nflag) return;
  nflag = true;
  int i = 0;
  while (i < nterms)
    if (terms[i].a == 0.0) {
      terms[i] = terms[nterms-1];
      --nterms;
    }
    else
      ++i;
  bool flag = false;
  for (i = 0; i < nterms && !flag; ++i)
    flag = terms[i].degree == 0;
  if (!flag) {
    for (int j = nterms; j > 0; --j)
      terms[j] = terms[j-1];
    terms[0] = Term(0.0, Ivector(nvars));
    ++nterms;
  }
  bool sflag = true;
  for (int i = 0; i < nterms - 1 && sflag; ++i)
    sflag = terms[i] < terms[i+1];
  if (!sflag)
    sort(terms, terms + nterms);
  degree = terms[nterms-1].degree;
}

Polynomial & Polynomial::operator= (const Polynomial &b)
{
  nvars = b.nvars;
  nterms = b.nterms;
  degree = b.degree;
  dmax = b.dmax;
  int nt = choose(nvars, nvars + dmax);
  delete [] terms;
  terms = new Term [nt];
  for (int i = 0; i < nterms; ++i)
    terms[i] = b.terms[i];
  nflag = b.nflag;
  vflag = b.vflag;
  if (vflag) {
    terms[0].pred = 0;
    for (int i = 1; i < nterms; ++i)
      terms[i].pred = terms + (b.terms[i].pred - b.terms);
  }
  return *this;
}

Polynomial Polynomial::operator+ (const Polynomial &b) const
{
  Polynomial res(nvars, max(dmax, b.dmax));
  for (int i = 0; i < nterms; ++i)
    res.add(terms[i]);
  for (int i = 0; i < b.nterms; ++i)
    res.add(b.terms[i]);
  return res;
}

Polynomial Polynomial::operator* (const Polynomial &b) const
{
  Polynomial res(nvars, dmax + b.dmax);
  for (int i = 0; i < nterms; ++i)
    for (int j = 0; j < b.nterms; ++j)
      res.add(terms[i]*b.terms[j]);
  return res;
}

Polynomial Polynomial::scale (double k) const
{
  Polynomial res(nvars, dmax);
  for (int i = 0; i < nterms; ++i)
    res.add(Term(k*terms[i].a, terms[i].exps));
  return res;
}

double Polynomial::val (double *x)
{
  if (!vflag) val_normalize();
  for (int i = 0; i < nterms; ++i)
    terms[i].v = 0.0;
  for (int i = nterms - 1; i >= 0; --i)
    terms[i].set_v(x);
  return terms[0].v;
}

void Polynomial::val_normalize ()
{
  if (vflag) return;
  if (!nflag) normalize();
  vflag = true;
  terms[0].pred = 0;
  for (int i = nterms - 1; i > 0; --i) {
    Term &ti = terms[i];
    for (int j = i - 1; j >= 0; --j)
      if (terms[j].degree < ti.degree &&
	  terms[j].exps.divides(ti.exps)) {
	ti.pred = terms + j;
	break;
      }
  }
}

Polynomial Polynomial::der (int v) const
{
  --v;
  Polynomial res(nvars, dmax - 1);
  for (int i = 0; i < nterms; ++i)
    res.add(terms[i].der(v));
  return res;
}

double Polynomial::coefficient (const Ivector &s) const
{
  for (int i = 0; i < nterms; ++i)
    if (terms[i].exps == s)
      return terms[i].a;
  return 0.0;
}

Polynomial Polynomial::reduce (int v, double val) const
{
  --v;
  Polynomial res(nvars - 1, dmax);
  for (int i = 0; i < nterms; ++i)
    res.add(terms[i].reduce(v, val));
  return res;
}

Polynomial Polynomial::reduce2 (int v, double val) const
{
  --v;
  int d = 0;
  for (int i = 0; i < nterms; ++i)
    d = max(d, terms[i].exps.elts[1-v]);
  double a[d+1];
  for (int i = 0; i <= d; ++i)
    a[i] = 0.0;
  for (int i = 0; i < nterms; ++i) {
    const Term &t = terms[i];
    a[t.exps.elts[1-v]] += t.a*pow(val, t.exps.elts[v]);
  }
  return unpack_univariate(a, d);
}

// deflate double root of univariate polynomial
Polynomial Polynomial::deflate2 (double r) const
{
  double a[degree+1], b[degree], c[degree-1];
  pack_univariate(a);
  deflate(a, b, degree, r);
  deflate(b, c, degree - 1, r);
  return unpack_univariate(c, degree - 2);
}

void deflate (double *a, double *b, int d, double r)
{
  b[d - 1] = a[d];
  for (int i = d - 2; i >= 0; --i)
    b[i] = a[i + 1] + r*b[i + 1];
}

void Polynomial::pack_univariate (double *a) const
{
  for (int i = 0; i <= degree; ++i)
    a[i] = 0.0;
  for (int i = 0; i < nterms; ++i)
    a[terms[i].degree] = terms[i].a;
}

Polynomial Polynomial::linsub2 (double a, double b, double c, double d, 
				double e, double f) const
{
  Polynomial res(2, dmax);
  for (int nt = 0; nt < nterms; ++nt) {
    int m = terms[nt].exps.elts[0], n = terms[nt].exps.elts[1];
    for (int i = 0; i <= m; ++i) {
      double ic = choose(i, m)*pow(a, i);
      for (int j = 0; j <= m - i; ++j) {
	double jc = choose(j, m - i)*pow(b, j)*pow(c, m - i - j);
	for (int k = 0; k <= n; ++k) {
	  double kc = choose(k, n)*pow(d, k);
	  for (int l = 0; l <= n - k; ++l) {
	    double lc = choose(l, n - k)*pow(e, l)*pow(f, n - k - l);
	    Ivector s(2);
	    s.elts[0] = i + k;
	    s.elts[1] = j + l;
	    res.add(Term(terms[nt].a*ic*jc*kc*lc, s));
	  }
	}
      }
    }
  }
  return res;
}

bool Polynomial::isHorizontal () {
  return (nvars == 2 &&
	  degree == 1 &&
	  nterms == 2 &&
	  terms[0].degree == 0 &&
	  terms[1].degree == 1 &&
	  terms[1].exps.n == 2 &&
	  terms[1].exps.elts[0] == 0 &&
	  terms[1].exps.elts[1] == 1);
}

double Polynomial::distance (const Polynomial &g) const {
  double distance = 0;
  for (int i = 0; i < nterms; i++) {
    double fcoeff = terms[i].a;
    double gcoeff = g.coefficient(terms[i].exps);
    double diff = fcoeff - gcoeff;
    distance += diff * diff;
  }
  for (int i = 0; i < g.nterms; i++) {
    double gcoeff = g.terms[i].a;
    double fcoeff = coefficient(g.terms[i].exps);
    double diff = fcoeff - gcoeff;
    distance += diff * diff;
  }
  distance = sqrt(distance / 2);
  return distance;
}

Polynomial Polynomial::homogenize () const
{
  Polynomial res(nvars + 1, dmax);
  Ivector v(nvars + 1);
  for (int i = 0; i < nterms; ++i) {
    const Term &t = terms[i];
    for (int j = 0; j < nvars; ++j)
      v.elts[j] = t.exps[j];
    v.elts[nvars] = degree - t.degree;
    res.add(Term(t.a, v));
  }
  return res;
}

Polynomial Polynomial::homogenize (const Polynomial &p) const
{
  Polynomial q(nvars, degree), res(nvars, degree);
  q.add(1.0);
  int d = degree;
  for (int i = nterms - 1; i >= 0; --i) {
    const Term &t = terms[i];
    if (t.degree < d) {
      for (int j = 0; j < d - t.degree; ++j)
	q = q*p;
      d = t.degree;
    }
    for (int j = 0; j < q.nterms; ++j) {
      const Term &qt = q.terms[j];
      res.add(Term(t.a*qt.a, t.exps + qt.exps));
    }
  }
  return res;
}

Polynomial operator* (double k, const Polynomial &a)
{
  return a.scale(k);
}

Polynomial unpack_univariate (double *a, int d)
{
  while (d > 0 && a[d] == 0.0)
    --d;
  Polynomial res(1, d);
  res.nterms = d + 1;
  res.degree = d;
  Ivector s(1);
  for (int i = 0; i <= d; ++i) {
    s.elts[0] = i;
    res.terms[i] = Term(a[i], s);
  }
  res.nflag = true;
  return res;
}

Polynomial linear_polynomial (int nvars, double *a)
{
  Polynomial res(nvars, 1);
  res.add(a[0]);
  for (int i = 0; i < nvars; ++i) {
    Ivector exps(nvars);
    exps.elts[i] = 1;
    res.add(Term(a[i+1], exps));
  }
  return res;
}

Polynomial factored_polynomial (int d, double *roots)
{
  double a[d+1];
  factored_polynomial_aux(d, roots, a);
  return unpack_univariate(a, d);
}

void factored_polynomial_aux (int d, double *roots, double *coeffs)
{
  if (d == 1) {
    coeffs[0] = - roots[0];
    coeffs[1] = 1.0;
  }
  else {
    double ncoeffs[d];
    factored_polynomial_aux(d - 1, roots + 1, ncoeffs);
    coeffs[0] = - roots[0]*ncoeffs[0];
    for (int i = 1; i < d; ++i)
      coeffs[i] = ncoeffs[i-1] - roots[0]*ncoeffs[i];
    coeffs[d] = 1.0;
  }
}

Polynomial bivariate_polynomial (int d, double *a)
{
  Ivector s(2);
  Polynomial f(2, d);
  for (int i = 0; i <= d; ++i)
    for (int j = 0; j <= d - i; ++j, ++a) {
      s.elts[0] = i;
      s.elts[1] = j;
      f.add(Term(*a, s));
    }
  return f;
}

bool fit_curve (int d, double *points, Polynomial &p)
{
  int n = (d+1)*(d+2)/2 - 1;
  double a[n*n], b[n], c[n+1];
  for (int i = 0; i <= n; ++i)
    c[i] = 1.0;
  p = bivariate_polynomial(d, c);
  for (int i = 0; i < n; ++i, points += 2) {
    b[i] = -1.0;
    for (int j = 0; j < n; ++j)
      a[i+n*j] = p.terms[j+1].val(points);
  }
  if (!linsolve(a, b, n))
    return false;
  for (int i = 0; i < n; ++i)
    p.terms[i+1].a = b[i];
  return true;
}

double random_number (double minX, double maxX) 
{
  return minX + (maxX - minX) * rand() / (double)RAND_MAX;
}

Polynomial random_linear_polynomial (int nvars)
{
  double k[nvars+1];
  for (int i = 0; i <= nvars; ++i)
    k[i] = random_number(-1.0, 1.0);
  return linear_polynomial(nvars, k);
}

Polynomial random_polynomial (int nvars, int dmax, double cmax)
{
  Ivector *monomials;
  int nt = form_monomials(nvars, dmax, monomials);
  Polynomial p(nvars, dmax);
  for (int i = 0; i < nt; ++i)
    p.add(Term(random_number(- cmax, cmax), monomials[i]));
  delete [] monomials;
  return p;
}

int form_monomials (int nvars, int dmax, Ivector *&pp)
{  
  int n = choose(nvars, nvars + dmax);
  pp = new Ivector [n];
  pp[0] = Ivector(nvars);
  int nprev = 1, npp = 1;
  for (int i = 0; i < nvars; ++i) {
    for (int j = 0; j < nprev; ++j) {
      const Ivector &ppj = pp[j];
      int dd = dmax - ppj.sum();
      for (int k = 1; k <= dd; ++k) {
	pp[npp] = ppj;
	pp[npp].elts[i] = k;
	++npp;
      }
    }
    nprev = npp;
  }
  sort(pp, pp + n);
  return n;
}

int choose (int k, int n)
{
  int res = 1;
  for (int i = 0; i < k; ++i)
    res *= n - i;
  for (int i = 2; i <= k; ++i)
    res /= i;
  return res;
}

Polynomial read_polynomial (istream &is)
{
  int nv, nt;
  is >> nv >> nt;
  double a;
  Term terms[nt];
  for (int i = 0; i < nt; ++i) {
    is >> a;
    Ivector iexps(nv);
    for (int j = 0; j < nv; ++j) {
      int k;
      is >> k;
      iexps.elts[j] = k;
    }
    terms[i] = Term(a, iexps);
  }
  int degree = 0;
  for (int i = 0; i < nt; ++i)
    degree = max(degree, terms[i].degree);
  Polynomial poly(nv, degree);
  for (int i = 0; i < nt; ++i)
    poly.add(terms[i]);
  return poly;
}

ostream& operator<< (ostream &os, const Polynomial &poly)
{
  os << endl << poly.nvars << " " << poly.nterms << endl;
  for (int i = 0; i < poly.nterms; ++i)
    os << poly.terms[i];
  os << endl;
  return os;
}

ostream& operator<< (ostream &os, const Term &term)
{
  os << setprecision(20) << term.a << " ";
  for (int i = 0; i < term.exps.n; ++i)
    os << term.exps[i] << " ";
  os << endl;
  return os;
}

void ppoly (const Polynomial &p)
{
  cout << p;
}

void pterm (const Term &t)
{
  cout << t;
}

Trie::Trie (int in) : n(in), v(0)
{
  c = new Trie * [n];
  for (int i = 0; i < n; ++i)
    c[i] = 0;
}

Trie::~Trie () 
{
  for (int i = 0; i < n; ++i)
    delete c[i];
  delete [] c;
}

void Trie::insert (const Ivector &a, int va, int i)
{
  if (i == a.n)
    v = va;
  else {
    int ai = a.elts[i];
    if (!c[ai])
      c[ai] = new Trie(n);
    c[ai]->insert(a, va, i + 1);
  }
}

int Trie::find (const Ivector &a, int i) const
{
  if (i == a.n)
    return v;
  return c[a.elts[i]]->find(a, i + 1);
}

Eigen::Eigen (const Eigen &b)
{ 
  n = b.n;
  k = b.k;
  val = b.val;
  vec = new double [n];
  for (int i = 0; i < n; ++i)
    vec[i] = b.vec[i];
}  

void Eigen::init (int ni, int ki, double ival, double *ivec)
{
  if (ni > n) {
    delete [] vec;
    vec = new double [ni];
  }
  n = ni;
  k = ki;
  val = ival;
  for (int i = 0; i < n; ++i)
    vec[i] = ivec[i];
}

bool Eigen::operator< (const Eigen &b) const
{
  return k < b.k || k == b.k && val < b.val;
}

Eigen & Eigen::operator= (const Eigen &b)
{
  init(b.n, b.k, b.val, b.vec);
  return *this;
}

Psystem::Psystem (const Polynomial &f)
{
  n = 1;
  polys = new Polynomial [n];
  polys[0] = f;
  init();
}

Psystem::Psystem (const Polynomial &f, const Polynomial &g)
{
  n = 2;
  polys = new Polynomial [n];
  polys[0] = f;
  polys[1] = g;
  init();
}

Psystem::Psystem (const Polynomial &f, const Polynomial &g, const Polynomial &h)
{
  n = 3;
  polys = new Polynomial [n];
  polys[0] = f;
  polys[1] = g;
  polys[2] = h;
  init();
}

Psystem::Psystem (const Polynomial &f, const Polynomial &g, const Polynomial &h,
		  const Polynomial &i)
{
  n = 4;
  polys = new Polynomial [n];
  polys[0] = f;
  polys[1] = g;
  polys[2] = h;
  polys[3] = i;
  init();
}

Psystem::Psystem (int in, Polynomial *ipolys)
{
  n = in;
  polys = new Polynomial [n];
  for (int i = 0; i < n; ++i)
    polys[i] = ipolys[i];
  init();
}

Psystem::~Psystem ()
{ 
  delete [] polys; 
  delete [] jac;
  delete [] pp; 
  delete [] p; 
  delete [] a; 
  delete [] bdata;
  delete [] bflag;
  delete ptrie; 
}

void Psystem::init ()
{
  m = 1;
  for (int i = 0; i < n; ++i) {
    polys[i].normalize();
    m *= polys[i].degree;
  }
  pp = 0;
  p = 0;
  a = 0;
  bdata = new double [n*m*m];
  bflag = new bool [n];
  for (int i = 0; i < n; ++i)
    bflag[i] = false;
  ptrie = 0;
  jac = 0;
}

void Psystem::init_jac ()
{
  if (jac) return;
  jac = new Polynomial [n*n];
  for (int i = 0; i < n; ++i)
    for (int j = 0; j < n; ++j)
      jac[i+n*j] = polys[i].der(j+1);
}

int Psystem::realroots (double *res, double *alleigen)
{
  int nr = simpleroots(res, alleigen);
  if (nr == -1) {
    srandom(1);
    double *fres = new double [n*m];
    for (int i = 0; i < 3; ++i) {
      Polynomial p = random_linear_polynomial(n);
      Psystem fsys = i < 2 ? homogenize(p) : perturb(0.01).homogenize(p);
      nr = fsys.simpleroots(fres, alleigen);
      if (nr != -1) {
	nr = dehomogenize(p, nr, fres, res);
	break;
      }
    }
    delete [] fres;
  }
  if (nr != -1 && n > 1)
    polish_roots(nr, res);
  return nr;
}

Psystem Psystem::homogenize (const Polynomial &p) const
{
  Polynomial fpolys[n];
  for (int i = 0; i < n; ++i)
    fpolys[i] = polys[i].homogenize(p);
  return Psystem(n, fpolys);
}

int Psystem::dehomogenize (Polynomial &p, int nf, double *fres, 
			   double *res) const
{
  int nr = 0;
  for (int i = 0; i < nf; ++i) {
    double z = p.val(fres + i*n);
    if (z != 0.0) {
      for (int j = 0; j < n; ++j)
	res[nr*n+j] = fres[i*n+j]/z;
      ++nr;
    }
  }
  return nr;
}

int Psystem::simpleroots (double *res, double *alleigen)
{
  if (m == 0) return 0;
  setup_pp();
  setup_a();
  setup_p();
  if (!solve_a())
    return -1;
  setup_ptrie();
  return simplesolve(res, alleigen);
}

void Psystem::setup_pp ()
{
  dmax = 1;
  for (int i = 0; i < n; ++i)
    dmax += polys[i].degree - 1;
  int npp = form_monomials(n, dmax, pp);
  N = npp - m;
  p = new integer [N+m];
  a = new double [(N+m)*N];
}

void Psystem::setup_a () const
{
  Trie trie(dmax + 1);
  for (int i = 0; i < N + m; ++i)
    trie.insert(pp[i], i);
  for (int i = 0; i < N*(N + m); ++i)
    a[i] = 0.0;
  int nrow = 0;
  Ivector s(n);
  for (int i = 0; i < N + m && nrow < N; ++i) {
    Ivector z = pp[i];
    for (int j = n - 1; j >= 0; --j) {
      const Polynomial &pj = polys[j];
      if (z.elts[j] >= pj.degree) {
	z.elts[j] -= pj.degree;
	for (int k = 0; k < pj.nterms; ++k) {
	  const Term &tk = pj.terms[k];
	  for (int l = 0; l < n; ++l)
	    s.elts[l] = z.elts[l] + tk.exps.elts[l];
	  int ns = trie.find(s);
	  a[ns*N+nrow] = tk.a;
	}
	++nrow;
	break;
      }
    }
  }
}

void Psystem::setup_p () const
{
  int pmax = pp[0].sum(), i0;
  for (int i = 0; i < N + m; ++i)
    if (pp[i].sum() == pmax)
      p[i] = i + 1;
    else {
      i0 = i;
      break;
    }
  for (int i = i0; i < N + m - 1; ++i)
    p[i] = 0;
  p[N+m-1] = N + m;
}

bool Psystem::solve_a () const
{
  return qr_column_pivot (a, N, N + m, p) &&
         linsolve_triangular(a, a + N*N, N, m);
}

void Psystem::setup_ptrie ()
{
  ptrie = new Trie(dmax + 1);
  for (int i = 0; i < N + m; ++i)
    ptrie->insert(pp[p[i]], i);
}

int Psystem::simplesolve (double *res, double *alleigen) const
{
  double *b = get_b(0);
  Eigen *eigen;
  int nr = real_eigen(b, m, eigen, alleigen);
  if (nr == -1)
    return -1;
  if (n == 1)
    for (int i = 0; i < nr; ++i)
      res[i] = eigen[i].val;
  else {
    check_eigen(nr, eigen);
    int np = 0;
    for (int i = 0; i < nr; ++i)
      if (root(eigen[i], res + n*np))
	++np;
    nr = np;
  }
  delete [] eigen;
  return nr;
}

double * Psystem::get_b (int k) const
{
  if (!bflag[k]) {
    bflag[k] = true;
    setup_b(k, bdata + k*m*m);
  }
  return bdata + k*m*m;
}

void Psystem::setup_b (int k, double *b) const
{
  Ivector s(n);
  s.elts[k] = 1;
  for (int i = 0; i < m; ++i) {
    Ivector x = s + pp[p[i+N]];
    int nx = ptrie->find(x);
    if (nx < N)
      for (int j = 0; j < m; ++j)
	b[m*j+i] = - a[N*(N+j)+nx];
    else
      for (int j = 0; j < m; ++j)
	b[m*j+i] = nx == j + N ? 1.0 : 0.0;
  }
}

void Psystem::check_eigen (int &nr, Eigen *eigen) const
{
  double dmin = 0.01;
  sort(eigen, eigen + nr);
  int i = 0;
  while (i < nr - 1) {
    int j = i + 1;
    while (j < nr && eigen[j].k == eigen[i].k && 
	   eigen[j].val - eigen[i].val < dmin)
      ++j;
    if (j == i + 1 || eigen[i].k == n - 1 || 
	!adjust_eigen(nr, eigen, i, j))
      i = j;
  }
}

bool Psystem::adjust_eigen (int nr, Eigen *eigen, int s, int e) const
{
  int na = e - s, k = eigen[s].k + 1;
  double *b = get_b(k), *c = new double [m*na], *d = new double[na*na];
  for (int i = 0; i < na; ++i)
    copy_v(m, eigen[s+i].vec, c + m*i);
  orthogonalize(c, m, na);
  at_times_b_times_a(m, na, c, b, d);
  Eigen *deigen;
  int nd = real_eigen(d, na, deigen);
  bool flag = nd == na && 
    fabs(deigen[na-1].val - deigen[0].val) > eigen[e-1].val - eigen[s].val;
  if (flag) {
    for (int i = 0; i < na; ++i) {
      eigen[s+i].k = k;
      eigen[s+i].val = deigen[i].val;
      m_times_v(m, na, c, deigen[i].vec, eigen[s+i].vec);
    }
    sort(eigen + s, eigen + e);
  }
  delete [] c;
  delete [] d;
  delete [] deigen;
  return flag;
}

bool Psystem::root (const Eigen &eigen, double *r) const
{
  double small = 1e-4; // was 1e-8
  if (eigen.vec[m-1] == 0.0)
    return false;
  if (fabs(eigen.vec[m-1]) < small)
    return root2(eigen, r);
  root1(eigen, r);
  return true;
}

void Psystem::root1 (const Eigen &eigen, double *r) const
{
  for (int i = 0; i < n; ++i)
    if (i == eigen.k)
      r[i] = eigen.val;
    else {
      Ivector x(n);
      x.elts[i] = 1;
      int nx = ptrie->find(x);
      if (nx < N) {
	r[i] = 0.0;
	for (int j = 0; j < m; ++j)
	  r[i] -= a[N*(N+j)+nx]*eigen.vec[j];
      }
      else
	r[i] = eigen.vec[nx-N];
      r[i] /= eigen.vec[m-1];  
    }
}

bool Psystem::root2 (const Eigen &eigen, double *r) const
{
  int nrow = 0;
  for (int i = 1; i < m; ++i)
    if (fabs(eigen.vec[i]) > fabs(eigen.vec[nrow]))
      nrow = i;
  if (eigen.vec[nrow] == 0.0)
    return false;
  for (int i = 0; i < n; ++i)
    if (i == eigen.k)
      r[i] = eigen.val;
    else {
      double *b = get_b(i), v = 0.0;
      for (int j = 0; j < m; ++j)
	v += b[nrow+j*m]*eigen.vec[j];
      r[i] = v/eigen.vec[nrow];
    }
  return true;
}

void Psystem::polish_roots (int nr, double *x)
{
  init_jac();
  for (int i = 0; i < nr; ++i)
    polish_root(x + n*i);
}

double Psystem::polish_root (double *x) const
{
  double v[n], dv[n*n], nx[n], err = 0.0;
  val(x, v);
  for (int j = 0; j < n; ++j)
    err += fabs(v[j]);
  for (int i = 0; i < 10; ++i) {
    dval(x, dv);
   if (!linsolve(dv, v, n)) break;
    for (int j = 0; j < n; ++j)
      nx[j] = x[j] - v[j];
    val(nx, v);
    double nerr = 0.0;
    for (int j = 0; j < n; ++j)
      nerr += fabs(v[j]);
    if (nerr >= 5.0*err)
      break;
    for (int j = 0; j < n; ++j)
      x[j] = nx[j];
    err = nerr;
    if (err == 0.0)
      break;
  }
  return err;
}

void Psystem::val (double *x, double *v) const
{
  for (int i = 0; i < n; ++i)
    v[i] = polys[i].val(x);
}

void Psystem::dval (double *x, double *dv) const
{
  for (int i = 0; i < n; ++i)
    for (int j = 0; j < n; ++j)
      dv[i+n*j] = jac[i+n*j].val(x);
}

Psystem Psystem::perturb (double eps)
{
  Polynomial *ppolys = new Polynomial [n];
  for (int i = 0; i < n; ++i)
    ppolys[i] = polys[i] + random_polynomial(n, polys[i].degree, eps);
  Psystem psys(n, ppolys);
  delete [] ppolys;
  return psys;
}

double max_norm (int n, double *x)
{
  double d = 0.0;
  for (int i = 0; i < n; ++i)
    d = max(d, fabs(x[i]));
  return d;
}

void copy_v (int n, double *v, double *w)
{
  for (int i = 0; i < n; ++i)
    w[i] = v[i];
}

void m_times_v (int m, int n, double *a, double *v, double *b)
{
  for (int i = 0; i < m; ++i)
    b[i] = 0.0;
  for (int j = 0; j < n; ++j)
    for (int i = 0; i < m; ++i, ++a)
      b[i] += (*a)*v[j];
}

void at_times_b_times_a (int m, int n, double *a, double *b, double *c)
{
  double *d = new double [m*n];
  for (int i = 0; i < m; ++i)
    for (int j = 0; j < n; ++j) {
      d[m*j+i] = 0.0;
      for (int k = 0; k < m; ++k)
	d[m*j+i] += b[m*k+i]*a[m*j+k];
    }
  for (int i = 0; i < n; ++i)
    for (int j = 0; j < n; ++j) {
      c[n*j+i] = 0.0;
      for (int k = 0; k < m; ++k)
	c[n*j+i] += a[m*i+k]*d[m*j+k];
    }
  delete [] d;
}

// clapack

bool qr_column_pivot (double *a, int m, int n, integer *p)
{
  double small = 1e-12;
  char side = 'L', trans = 'T';
  integer nr = m, nc = n - 1, nc0 = 1, lwork = 1000*nc, info;
  double *tau = new double [nr], *work = new double [lwork];
  double *a0 = a + nc*nr;
  dgeqp3_(&nr, &nc, a, &nr, p, tau, work, &lwork, &info);
  bool flag = info == 0;
  for (int i = 0; flag && i < m; ++i)
    flag = fabs(a[i+m*i]) > small;
  if (flag) {
    dormqr_(&side, &trans, &nr, &nc0, &nr, a, &nr, tau, a0, &nr, 
	    work, &lwork, &info);
    for (int i = 0; i < n; ++i)
      --p[i];
  }
  delete [] tau;
  delete [] work;
  return flag;
}

bool linsolve_triangular (double *a, double *b, int m, int k)
{
  char uplo = 'U', trans = 'N', diag = 'N';
  integer n = m, nrhs = k, info;
  dtrtrs_(&uplo, &trans, &diag, &n, &nrhs, a, &n, b, &n, &info);
  return info == 0;
}

bool linsolve (double *a, double *b, int k)
{
  integer n = k, nrhs = 1, lda = k, ipiv[k], ldb = k, info;
  dgesv_(&n, &nrhs, a, &lda, ipiv, b, &ldb, &info);
  return info == 0;
}

int real_eigen (double *a, int k, Eigen *&eigen, double *alleigen)
{
  char balanc = 'B', jobvl = 'N', jobvr = 'V', sense = 'N';
  integer n = k, lda = k, ldvl = 1, ldvr = k, ilo, ihi, 
    lwork = 10*k*(k+6), iwork[2*k-2], info;
  double wr[k], wi[k], vl[1], vr[k*k], scale[k], abnrm, rconde[k], 
    rcondv[k], work[lwork];
  double aa[k*k];
  copy_v(k*k, a, aa);
  dgeevx_(&balanc, &jobvl, &jobvr, &sense, &n, aa, &lda, wr, wi,
	  vl, &ldvl, vr, &ldvr, &ilo, &ihi, scale, &abnrm,
	  rconde, rcondv, work, &lwork, iwork, &info);
  if (info != 0) 
    return -1;
  double large = 1.0e4, small = 1e-10;
  int nr = 0;
  for (int i = 0; i < k; ++i) {
    if (fabs(wr[i]) + fabs(wi[i]) > large)
      return -1;
    if (fabs(wi[i]) < small)
      ++nr;
  }
  eigen = new Eigen [nr];
  int er = 0;
  for (int i = 0; i < k; ++i)
    if (fabs(wi[i]) < small) {
      eigen[er].init(k, 0, wr[i], vr + i*k);
      ++er;
    }
  if (alleigen)
    for (int i = 0; i < k; ++i) {
      alleigen[2*i] = wr[i];
      alleigen[2*i+1] = wi[i];
    }
  return nr;
}

bool orthogonalize (double *a, int nr, int nc)
{
  integer m = nr, n = nc, lda = nr, lwork = n*64, info;
  double tau[nr], work[lwork];
  dgeqrf_(&m, &n, a, &lda, tau, work, &lwork, &info);
  if (info != 0) 
    return false;
  dorgqr_(&m, &n, &n, a, &lda, tau, work, &lwork, &info);
  return info == 0;
}

// debugging

void print_matrix (double *a, int m, int n)
{
  cerr << endl;
  for (int i = 0; i < m; ++i) {
    for (int j = 0; j < n; ++j)
      cerr << a[i+m*j] << " ";
    cerr << endl;
  }
}

double check_eigenvector (double *a, Eigen &eigen)
{
  int n = eigen.n;
  double err = 0.0;
  for (int i = 0; i < n; ++i) {
    double ei = 0.0, v = 0.0;
    for (int j = 0; j < n; ++j)
      v += a[i+j*n]*eigen.vec[j];
    ei = fabs(eigen.val*eigen.vec[i] - v);
    err = max(err, ei);
  }
  return err;
}

void roots_at_infinity (Polynomial &f, Polynomial &g)
{
  roots_at_infinity(f, g, 1.0);
  roots_at_infinity(f, g, 0.0);
}

void roots_at_infinity (Polynomial &f, Polynomial &g, double v)
{
  Polynomial f1 = f.reduce(2, v), g1 = g.reduce(2, v);
  double *r = new double [f1.degree];
  Psystem psys(f1);
  int nr = psys.realroots(r);
  double dmin = 1e20;
  int imin = -1;
  for (int i = 0; i < nr; ++i) {
    double di = fabs(g1.val(r + i));
    if (di < dmin) {
      dmin = di;
      imin = i;
    }
  }
  if (imin != -1)
    cout << "root 1 = " << r[imin] << " " << v 
	 << " val 2 = " << dmin << endl;
}

/* running time
int main ()
{
  int n = 100, dmax = 10;
  double res[dmax*dmax*2], a = 0.0, b = 0.0;
  for (int d = 1; d <= dmax; ++d) {
    clock_t t0 = clock();
    for (int i = 0; i < n; ++i) {
      Psystem psys(random_polynomial(2, d), random_polynomial(2, d));
      psys.realroots(res);
    }
    double bi = (clock() - t0)/(1.0*CLOCKS_PER_SEC*n);
    double ai = d*d*d*d;
    a += ai*ai;
    b += ai*bi;
    cout << "degree " << d << " c estimate " << bi/ai << endl;
  }
  cout << "least squares c " << b/a << endl;
  return 0;
}
*/

/*
int main ()
{
  double dummy;
  for (int i = 0; i < 5; ++i)
    cin >> dummy;
  int n;
  cin >> n;
  Polynomial *polys = new Polynomial [n];
  for (int i = 0; i < n; ++i) {
    polys[i] = read_polynomial(cin);
    polys[i].normalize_coeffs();
  }
  Psystem psys(n, polys);
  double *roots = new double [n*psys.degree()], 
    *all = new double[2*psys.degree()];
  int nr = psys.realroots(roots, all);
  cout << setprecision(20) << "degree/roots " << psys.degree() 
       << "/" << nr << endl;
  for (int i = 0; i < nr; ++i) {
    cout << "(";
    for (int j = 0; j < n; ++j)
      cout << roots[n*i+j] << " ";
    cout << ")" << endl;
  }
  delete [] roots;
  delete [] polys;
  return 0;
}
*/
