#include <iostream>
#include <iomanip>
#include <math.h>

using namespace std;

//typedef long int integer;
typedef int integer;
typedef double doublereal;

extern "C" {
  int dgeqp3_ (integer *m, integer *n, doublereal *a, integer *
	       lda, integer *jpvt, doublereal *tau, doublereal *work, 
	       integer *lwork, integer *info);
  int dtrtrs_ (char *uplo, char *trans, char *diag, integer *n, 
	       integer *nrhs, doublereal *a, integer *lda, doublereal *b,
	       integer *ldb, integer *info);
  int dormqr_ (char *side, char *trans, integer *m, integer *n, 
	       integer *k, doublereal *a, integer *lda, doublereal *tau, 
	       doublereal *c, integer *ldc, doublereal *work, integer *lwork, 
	       integer *info);
  int dgeevx_ (char *balanc, char *jobvl, char *jobvr, char *sense, integer *n, 
	       doublereal *a, integer *lda, doublereal *wr, doublereal *wi, 
	       doublereal *vl, integer *ldvl, doublereal *vr, integer *ldvr, 
	       integer *ilo, integer *ihi, doublereal *scale, doublereal *abnrm, 
	       doublereal *rconde, doublereal *rcondv, doublereal *work, 
	       integer *lwork, integer *iwork, integer *info);
  int dgesv_ (integer *n, integer *nrhs, doublereal *a, integer *lda, 
	      integer *ipiv, doublereal *b, integer *ldb, integer *info);
  int dgeqrf_ (integer *m, integer *n, doublereal *a, integer *lda, 
	       doublereal *tau, double *work, integer *lwork, integer *info);
  int dorgqr_ (integer *m, integer *n, integer *k, doublereal *a, integer *lda, 
	       doublereal *tau, double *work, integer *lwork, integer *info);
}

class Ivector {
 public:
  Ivector (int in = 0);
  Ivector (const Ivector &b);
  ~Ivector () { delete [] elts; }
  int operator[] (int i) const { return elts[i]; }
  Ivector & operator= (const Ivector &b);
  int sum () const;
  bool divides (const Ivector &b) const;
  Ivector operator+ (const Ivector &b) const;
  bool operator== (const Ivector &b) const;
  bool operator< (const Ivector &b) const { return sum() > b.sum(); }

  int n, *elts;
};

class Term {
 public:
  Term () : a(0.0), degree(0), pred(0) {}
  Term (double ia, const Ivector &iexps) : 
    a(ia), exps(iexps),  degree(exps.sum()), pred(0) {}
  double val (double *x) const;
  void set_v (double *x);
  Term der (int v) const;
  Term reduce (int v, double val) const;
  bool operator== (const Term &b) const { return a == b.a && exps == b.exps; }
  bool operator< (const Term &b) const { return degree <= b.degree; }
  Term operator* (const Term &b) const;

  double a, v;
  Ivector exps;
  int degree;
  Term *pred;
};

class Polynomial {
 public:
  Polynomial () 
    : nvars(0), nterms(0), degree(0), dmax(0), terms(0), 
      nflag(false), vflag(false) {}
  Polynomial (int invars, int idmax);
  Polynomial (const Polynomial &p);
  ~Polynomial ();
  void add (double a);
  void add (double a, int e1);
  void add (double a, int e1, int e2);
  void add (double a, int e1, int e2, int e3);
  void add (double a, int e1, int e2, int e3, int e4);
  void add (const Term &t);
  void normalize_coeffs ();
  void normalize ();
  Polynomial & operator= (const Polynomial &b);
  Polynomial operator+ (const Polynomial &b) const;
  Polynomial operator* (const Polynomial &b) const;
  Polynomial scale (double k) const;
  double val (double *x);
  void val_normalize ();
  double coefficient (const Ivector &s) const;
  Polynomial derPrint (int v) const;
  Polynomial der (int v) const;
  Polynomial reduce (int v, double val) const;
  Polynomial reduce2 (int v, double val) const;
  Polynomial deflate2 (double r) const;
  void pack_univariate (double *a) const;
  Polynomial linsub2 (double a, double b, double c, double d,
		      double e, double f) const;
  bool isHorizontal ();
  double distance (const Polynomial &that) const;
  Polynomial homogenize () const;
  Polynomial homogenize (const Polynomial &p) const;
  Polynomial lterms () const;

  int nvars, nterms, degree, dmax;
  Term *terms;
  bool nflag, vflag;
};

Polynomial operator* (double k, const Polynomial &a);

void deflate (double *a, double *b, int d, double r);

Polynomial unpack_univariate (double *a, int d);

Polynomial linear_polynomial (int nvars, double *a);

Polynomial factored_polynomial (int d, double *roots);

void factored_polynomial_aux (int d, double *roots, double *coeffs);

Polynomial bivariate_polynomial (int d, double *a);

bool fit_curve (int d, double *points, Polynomial &p);

double random_number (double minX, double maxX);

Polynomial random_linear_polynomial (int nvars);

Polynomial random_polynomial (int nvars, int dmax, double cmax = 1.0);

int form_monomials (int nvars, int dmax, Ivector *&pp);

int choose (int k, int n);

Polynomial read_polynomial (istream &is);

ostream& operator<< (ostream &os, const Polynomial &poly);

ostream& operator<< (ostream &os, const Term &term);

class Trie {
 public:
  Trie (int in = 0);
  ~Trie ();
  void insert (const Ivector &a, int va, int i = 0);
  int find (const Ivector &a, int i = 0) const;

  int n, v;
  Trie **c;
};

class Eigen {
 public:
  Eigen () : n(0), k(0), val(0.0), vec(0) {}
  ~Eigen () { delete [] vec; }
  Eigen (const Eigen &b);
  void init (int ni, int ki, double ival, double *ivec);
  bool operator< (const Eigen &b) const;
  Eigen & operator= (const Eigen &b);

  int n, k;
  double val, *vec;
};

class Psystem {
 public:
  Psystem (const Polynomial &f);
  Psystem (const Polynomial &f, const Polynomial &g);
  Psystem (const Polynomial &f, const Polynomial &g,
	   const Polynomial &fx, const Polynomial &fy,
	   const Polynomial &gx, const Polynomial &gy);
  Psystem (const Polynomial &f, const Polynomial &g, const Polynomial &h);
  Psystem (const Polynomial &f, const Polynomial &g, const Polynomial &h,
	   const Polynomial &i);
  Psystem (int in, Polynomial *ipolys);
  ~Psystem ();
  void init ();
  void init_jac ();
  int degree () const { return m; }
  int realroots (double *res, double *alleigen = 0);
  Psystem homogenize (const Polynomial &p) const;
  int dehomogenize (Polynomial &p, int nf, double *fres, double *res) const;
  int simpleroots (double *res, double *alleigen = 0);
  void setup_pp ();
  void setup_a () const;
  void setup_p () const;
  bool solve_a () const;
  void setup_ptrie ();
  int simplesolve (double *res, double *alleigen) const;
  double *get_b (int k) const;
  void setup_b (int k, double *b) const;
  void check_eigen (int &nr, Eigen *eigen) const;
  bool adjust_eigen (int nr, Eigen *eigen, int s, int e) const;
  bool root (const Eigen &eigen, double *r) const;
  void root1 (const Eigen &eigen, double *r) const;
  bool root2 (const Eigen &eigen, double *r) const;
  void polish_roots (int nr, double *res);
  double polish_root (double *x) const;
  void val (double *x, double *v) const;
  void dval (double *x, double *dv) const;
  Psystem perturb (double eps);

  Polynomial *polys, *jac;
  Ivector *pp;
  int n, m, N, dmax;
  integer *p;
  double *a, *bdata;
  bool *bflag;
  Trie *ptrie;
};

double max_norm (int n, double *x);

void copy_v (int n, double *v, double *w);

void m_times_v (int m, int n, double *a, double *v, double *b);

void at_times_b_times_a (int m, int n, double *a, double *b, double *c);

// clapack

bool qr_column_pivot (double *a, int m, int n, integer *p);

bool linsolve_triangular (double *a, double *b, int m, int k);

bool linsolve (double *a, double *b, int k);

int real_eigen (double *a, int k, Eigen *&eigen, double *alleigen = 0);

bool orthogonalize (double *a, int nr, int nc);

// debugging

void print_matrix (double *a, int m, int n);

double check_eigenvector (double *a, Eigen &eigen);

void roots_at_infinity (Polynomial &f, Polynomial &g);

void roots_at_infinity (Polynomial &f, Polynomial &g, double v);
